(*************************************************************************)
(* Memory Model in ssreflect                                             *)
(*                                                                       *)
(* Definition of heaps with decidable checks for emptiness, disjointness *)
(* and validity of locations.                                            *)
(*                                                                       *)
(* The exported heap primitives are:                                     *)
(*                                                                       *)
(* locn : Type (locations, including null)                               *)
(* loc  : Type (non-null locations)                                      *)
(* null : locn (the null pointer)                                        *)
(*                                                                       *)
(* upd (h : heap) (x : loc) (d : dynamic) : heap                         *)
(* look (x : locn) (h : heap) : option dynamic                           *)
(* free (x : locn) (h : heap) : heap                                     *)
(* valid (x : locn) (h : heap) : bool                                    *)
(* union (h1 h2 : heap) : heap                                           *)
(* empty : heap                                                          *)
(* is_empty (h : heap) : bool                                            *)
(* def (h : heap) : bool                                                 *)
(* fresh (h : heap) : loc                                                *)
(*                                                                       *)
(* Also exported is an exhaustive set of McCarthy-style axioms for       *)
(* describing all possible interactions between the primitives.          *)
(*************************************************************************)

Require Import ssreflect.
Require Import eqtype.
Require Import ssrnat.
Require Import ssrbool.
Require Import seq.
Require Import Eqdep.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(**************************)
(* Preliminaries          *)
(* (move to another file) *)
(**************************)

Definition dec (b : bool) := decP (@idP b).

(* type dynamic *)

Record dynamic : Type := dyn {typ : Type; val : typ}.

Lemma dyn_inj (A : Type) (x y : A) : dyn x = dyn y -> x = y. 
Proof. move=> A x y [H]; apply: inj_pairT2 H. Qed.

Lemma dyn_eta (d : dynamic) : d = dyn (val d).
Proof. by case. Qed.

Lemma dyn_injT (A1 A2 : Type) (x1 : A1) (x2 : A2) : dyn x1 = dyn x2 -> A1 = A2.
Proof. by move=>A1 A2 x1 x2 [H]. Qed.

(*********************************************)
(* Concrete heaps are finite maps of dynamic *)
(*********************************************)

Inductive heap_c : Type :=
| empty_c : heap_c
| cons_c : nat -> dynamic -> heap_c -> heap_c.

Fixpoint last_loc (xs : heap_c) : nat :=
  match xs with 
  | empty_c => 0
  | cons_c x d h => x
  end.

Fixpoint sorted_c (xs : heap_c) : bool :=
  match xs with 
  | empty_c => true
  | cons_c x d ys => (x > (last_loc ys)) && sorted_c ys (* sorted heaps can't have null locs *)
  end.   

(********************************)
(* Operations on concrete heaps *)
(********************************)

Fixpoint update_c (xs : heap_c) (x : nat) (d : dynamic) {struct xs} : heap_c :=
  match xs with 
    empty_c => cons_c x d empty_c
  | cons_c y d1 ys => 
      if (x > y) then cons_c x d xs else
        if (x == y) then cons_c x d ys else
          cons_c y d1 (update_c ys x d)
  end.

Fixpoint valid_c (x : nat) (xs : heap_c) {struct xs} : bool :=
  match xs with
    empty_c => false
  | cons_c y d ys => if x == y then true else valid_c x ys
  end.

Fixpoint free_c (x : nat) (xs : heap_c) {struct xs} : heap_c :=
  match xs with
    empty_c => empty_c
  | cons_c y d ys => if x == y then ys else cons_c y d (free_c x ys)
  end.

Fixpoint lookup_c (x : nat) (xs : heap_c) {struct xs} : dynamic :=
  match xs with
    empty_c => dyn tt
  | cons_c y d ys => if x == y then d else lookup_c x ys
  end.

Fixpoint union2_c (xs1 xs2 : heap_c) {struct xs2} : heap_c :=
  match xs2 with
    empty_c => xs1
  | cons_c y d ys => update_c (union2_c xs1 ys) y d
  end.

Fixpoint disjoint2_c (xs1 xs2 : heap_c) {struct xs2} : bool :=
  match xs2 with
    empty_c => true
  | cons_c y _ ys => (~~ (valid_c y xs1)) && disjoint2_c xs1 ys
  end.

Definition is_empty_c (xs : heap_c) : bool := 
  match xs with
    empty_c => true
  | _ => false
  end.

Fixpoint subdom_c (h1 h2 : heap_c) {struct h2} : bool :=
  match h1, h2 with
    empty_c, _ => true
  | cons_c x1 _ t1, cons_c x2 _ t2 => 
     if x1 > x2 then false
     else if x1 == x2 then subdom_c t1 t2
     else subdom_c h1 t2
  | _, _ => false
  end.

Definition pick_c (h : heap_c) : nat :=
  if h is cons_c n _ _ then n else 0. 

(*************************)
(* Lemmas about last_loc *)
(*************************)

Lemma last_update (xs : heap_c) (x : nat) (d : dynamic) : 
        x > last_loc xs -> update_c xs x d = cons_c x d xs.
Proof. by case=> [|l a xs] //= x d [->]. Qed. 

Lemma last_free (xs : heap_c) (x : nat) : 
        x > last_loc xs -> sorted_c xs -> free_c x xs = xs.
Proof.
elim=>[|l a xs IH] //= x H1; case/andP=>H2 H3.
case: eqP H1 => [->|_]; first by rewrite ltnn.
by move/(ltn_trans H2)=> H1; rewrite IH.
Qed.

Lemma last_valid (xs : heap_c) (x : nat) : 
        x > last_loc xs -> sorted_c xs -> ~~ valid_c x xs.
Proof.
elim=>[|l a xs IH] //= x H1; case/andP=>H2 H3.
case: eqP H1 => [->|_]; first by rewrite ltnn.
by move/(ltn_trans H2)=> H1; rewrite IH.
Qed.

Lemma last_update_lt (xs : heap_c) (x : nat) (d : dynamic) : 
        x <= last_loc xs -> last_loc (update_c xs x d) = last_loc xs.
Proof.
case=>[|l a xs] //= x d; first by rewrite leqn0; move/eqP.
by case: ltngtP=> //=; rewrite ltnNge; move/negbET=> ->.
Qed.

Lemma last_free_lt (xs : heap_c) (x : nat) : 
        sorted_c xs -> last_loc (free_c x xs) <= last_loc xs.
Proof. 
case=>[|l a xs] //= x; case/andP=>H1 H2.
case: eqP=> _ //; by apply ltnW. 
Qed. 

(*************************************************************)
(* Compatibility lemmas: heap operations preserve sortedness *)
(*************************************************************)

Lemma update_compat (xs : heap_c) (x : nat) (d : dynamic) : 
        sorted_c xs -> x != 0 -> sorted_c (update_c xs x d).
Proof.
elim=>[|l a xs IH] //= x d; first by move=>_; move/(elimT negPf); move/neq0_lt0n=> -> //.
case/andP=> H1 H2 H3. 
case: ltngtP=> H4 //=; try rewrite H4 H1 H2 //.
rewrite IH //.
case H5: (x <= last_loc xs); first by rewrite last_update_lt // H1 //.
rewrite last_update //=; first by rewrite H4 //.
by rewrite ltnNge H5.
Qed.
 
Lemma free_compat (xs : heap_c) (x : nat) : 
        sorted_c xs -> sorted_c (free_c x xs).
Proof.
elim=>[|l a xs IH] //= x; case/andP=> H1 H2.
case: ltngtP=> H3 //=; rewrite IH //= andb_comm //=.
- by [apply: leq_ltn_trans H1; apply: last_free_lt=> //].
rewrite last_free //; apply: ltn_trans H1 H3.
Qed.

Lemma union2_compat (xs1 xs2 : heap_c) : 
        sorted_c xs1 -> sorted_c xs2 -> sorted_c (union2_c xs1 xs2).
Proof.
move=>xs1 xs2; elim: xs2 xs1=>[|l a xs2 IH] //= xs1 H1; case/andP=> H2 H3. 
apply: update_compat; first by apply IH; auto.
by case: eqP H2=>[->|].
Qed. 

(********************************)
(* Properties of concrete heaps *)
(********************************)

Lemma update_inj_c (h : heap_c) (x : nat) (d1 d2 : dynamic) :
        update_c h x d1 = update_c h x d2 -> d1 = d2.
Proof.
elim=>[|l a xs IH] //= x d1 d2; first by move=>[->].
by case: ltngtP=>eq1 [eq2]; rewrite ?(IH _ _ _ eq2).
Qed.

Lemma update_eta_c (h : heap_c) (x : nat) : 
        valid_c x h -> sorted_c h -> h = update_c h x (lookup_c x h).
Proof.
elim=>[|l a xs IH] //= x.
case: ltngtP=>[H1 H2|H1 H2|->] //; move/andP=>[H3 H4].
- by rewrite -IH.
have H5 : last_loc xs < x by apply: ltn_trans H1.
by move: H2 (last_valid H5 H4)=>->.
Qed. 

Lemma update_update_eq_c (xs : heap_c) (x : nat) (d1 d2 : dynamic) : 
        update_c (update_c xs x d1) x d2 = update_c xs x d2.
Proof.
elim=> [|l a xs IH] //= x d1 d2; do 3 case: ltngtP=> //=; try rewrite ltnn //.
by rewrite IH.
Qed.

Lemma update_update_neq_c (xs : heap_c) (x y : nat) (d1 d2 : dynamic) :
        x != y -> sorted_c xs -> 
          update_c (update_c xs x d1) y d2 = update_c (update_c xs y d2) x d1.
Proof.
elim=> [| l a xs IH] //= x y d1 d2 //=; first by do 2 case: ltngtP=> //=.
move=> H1; case/andP=> H2 H3.
case: eqP H2=>[<-|] //=; try rewrite ltnn.
- case: eqP (H1) => _ _ H2 //=; rewrite eq_sym. 
  move/negbET: (H1) => -> //=.
  by do 2 case: ltngtP => //=; try rewrite ltnn //= eq_refl.
move/(introN eqP)=> H4 H5; move: (H1) (H4).
do 6 case: ltngtP=> //=; try rewrite (negbET H1); 
  try rewrite (negbET H4); try rewrite IH //=; move=> t1 t2 t3 t4 t5 _ _ _.
- by move: (ltn_trans t1 t4) t3; case: ltngtP.
- by move: t3 t1 t5 => [->]; case: ltngtP.
- by move: (ltn_trans t1 t5) t4; case: ltngtP.
by move: t4 t1 t5 => [->]; case: ltngtP.
Qed.

Lemma valid_update_eq_c (xs : heap_c) (x : nat) (d : dynamic) : valid_c x (update_c xs x d).
Proof. 
by elim=> [|l a xs IH] /= *; [rewrite eq_refl | 
                             do 2 case: ltngtP=> //=; rewrite ltnn]. 
Qed.

Lemma valid_update_neq_c (xs : heap_c) (x y : nat) (d : dynamic) : 
        x != y -> valid_c x xs = valid_c x (update_c xs y d).
Proof.
elim=> [|l a xs IH] //= x y d H1; first by rewrite (negbET H1). 
case: eqP=>[<-|].
- rewrite eq_sym; case: ltngtP (H1) => //=; rewrite eq_refl //.
  by rewrite (negbET H1).
case: ltngtP=> //=; try rewrite (negbET H1) //; case: eqP=> //.
by rewrite (IH x y d).
Qed.

Lemma free_empty_c (x : nat) : free_c x empty_c = empty_c.
Proof. trivial. Qed.

Lemma valid_free_eq_c (xs : heap_c) (x : nat) : sorted_c xs -> ~~ valid_c x (free_c x xs).
Proof.
elim=>[|l a xs IH] //= x; case/andP=> H1 H2.
case: eqP (IH x H2) H1=>[<-|] => [H1 H3|] //=; first by rewrite last_free in H1.
by case: eqP.
Qed.

Lemma valid_free_neq_c (xs : heap_c) (x y : nat) : 
        x != y -> valid_c x (free_c y xs) = valid_c x xs.
Proof. 
by elim=>[|l a xs IH] //= x y H; move: H (IH x y H); 
   do 3 case: eqP=> //=; move=> -> ->. 
Qed.

Lemma valid_free_weaken_c (xs : heap_c) (x y : nat) : 
        sorted_c xs -> valid_c x (free_c y xs) -> valid_c x xs.
Proof.
  move => xs x y s v; case H: (x == y).
  - by rewrite (eqP H) in v; move: (valid_free_eq_c y s) => ns; rewrite v in ns.
  - by rewrite valid_free_neq_c in v; last apply negbT.
Qed.

Lemma valid_null_c (xs : heap_c) : sorted_c xs -> ~~ valid_c 0 xs.
Proof.
elim=>[|l a xs IH] //=; move/andP=>[H1 H2]; move: H1.
case: eqP=>[<-|*] //=.
by apply: IH.
Qed.

Lemma free_update_eq_c (xs : heap_c) (x : nat) (d : dynamic) : 
        sorted_c xs -> free_c x (update_c xs x d) = free_c x xs.
Proof.
elim=> [|l a xs IH] //= x d; first by rewrite eq_refl.
case/andP=> H1 H2.
do 2 case: ltngtP=> //=; try rewrite ltnn //=; try rewrite (IH x d H2) //.
move=> _ H3; rewrite last_free //=.
apply: ltn_trans H1 H3.
Qed.

Lemma free_update_neq_c (xs : heap_c) (x y : nat) (d : dynamic) : 
        x != y -> sorted_c xs -> free_c x (update_c xs y d) = update_c (free_c x xs) y d.
Proof.
elim=> [|l a xs IH] //= x y d H; first by rewrite (negbET H) //. 
case/andP=> H1 H2. 
case: ltngtP=>[||<-] //=; try rewrite (negbET H) //=; try rewrite ltnn eq_refl //.
- case: eqP=>//=; case: ltngtP=>//.
  by rewrite IH.
case: eqP=>//= => _ H4; last by rewrite H4.
rewrite last_update //.
apply: ltn_trans H1 H4.
Qed.

Lemma update_free_eq_c (xs : heap_c) (x : nat) (d : dynamic) : 
        sorted_c xs -> update_c (free_c x xs) x d = update_c xs x d.
Proof.
elim=>[|l a xs IH] //= x d; move/andP=>[H1 H2]. 
case: ltngtP=>[||->] /=; try case: ltngtP=>// H3 _.
- by rewrite IH.
- by rewrite last_free //; apply: ltn_trans H3.
by rewrite last_update.
Qed.

Lemma free_valid_neq_c (xs : heap_c) (x : nat) : ~~ valid_c x xs -> free_c x xs = xs.
Proof. by elim=>[|l a xs IH] //= x; case: eqP=> _ H //=; rewrite IH. Qed.

Lemma free_valid_eq_c (xs : heap_c) (x : nat) : free_c x xs = xs -> ~~ valid_c x xs.
Proof.
elim=>[|l a xs IH] //= x; case: eqP=>[<- H|H]//=; last by move=>[]; auto.
elim: {l IH} xs H=>[| l1 a1 xs1 IH1] //=.
by move=>[->->].
Qed.

Lemma lookup_empty_c (x : nat) : lookup_c x empty_c = dyn tt. 
Proof. trivial. Qed.

Lemma lookup_update_eq_c (xs : heap_c) (x : nat) (d : dynamic) : 
        lookup_c x (update_c xs x d) = d.
Proof.
elim=> [|l a xs IH] //= x d; first by rewrite eq_refl.
case: eqP=>[->|]; first by rewrite ltnn //= eq_refl.
case: ltngtP=> //=; last by rewrite eq_refl.
by case: eqP.
Qed.

Lemma lookup_update_neq_c (xs : heap_c) (x y : nat) (d : dynamic) : 
        x != y -> lookup_c x (update_c xs y d) = lookup_c x xs.
Proof.
elim=> [|l a xs IH] //= x y d H; first by rewrite (negbET H).
case: ltngtP=>[||<-] //=; try rewrite (negbET H) //.
by rewrite IH.
Qed.

Lemma union2_update_right_c (xs1 xs2 : heap_c) (x : nat) (d : dynamic) : 
        sorted_c xs1 -> sorted_c xs2 -> 
          union2_c xs1 (update_c xs2 x d) = update_c (union2_c xs1 xs2) x d. 
Proof.
move=> x1 x2; elim: x2 x1=> [|l a xs2 IH] //= => x1 x d H1; move/andP=> [H2 H3]. 
case: ltngtP=>[H4||<-] //=; last by rewrite update_update_eq_c.
rewrite IH // update_update_neq_c //=; last by apply: union2_compat.
by case: eqP H4=> // => ->; rewrite ltnn.
Qed.

Lemma union2_update_left_c (xs1 xs2 : heap_c) (x : nat) (d : dynamic) : 
        sorted_c xs1 -> sorted_c xs2 -> ~~ valid_c x xs2 -> 
          union2_c (update_c xs1 x d) xs2 = update_c (union2_c xs1 xs2) x d.
Proof.
move=> x1 x2; elim: x2 x1=> [|l a xs2 IH] //= x1 x d H1; move/andP=> [H2 H3].
case: eqP=> H4 H5 //; rewrite IH // update_update_neq_c //; first by case: eqP.
by apply: union2_compat.
Qed.

Lemma disjoint2_valid_c (xs1 xs2 : heap_c) (x : nat) : 
        valid_c x xs1 -> valid_c x xs2 -> ~~ disjoint2_c xs1 xs2. 
Proof.
move=> x1 x2; elim: x2 x1=> [|l a xs2 IH] //= x1 x.
case: eqP=> [<- ->|H1 H2 H3] //.
by rewrite negb_and (IH x1 x) // orb_comm.
Qed.

Lemma disjoint2_update_right_c (xs1 xs2 : heap_c) (x : nat) (d : dynamic) : 
        ~~ valid_c x xs1 -> disjoint2_c xs1 (update_c xs2 x d) = disjoint2_c xs1 xs2.
Proof.
move=> x1 x2; elim: x2 x1=>[| l a xs IH] /= xs1 x d=>[->|] //.
case: ltngtP=> //= [_ H2|H1 ->|->] //.
by rewrite IH.
Qed.

Lemma update_not_empty_c (xs : heap_c) (x : nat) (d : dynamic) : update_c xs x d <> empty_c.
Proof. by elim=> [|l a xs IH] //= x d; case: ltngtP. Qed.

(* ring properties of union2_c and disjoint2_c *)

Lemma union2_empty_right_c (xs : heap_c) : union2_c xs empty_c = xs.
Proof. trivial. Qed.

Lemma union2_empty_left_c (xs : heap_c) : sorted_c xs -> union2_c empty_c xs = xs.
Proof.
elim=> [|l a xs IH] //=; move/andP=> [H1 H2].
by rewrite IH // last_update.
Qed.

Lemma union2_commute_c (xs1 xs2 : heap_c) : 
        sorted_c xs1 -> sorted_c xs2 -> disjoint2_c xs1 xs2 -> 
          union2_c xs1 xs2 = union2_c xs2 xs1.
Proof.
move=> xs1 xs2; elim: xs2 xs1=> [|l a xs2 IH] //= xs1 H; first by rewrite union2_empty_left_c.
move/andP=>[H1 H2]; move/andP=>[H3 H4].
by rewrite -last_update // union2_update_left_c // IH.
Qed.

Lemma union2_assoc_c (xs1 xs2 xs3 : heap_c) : 
        sorted_c xs1 -> sorted_c xs2 -> sorted_c xs3 ->
          union2_c (union2_c xs1 xs2) xs3 = union2_c xs1 (union2_c xs2 xs3).
Proof.
move=> xs1 xs2 xs3; elim: xs3 xs1 xs2 => [|l a xs3 IH] //= xs1 xs2 H1 H2; move/andP=>[H3 H4].
rewrite union2_update_right_c //; last by apply: union2_compat.
by rewrite IH.
Qed.

Lemma disjoint2_empty_left_c (xs : heap_c) : disjoint2_c empty_c xs.
Proof. by elim. Qed.
   
Lemma disjoint2_empty_right_c (xs : heap_c) : disjoint2_c xs empty_c.
Proof. trivial. Qed.

Lemma disjoint2_sym_c (xs1 xs2 : heap_c) : disjoint2_c xs1 xs2 = disjoint2_c xs2 xs1.
Proof.
elim=>[|l1 a1 xs1 IH1]; elim=>[|l2 a2 xs2 IH2] //=; first by rewrite disjoint2_empty_left_c.
rewrite eq_sym; case: eqP=>_ //.
rewrite IH2 -IH1 //= IH1.
bool_congr.
Qed.

Lemma valid_union2_c (xs1 xs2 : heap_c) (x : nat) : 
        valid_c x (union2_c xs1 xs2) = valid_c x xs1 || valid_c x xs2.
Proof.
move=> xs1 xs2; elim: xs2 xs1 => [|l a xs2 IH] //= xs1 x; first by rewrite orb_comm.
case: eqP=>[->|H] //=; first by rewrite valid_update_eq_c orb_comm.
rewrite -valid_update_neq_c //.
by case: eqP.
Qed.

Lemma valid_disjoint2_c (xs1 xs2 : heap_c) (x : nat) : 
        disjoint2_c xs1 xs2 -> valid_c x xs1 -> ~~ valid_c x xs2.
Proof.
move=>xs1 xs2; elim: xs2 xs1=>[|l a xs2 IH] //= xs1 x; move/andP=>[H1 H2] H3.
case: eqP H3 H1=>[->->|] //=. 
firstorder.
Qed.

Lemma disjoint2_union2_c (xs xs1 xs2 : heap_c) :
        disjoint2_c (union2_c xs1 xs2) xs = disjoint2_c xs1 xs && disjoint2_c xs2 xs.
Proof. elim => [|l a xs IH] //= xs1 xs2; rewrite IH valid_union2_c negb_or; ring. Qed.

(* cancelation properties *)

Lemma update_cancel_c (xs1 xs2 : heap_c) (x : nat) (d : dynamic) :
        update_c xs1 x d = update_c xs2 x d -> 
          ~~ valid_c x xs1 -> ~~ valid_c x xs2 -> xs1 = xs2.
Proof.
elim=>[|l1 a1 xs1 IH1]; case=>[|l2 a2 xs2] //= x d //=; try by case: ltngtP=> // H1 [H2]; 
  rewrite H2 ltnn in H1.
do 2 case: ltngtP=> //=; move=> H3 H4 [H5 H6 H7] *.
- by rewrite H5 H6 (IH1 xs2 x d).
- by rewrite H5 ltnn in H4.
- by rewrite H5 ltnn in H3.
by rewrite H5 H6 H7.
Qed.

Lemma union2_cancel_right_c (xs xs1 xs2 : heap_c) : 
        sorted_c xs -> union2_c xs1 xs = union2_c xs2 xs -> 
          disjoint2_c xs1 xs -> disjoint2_c xs2 xs -> xs1 = xs2. 
Proof.
elim=>[|l a xs IH] //= xs1 xs2; move/andP=>[H1 H2] // H; move/andP=>[H3 H4]; 
  move/andP=>[H5 H6].
apply: IH=>//.
by apply: (update_cancel_c H); 
rewrite valid_union2_c negb_or [~~ _ xs] last_valid // andb_comm.
Qed.

Lemma union2_cancel_left_c (xs xs1 xs2 : heap_c) : 
        sorted_c xs -> sorted_c xs1 -> sorted_c xs2 -> 
          union2_c xs xs1 = union2_c xs xs2 -> 
            disjoint2_c xs xs1 -> disjoint2_c xs xs2 -> xs1 = xs2.
Proof.
move=> xs xs1 xs2 H1 H2 H3; rewrite 2![disjoint2_c xs _]disjoint2_sym_c => H4 H5 H6.
apply: (union2_cancel_right_c H1)=>//.
by rewrite union2_commute_c // H4 union2_commute_c // disjoint2_sym_c.
Qed.

(* Some further McCarthy-style properties *)

(*
Lemma lookup_union2_none_intro_c (xs1 xs2 : heap_c) (x : nat) : 
        sorted_c xs1 -> sorted_c xs2 -> 
          lookup_c x xs1 = None -> lookup_c x xs2 = None -> lookup_c x (union2_c xs1 xs2) = None.
Proof.
move=>xs1 xs2; elim: xs2 xs1=>[|l a xs2 IH] //= xs1 x H1; move/andP=>[H3 H4] H5.
case: eqP=>[|H6 H7]//.
rewrite -union2_update_left_c //; last by apply:last_valid.
apply: IH=>//; first by apply update_compat=>//; case: eqP H3=>[->|].
by rewrite lookup_update_neq_c //; case: eqP.
Qed.

Lemma lookup_union2_none_elim_c (xs1 xs2 : heap_c) (x : nat) : 
        sorted_c xs1 -> sorted_c xs2 -> 
          lookup_c x (union2_c xs1 xs2) = None -> lookup_c x xs1 = None /\ lookup_c x xs2 = None.
Proof.
move=>xs1 xs2; elim: xs2 xs1=>[|l a xs2 IH] xs1 x //= H1; move/andP=>[H2 H3].
case: eqP=>[->|H4]//; first by rewrite lookup_update_eq_c.
rewrite lookup_update_neq_c; last by case: eqP.
by apply: IH.
Qed.
*)

Lemma lookup_union2_some_intro1_c (xs1 xs2 : heap_c) (x : nat) : 
        sorted_c xs1 -> sorted_c xs2 -> ~~ valid_c x xs2 -> 
          lookup_c x (union2_c xs1 xs2) = lookup_c x xs1.
Proof.
move=>xs1 xs2; elim: xs2 xs1=>[|l a xs2 IH] //= xs1 x d; move/andP=>[H2 H3].
case: eqP=>// H5 H6.
rewrite lookup_update_neq_c; last by case: eqP.
by apply: IH.
Qed.

Lemma lookup_union2_some_intro2_c (xs1 xs2 : heap_c) (x : nat) : 
        sorted_c xs1 -> sorted_c xs2 -> valid_c x xs2 -> 
          lookup_c x (union2_c xs1 xs2) = lookup_c x xs2.
Proof.
move=>xs1 xs2; elim: xs2 xs1=>[|l a xs2 IH] //= xs1 x H1; move/andP=>[H2 H3].
case: eqP=>[->|H4]//=; first by rewrite lookup_update_eq_c.
rewrite lookup_update_neq_c; last by case: eqP.
exact: IH.
Qed.

Lemma lookup_union2_some_elim_c (xs1 xs2 : heap_c) (x : nat) (d : dynamic) : 
        sorted_c xs1 -> sorted_c xs2 -> valid_c x (union2_c xs1 xs2) ->
          lookup_c x (union2_c xs1 xs2) = d -> 
            (lookup_c x xs1 = d /\ ~~ valid_c x xs2) \/ lookup_c x xs2 = d.
Proof.
move=>xs1 xs2; elim: xs2 xs1=>[|l a xs2 IH] //= xs1 x d H1; try tauto; move/andP=>[H2 H3].
case: eqP=>[->|H4]//; first by rewrite lookup_update_eq_c; tauto. 
rewrite -valid_update_neq_c; last by case: eqP. 
rewrite lookup_update_neq_c; last by case: eqP.
by apply: IH.
Qed.

Lemma is_empty_update_c (xs : heap_c) (x : nat) (d : dynamic) : 
        ~~ is_empty_c (update_c xs x d).
Proof. by elim=>[|l a xs IH] x d //=; case: ltngtP. Qed.

Lemma valid_lookup_c (xs : heap_c) (x : nat) : ~~ valid_c x xs -> lookup_c x xs = dyn tt.
Proof. by elim=>[|l a xs IH] //= x; case: eqP=>// _; apply: IH. Qed.

(*
Lemma valid_lookup_P_c (xs : heap_c) (x : nat) : reflect (exists d, lookup_c x xs = Some d) (valid_c x xs).
Proof.
elim=>[|l a xs IH] //= x; first by constructor; move=>[d H].
case: eqP=>_; [by constructor; exists a | apply: IH].
Qed.
*)

Lemma valid_disjoint2_P1_c (xs1 xs2 : heap_c) : 
        reflect (forall x, valid_c x xs1 -> ~~ valid_c x xs2) (disjoint2_c xs1 xs2).
Proof.
move=>xs1 xs2; elim: xs2 xs1=>[|l a xs2 IH] //= xs1; first by constructor.
rewrite andb_comm; case:IH=>H1/=.
- case H2:(valid_c _ _)=> //=; constructor.
  - by move/(_ l H2); rewrite eq_refl.
  by move=> x H3; move: H2 H3 (H1 x H3); case: eqP=>[->->|].
constructor.
move=> H3; apply: H1; move=> x H1; move: {a xs1 H1 H3} (H3 x H1).
by case: eqP.
Qed.

Lemma valid_disjoint2_PN_c (xs1 xs2 : heap_c) : 
        reflect (exists x, valid_c x xs1 /\ valid_c x xs2) (~~ disjoint2_c xs1 xs2).
Proof.
move=>xs1 xs2; elim: xs2 xs1=>[|x d xs2 IH] xs1 /=; first by right=>[[x [H]]].
case E1: (valid_c x xs1)=>/=; first by left; exists x; rewrite E1 eq_refl.
case E2: (~~ disjoint2_c xs1 xs2).
- left; case: IH E2=>// [[y [H1 H2]]] _.
  exists y; split=>//. 
  by rewrite H2; case: eqP.
right; case: IH E2=>// H _ [y [H1 H2]]; elim: H.
case: eqP H1 H2=>[->|_ H1 H2]; first by rewrite E1.
by exists y.
Qed.   

Lemma free_free_eq_c (xs : heap_c) (x : nat) :  
        sorted_c xs -> free_c x (free_c x xs) = free_c x xs.
Proof.
elim=>[|l a xs IH] //= x; case/andP=>[H1 H2]. 
case: eqP H1=>[-> H1|H1 H3] /=; first by rewrite last_free.
case: eqP H1=>[->|H1 _]//=.
by rewrite IH.
Qed.

Lemma free_free_neq_c (xs : heap_c) (x y : nat) : 
        x != y -> free_c x (free_c y xs) = free_c y (free_c x xs).
Proof.
elim=>[|l a xs IH] //= x y H. 
case: eqP=>[<-|]//=; first by case: eqP H=>//=; rewrite eq_refl.
do 2 case: eqP=>//=.
by rewrite IH.
Qed.

Lemma lookup_free_eq_c (xs : heap_c) (x : nat) : 
        sorted_c xs -> lookup_c x (free_c x xs) = dyn tt. 
Proof.
elim=>[|l a xs IH] //= x; case/andP=>[H1 H2].
case: eqP H1=>[-> H1|] //=; first by rewrite -(IH l) //= last_free.
case: eqP=>//= => _ H1 H3.
by apply: IH.
Qed.

Lemma lookup_free_neq_c (xs : heap_c) (x y : nat) : 
        x != y -> lookup_c x (free_c y xs) = lookup_c x xs.
Proof.
elim=>[|l a xs IH] //= x y H.
case: eqP=>[<-|]//=; first by case: eqP H.
case: eqP=>//= H1 H2.
by apply: IH.
Qed.

(*
Lemma lookup_disjoint2_c (xs1 xs2 : heap_c) (x : nat) (d : dynamic) : 
        disjoint2_c xs1 xs2 -> lookup_c x xs1 = Some d -> lookup_c x xs2 = None.
Proof.
move=>xs1 xs2; elim: xs2 xs1=>[|l a xs1 IH] //= xs2 x d.
case: valid_lookup_P_c=>//=; move=>H1 H2 H3. 
case: eqP H1=>[<-|]; by firstorder.
Qed.
*)

Lemma free_union2_c (xs1 xs2 : heap_c) (x : nat) : 
        sorted_c xs1 -> sorted_c xs2 -> 
          free_c x (union2_c xs1 xs2) = union2_c (free_c x xs1) (free_c x xs2).
Proof.
move=>xs1 xs2; elim: xs2 xs1=>[|l a xs2 IH] //= xs1 x H1; case/andP=>[H2 H3].
case: eqP=>[->|H4]//=. 
- rewrite free_update_eq_c //=; last by apply: union2_compat.
  by rewrite IH // (last_free H2) //.
rewrite free_update_neq_c //=; last by apply: union2_compat. 
- by rewrite IH.
by case: eqP.
Qed.

Lemma disjoint2_free_left_eq_c (xs1 xs2 : heap_c) (x : nat) : 
        ~~ valid_c x xs2 -> disjoint2_c (free_c x xs1) xs2 = disjoint2_c xs1 xs2.
Proof.
move=>xs1 xs2; elim: xs2 xs1=>[|l a xs2 IH] //= xs1 x.
case: eqP=>//= => H1 H2. rewrite valid_free_neq_c ; last by case: eqP H1=>[->|].
by rewrite IH. 
Qed.

Lemma disjoint2_free_left_c (xs1 xs2 : heap_c) (x : nat) : 
        sorted_c xs1 -> disjoint2_c xs1 xs2 -> disjoint2_c (free_c x xs1) xs2.
Proof.
move=>xs1 xs2; elim: xs2 xs1=>[|l a xs2 IH] //= xs1 x s. move/andP => [v1 dis1].
rewrite IH // andbT. move: v1. apply contra.
by apply: valid_free_weaken_c.
Qed.

Lemma is_empty_union2_c (xs1 xs2 : heap_c) : 
        is_empty_c (union2_c xs1 xs2) = is_empty_c xs1 && is_empty_c xs2.
Proof.
move=>xs1 xs2; case: xs2 xs1=>[|l a xs2] xs1 //=; rewrite andb_comm //=. 
apply/idPn; apply: is_empty_update_c.
Qed.

(*********************)
(* Reflecting subdom *)
(*********************)

Lemma subdomP_c (h1 h2 : heap_c) : 
        sorted_c h1 -> sorted_c h2 -> 
          reflect (forall x, valid_c x h1 -> valid_c x h2) 
                  (subdom_c h1 h2).
Proof.
move=>h1 h2 H1 H2; case E: (subdom_c h1 h2); constructor.
- elim: h2 h1 H1 H2 E=>[[]//|x2 d2 h2 IH] [//|x1 d1 h1] H1 /=.
  case/andP=>H2 H3.
  case: ltngtP H1 H2 H3=>[H4 H1||<- /= H1]// H2 H3 H5 x H6.
  - by case: eqP=>// H7; apply: (IH _ H1 H3 H5 x H6).
  case/andP: H1=>H7 H8.
  case: (x =P x1) H6=>// _ H6.
  by apply: IH H6.
elim: h2 h1 H1 H2 E=>[|x2 d2 h2 IH].
- case=>[|x1 d1 h1] //=; case/andP=>_ _ _ _.
  by move/(_ x1); rewrite eq_refl=>H; auto.
case=>[|x1 d1 h1] // H1 /=.
case/andP=>H2 H3.
case: ltngtP H2=>[H4 H2 H5|H4 H2 _|<- H2 H5] H6.
- elim: (IH _ H1 H3 H5)=>x H7.
  case: (x =P x2) H7=>[-> /=|H8 H9].
  - case: eqP H4 H1=>[->|_ H4 /=]; first by rewrite ltnn.
    case/andP=>H7 H8 H9.
    move: (last_valid (ltn_trans H7 H4) H8).
    by rewrite H9.
  by case: eqP H8 (H6 x H9)=>[->|].
- case: (x1 =P x2) H4 (H6 x1)=>[->|_]; first by rewrite ltnn.
  rewrite eq_refl=>H4 H5; have {H5} H5: valid_c x1 h2 by apply: H5.
  move: (last_valid (ltn_trans H2 H4) H3).
  by rewrite H5.
simpl in H1. case/andP: H1=>H1 H4.
elim: (IH _ H4 H3 H5)=>x H7.
case: (x =P x1) H7 (H6 x)=>[-> H7 _|]; last by move=>_ T; apply.
by move: (last_valid H1 H4); rewrite H7.
Qed. 

Lemma subdomE_c (h1 h2 h : heap_c) : 
        sorted_c h1 -> sorted_c h2 -> sorted_c h -> 
          disjoint2_c h1 h -> disjoint2_c h2 h -> 
            subdom_c h1 h2 -> subdom_c (union2_c h1 h) (union2_c h2 h).
Proof.
move=>h1 h2 h H1 H2 H3 H4 H5; case: subdomP_c=>// H _.
case: subdomP_c=>//; try by apply: union2_compat.
elim=>x; rewrite !valid_union2_c.
case/orP=>[|->]; last by apply: orbT.
by move/H=>->.
Qed.   

Lemma subdom_disjoint2_c (h1 h2 h : heap_c) : 
        sorted_c h1 -> sorted_c h2 -> sorted_c h -> 
        subdom_c h1 h2 -> disjoint2_c h2 h -> disjoint2_c h1 h.
Proof.
move=>h1 h2 h H1 H2 H.
case: subdomP_c=>// F H3 H4.
case: valid_disjoint2_P1_c=>//.
elim=>x H5.
move: (F x H5)=>H6.
case: valid_disjoint2_P1_c H4=>// H7 _.
by apply: H7.
Qed.   

Lemma subdom_union2_c (h1 h2 h1' h2' : heap_c) : 
        sorted_c h1 -> sorted_c h2 -> sorted_c h1' -> sorted_c h2' -> 
        subdom_c h1 h2 -> subdom_c h1' h2' -> 
        disjoint2_c h2 h2' -> subdom_c (union2_c h1 h1') (union2_c h2 h2').
Proof.
move=>h1 h2 h1' h2' H1 H2 H3 H4 H5 H6 H7.
case: subdomP_c=>//; try by apply: union2_compat.
elim=>x.
rewrite !valid_union2_c.
case/orP=>H8.
- by case: subdomP_c H5=>// H9 _; rewrite H9.
by case: subdomP_c H6=>// H9 _; rewrite H9 ?orbT.
Qed.

Lemma subdom_emp_c (h : heap_c) : 
        sorted_c h -> subdom_c empty_c h.
Proof. by case. Qed.

Lemma subdom_emp_inv_c (h : heap_c) : 
        sorted_c h -> subdom_c h empty_c -> h = empty_c.
Proof. by case. Qed. 

(**********************)
(* Properties of pick *)
(**********************)

Lemma pickP_c (h : heap_c) : ~~ is_empty_c h = valid_c (pick_c h) h.
Proof. by case=>[|n d h] //=; rewrite eq_refl. Qed.

(**********************************)
(* Locations are an equality type *)
(**********************************)

(* locs are isomorphic to nat; wrap into    *)
(* inductive, to get the explicit inclusion *)
Inductive loc : Type := loc_from_nat : nat -> loc.

Definition null : loc := loc_from_nat 0.

Definition loc_to_nat (x : loc) : nat := 
  match x with loc_from_nat y => y end.

Definition eqloc (x y : loc) : bool :=
  match x, y with loc_from_nat m, loc_from_nat n => m == n end.

Lemma eqlocP : reflect_eq eqloc.
Proof. 
move=>[x][y] /=. 
case: eqP=>[->|*];constructor=>//.
by move=>[*].
Qed.

Canonical Structure loc_eqMixin := EqMixin eqlocP.
Canonical Structure loc_eqType := EqClass loc_eqMixin. 

(*******************************************)
(* Positive locations are an equality type *)
(*******************************************)

(* Some pointer arithmetic: offsetting from a base pointer *)
Definition loc_offset (x : loc) (i : nat) : loc := loc_from_nat (loc_to_nat x + i).

Notation "x .+ i" := (loc_offset x i) (at level 20).

Lemma offset0 (x : loc) : x .+ 0 = x.
Proof. by move=>[x]; rewrite /loc_offset addn0. Qed. 

Lemma loc_inv x y : (x == y) = (loc_to_nat x == loc_to_nat y).
Proof. by move=>[x][y]. Qed. 

Lemma loc_inj x y : loc_to_nat x = loc_to_nat y -> x = y.
Proof. by move=>[x][y] /= ->. Qed. 
 
Lemma offset_null x m : (x .+ m == null) = (m == 0) && (x == null).
Proof. by move=>[x] m; rewrite !loc_inv /= eqn_add0 andbC. Qed.

Lemma offsetA (x : loc) (i j : nat) : x .+ i .+ j = x .+ (i + j).
Proof. by move=>[x] i j;  rewrite /loc_offset addnA. Qed.

Lemma offsetK (x : loc) (i j : nat) : (x .+ i == x .+ j) = (i == j).
Proof.
move=>[x] i j; rewrite /loc_offset /=.
case: eqP; first by case; move/eqP; rewrite eqn_addl=>->.
by case: eqP=>// ->.
Qed. 

Lemma offset_total (x y : loc) : exists m, (x == y .+ m) || (y == x .+ m).
Proof.
move=>[x][y]; case/orP: (leq_total x y)=>H; 
[exists (y - x) | exists (x - y)]; apply/orP; [right | left];
by rewrite /loc_offset /= subnK.
Qed.

(****************)
(* Sorted_C heaps *)
(****************)

Record heap_s : Type := Heap {heapc :> heap_c; _ : sorted_c heapc}.

Program Definition empty_s := @Heap empty_c _.

Lemma update_pf (h : heap_s) (x : loc) (d : dynamic) :
    x != null -> sorted_c (update_c h (loc_to_nat x) d). 
Proof.
elim=>hc hs x d pfx; apply: update_compat=>//.
by apply: contra pfx; case: x=>n /=; move/eqP=>->.
Qed.   

Definition update_s (h : heap_s) (x : loc) (d : dynamic) := 
  if (dec (x != null)) is left pf then 
      @Heap (update_c h (loc_to_nat x) d) (@update_pf h x d pf) 
  else empty_s. 

Definition valid_s (x : loc) (h : heap_s) := valid_c (loc_to_nat x) h.

Program Definition free_s (x : loc) (h : heap_s) := @Heap (free_c (loc_to_nat x) h) _.
Next Obligation. by elim: h x=> * //=; apply: free_compat. Qed.

Definition lookup_s (x : loc) (h : heap_s) := lookup_c (loc_to_nat x) h.

Program Definition union2_s (h1 h2 : heap_s) := @Heap (union2_c h1 h2) _.
Next Obligation. by elim: h1 h2=>[h1 p1][h2 p2]//=; apply union2_compat. Qed.

Definition disjoint2_s (h1 h2 : heap_s) := disjoint2_c h1 h2.

Definition is_empty_s (h : heap_s) := is_empty_c h.

Program
Definition fresh'_s (h : heap_s) : loc := loc_from_nat ((last_loc h).+1).

Definition fresh_s := nosimpl fresh'_s.

Definition subdom_s (h1 h2 : heap_s) := subdom_c h1 h2. 
  
Definition pick_s (h : heap_s) : loc := nosimpl 
             loc_from_nat (pick_c h).

(***********************)
(* Properties of heaps *)
(***********************)

Lemma heap_func_s (h1 h2 : heap_s) : h1 = h2 :> heap_c -> h1 = h2.
Proof.
move=>[x1 p1][x2 p2] /= H; move:H p1 p2=>->{x1} *.
by congr Heap; apply: bool_irrelevance.
Qed. 

Lemma is_emptyP_s (h : heap_s) : reflect (h = empty_s) (is_empty_s h).
Proof.
rewrite /empty_s; elim=>[[|h]H]; constructor=> //=.
by congr Heap; apply:bool_irrelevance.
Qed.

(*******************************************)
(* Ring properties of union2 and disjoint2 *)
(*******************************************)

Lemma union2_commute_s (h1 h2 : heap_s) : 
        disjoint2_s h1 h2 -> union2_s h1 h2 = union2_s h2 h1.
Proof. by move=>[h1 p1][h2 p2] //= H; apply: heap_func_s=>/=; apply union2_commute_c. Qed.

Lemma union2_assoc_s (h1 h2 h3 : heap_s) : 
        union2_s (union2_s h1 h2) h3 = union2_s h1 (union2_s h2 h3).
Proof. by move=>[h1 p1][h2 p2][h3 p3] //=; apply: heap_func_s; apply: union2_assoc_c. Qed.

Lemma union2_empty_left_s (h : heap_s) : union2_s empty_s h = h.
Proof. by move=>[h p] //=; apply: heap_func_s; apply: union2_empty_left_c. Qed.

Lemma union2_empty_right_s (h : heap_s) : union2_s h empty_s = h.
Proof. by move=>[h p] //=; apply: heap_func_s. Qed.

Lemma disjoint2_empty_left_s (h : heap_s) : disjoint2_s empty_s h.
Proof. by move=>[h p]; apply: disjoint2_empty_left_c. Qed.

Lemma disjoint2_empty_right_s (h : heap_s) : disjoint2_s h empty_s.
Proof. trivial. Qed.

Lemma disjoint2_sym_s (h1 h2 : heap_s) : disjoint2_s h1 h2 = disjoint2_s h2 h1.
Proof. by move=>[h1 p1][h2 p2] //=; apply: disjoint2_sym_c. Qed.

Lemma disjoint2_union2_left_s (h1 h2 h : heap_s) : 
  disjoint2_s (union2_s h1 h2) h = disjoint2_s h1 h && disjoint2_s h2 h.
Proof. 
by move=>[h1 p1][h2 p2][h3 p3]; rewrite /disjoint2_s =>//=; apply disjoint2_union2_c. 
Qed.

Lemma disjoint2_union2_right_s (h h1 h2 : heap_s) :
        disjoint2_s h (union2_s h1 h2) = disjoint2_s h h1 && disjoint2_s h h2.
Proof. by move=>h h1 h2; rewrite !(disjoint2_sym_s h); apply: disjoint2_union2_left_s. Qed.

(************************)
(* Cancelation of union *)
(************************)

Lemma union2_cancel_left_s (h h1 h2 : heap_s) : 
        union2_s h h1 = union2_s h h2 -> disjoint2_s h h1 -> disjoint2_s h h2 -> h1 = h2.
Proof.
move=>[x p][x1 p1][x2 p2]; rewrite /union2_s /disjoint2_s //=; move=> [H1] H2 H3.
apply: heap_func_s=> //=.
by apply: union2_cancel_left_c H3.
Qed.

Lemma union2_cancel_right_s (h1 h2 h : heap_s) : 
        union2_s h1 h = union2_s h2 h -> disjoint2_s h1 h -> disjoint2_s h2 h -> h1 = h2.
Proof.
move=>[x p][x1 p1][x2 p2]; rewrite /union2_s /disjoint2_s //=; move=> [H1] H2 H3.
apply: heap_func_s=> //=.
by apply: union2_cancel_right_c H3.
Qed.

(*******************************************************)
(* McCarthy-style properties                           *)
(*                                                     *)
(* This list pairs up every primitive with every other *)
(* I have tried to make sure I didn't omit any pair    *)
(*******************************************************)

(************************)
(* properties of update *)
(************************)

Lemma valid_nonull_s (h : heap_s) (x : loc) : valid_s x h -> x != null.
Proof.
by move=>[h p] x; case: eqP=>// ->; rewrite /valid_s /= (negbET (valid_null_c p)).
Qed.

Lemma update_eta_s (h : heap_s) (x : loc) : 
        valid_s x h -> h = update_s h x (lookup_s x h).
Proof. 
move=>[h p] x H; apply: heap_func_s=>/=; rewrite /update_s /=.
case: dec=>[i|]; last by rewrite (valid_nonull_s H).
by apply: update_eta_c.
Qed.

Lemma update_update_eq_s (h : heap_s) (x : loc) (d1 d2 : dynamic) : 
        update_s (update_s h x d1) x d2 = update_s h x d2.
Proof.
move=>[h ph] x *; apply: heap_func_s; rewrite /update_s; case: dec=>//= i.
by apply: update_update_eq_c.
Qed.
 
Lemma update_update_neq_s (h : heap_s) (x y : loc) (d1 d2 : dynamic) : 
        x != y -> x != null -> y != null -> 
        update_s (update_s h x d1) y d2 = update_s (update_s h y d2) x d1.
Proof.
move=>[h ph] x y d1 d2 H H1 H2; apply: heap_func_s=>//=.
rewrite /update_s; do 2![case: dec]=>//= _ _ {H1 H2}.
by apply: update_update_neq_c=>//; apply: contra H; move/eqP; move/loc_inj=>->.
Qed.   

Lemma valid_update_eq_s (h : heap_s) (x : loc) (d : dynamic) : 
        x != null -> valid_s x (update_s h x d).
Proof. 
move=>[h ph] [x] d; rewrite /valid_s /update_s /=. 
by case: dec=>// i _; apply: valid_update_eq_c.
Qed.

Lemma valid_update_neq_s (h : heap_s) (x : loc) (y : loc) (d : dynamic) : 
        x != y -> y != null -> valid_s x h = valid_s x (update_s h y d).
Proof. 
move=>[h ph][x][y] d H H1; rewrite /update_s; case: dec=>// H2.
by apply: valid_update_neq_c.
Qed.

Lemma free_update_eq_s (h : heap_s) (x : loc) (d : dynamic) : 
        x != null -> free_s x (update_s h x d) = free_s x h.
Proof. 
move=>[h ph][x] d H; apply: heap_func_s=>//=; rewrite /update_s; case: dec=>// H1. 
by apply: free_update_eq_c.
Qed.

Lemma free_update_neq_s (h : heap_s) (x : loc) (y : loc) (d : dynamic) : 
        x != y -> y != null -> free_s x (update_s h y d) = update_s (free_s x h) y d.
Proof. 
move=>[h ph][x][y] d H1 H2; apply: heap_func_s=>//=.
by rewrite /update_s; case: dec=>// H3; apply: free_update_neq_c.
Qed.

Lemma update_free_neq_s (h : heap_s) (x : loc) (y : loc) (d : dynamic) : 
        x != y -> y != null -> update_s (free_s x h) y d = free_s x (update_s h y d).
Proof. by move=>*; apply: sym_equal; apply: free_update_neq_s. Qed.

Lemma update_free_eq_s (h : heap_s) (x : loc) (d : dynamic) : 
        update_s (free_s x h) x d = update_s h x d.
Proof.
move=>[h p] x d; apply: heap_func_s=>/=; rewrite /update_s; case: dec=>// H.
by apply: update_free_eq_c.
Qed.

Lemma lookup_update_eq_s (h : heap_s) (x : loc) (d : dynamic) : 
        x != null -> lookup_s x (update_s h x d) = d.
Proof.
move=>[h ph] x d; rewrite /update_s; case: dec=>// H _.
by apply: lookup_update_eq_c.
Qed.

Lemma lookup_update_neq_s (h : heap_s) (x : loc) (y : loc) (d : dynamic) : 
        x != y -> y != null -> lookup_s x (update_s h y d) = lookup_s x h.
Proof. 
move=>[h ph][x][y] d H1 H2; rewrite /update_s; case: dec=>// H3.
by apply: lookup_update_neq_c.
Qed.

Lemma update_union2_left_s (h1 h2 : heap_s) (x : loc) (d : dynamic) : 
        ~~ valid_s x h2 -> x != null -> 
        update_s (union2_s h1 h2) x d = union2_s (update_s h1 x d) h2.
Proof.
move=>[h1 p1][h2 p2][x] d H1 H2; apply: heap_func_s=>//=; apply: sym_equal.
by rewrite /update_s; case: dec=>// H3; apply: union2_update_left_c.
Qed.

Lemma union2_update_left_s (h1 h2 : heap_s) (x : loc) (d : dynamic) : 
         ~~ valid_s x h2 -> x != null -> 
         union2_s (update_s h1 x d) h2 = update_s (union2_s h1 h2) x d.
Proof. by move=>*; apply: sym_equal; apply: update_union2_left_s. Qed.

Lemma update_union2_right_s (h1 h2 : heap_s) (x : loc) (d : dynamic) : 
        x != null -> update_s (union2_s h1 h2) x d = union2_s h1 (update_s h2 x d). 
Proof.
move=>[h1 p1][h2 p2][x] d H; apply: heap_func_s=>//; apply: sym_equal.
by rewrite /update_s; case: dec=>// H'; apply: union2_update_right_c.
Qed.

Lemma union2_update_right_s (h1 h2 : heap_s) (x : loc) (d : dynamic) : 
        x != null -> union2_s h1 (update_s h2 x d) = update_s (union2_s h1 h2) x d.
Proof. by move=>*; apply: sym_equal; apply: update_union2_right_s. Qed.

Lemma disjoint2_update_left_s (h1 h2 : heap_s) (x : loc) (d : dynamic) : 
        ~~ valid_s x h2 -> x != null -> 
        disjoint2_s (update_s h1 x d) h2 = disjoint2_s h1 h2.
Proof.
move=>[h1 p1][h2 p2][x] d //= H1 H2.
rewrite /disjoint2_s !(disjoint2_sym_c _ h2) /update_s.
by case: dec=>//= H3; rewrite disjoint2_update_right_c.
Qed.

Lemma disjoint2_update_right_s (h1 h2 : heap_s) (x : loc) (d : dynamic) : 
        ~~ valid_s x h1 -> x != null -> 
        disjoint2_s h1 (update_s h2 x d) = disjoint2_s h1 h2.
Proof. by move=>h1 *; rewrite !(disjoint2_sym_s h1); apply: disjoint2_update_left_s. Qed.

Lemma is_empty_update_s (h : heap_s) (x : loc) (d : dynamic) : 
        x != null -> ~~ is_empty_s (update_s h x d).
Proof. 
move=>[h p][x] d H; rewrite /update_s; case: dec=>//= H1.  
by apply: is_empty_update_c. 
Qed. 

(***************************)
(* properties of valid_loc *)
(***************************)

Lemma valid_empty_s (x : loc) : ~~ valid_s x empty_s.
Proof. trivial. Qed.

Lemma valid_null_s (h : heap_s) : ~~ valid_s null h.
Proof. by move=>[h p]; apply: valid_null_c. Qed.

Lemma valid_free_eq_s (h : heap_s) (x : loc) : ~~ valid_s x (free_s x h).
Proof. by move=>[h p][x]; apply: valid_free_eq_c. Qed.

Lemma valid_free_neq_s (h : heap_s) (x y : loc) : 
        x != y -> valid_s x (free_s y h) = valid_s x h.
Proof. move=>[h p][x][y]; apply: valid_free_neq_c. Qed.

Lemma valid_lookup_s (h : heap_s) (x : loc) : ~~ valid_s x h -> lookup_s x h = dyn tt.
Proof. by move=>[h p][x][y]; apply: valid_lookup_c. Qed.

(*
Lemma valid_lookup_P_s (h : heap_s) (x : locn) : reflect (exists d, lookup_s x h = Some d) (valid_s x h).
Proof. move=>[h p][x]; apply: valid_lookup_P_c. Qed.

Lemma valid_lookup_PN_s (h : heap_s) (x : locn) : reflect (lookup_s x h = None) (~~ valid_s x h).
Proof.
move=>h x; case: valid_lookup_P_s=>H; constructor; move:H; first by move=>[d ->].
case: lookup_s=>d//; elim.
by exists d.
Qed.
*)

Lemma valid_union2_s (h1 h2 : heap_s) (x : loc) : 
        valid_s x (union2_s h1 h2) = valid_s x h1 || valid_s x h2.
Proof. move=>[h1 p1][h2 p2][x] //=; apply: valid_union2_c. Qed.

Lemma valid_disjoint2_s (h1 h2 : heap_s) (x : loc) : 
        disjoint2_s h1 h2 -> valid_s x h1 -> ~~ valid_s x h2.
Proof. by move=>[h1 p1][h2 p2][x] //= H1 H2; apply: (valid_disjoint2_c H1). Qed.

Lemma valid_disjoint2_P1_s (h1 h2 : heap_s) : 
        reflect (forall x, valid_s x h1 -> ~~ valid_s x h2) (disjoint2_s h1 h2).
Proof.
move=>[h1 p1][h2 p2]; rewrite /disjoint2_s /valid_s //=.
case: valid_disjoint2_P1_c=> H; constructor; first by move=>[x]; exact: H.
by move=> H1; apply: H => x H; apply: (H1 (loc_from_nat x)).
Qed.

Lemma valid_disjoint2_P2_s (h1 h2 : heap_s) : reflect (forall x, valid_s x h2 -> ~~ valid_s x h1) (disjoint2_s h1 h2).
Proof. move=>*; rewrite disjoint2_sym_s; apply: valid_disjoint2_P1_s. Qed.

Lemma valid_disjoint2_PN_s (h1 h2 : heap_s) : reflect (exists x, valid_s x h1 /\ valid_s x h2) (~~ disjoint2_s h1 h2).
Proof.
move=>[h1 p1][h2 p2]; rewrite /disjoint2_s /valid_s /=.
case: valid_disjoint2_PN_c=>H; [left | right].
- by move: H=>[x [H1 H2]]; exists (loc_from_nat x).
by move=>[x [H1 H2]]; elim: H; exists (loc_to_nat x).
Qed.

(**************************)
(* properties of free_loc *)
(**************************)

Lemma free_valid_P_s (h : heap_s) (x : loc) : reflect (free_s x h <> h) (valid_s x h). 
Proof.
move=>[h p][x] //=; rewrite /free_s /valid_s //=; apply: (iffP idP); move=>H1. 
- by move=>[H2]; move: H1 (free_valid_eq_c H2)=>->.
apply: negbE2; apply: (introN idP) => H2.
apply: H1; apply: heap_func_s=>/=.
by apply: free_valid_neq_c.
Qed.

Lemma free_valid_PN_s (h : heap_s) (x : loc) : reflect (free_s x h = h) (~~ valid_s x h). 
Proof.
move=>[h p][x] //=; rewrite /free_s /valid_s //=.
apply: (iffP idP); move=>[H1]; last by apply: free_valid_eq_c. 
apply: heap_func_s=> //=.
by apply free_valid_neq_c.
Qed.

Lemma free_empty_s (x : loc) : free_s x empty_s = empty_s.
Proof. by move=>*; apply: heap_func_s. Qed.
 
Lemma free_free_eq_s (h : heap_s) (x : loc) :  free_s x (free_s x h) = free_s x h.
Proof. by move=>[h p][x] //=; apply: heap_func_s=>//=; apply: free_free_eq_c. Qed.

Lemma free_free_neq_s (h : heap_s) (x y : loc) :  x != y -> free_s x (free_s y h) = free_s y (free_s x h).
Proof. by move=>[h p][x][y] H; apply: heap_func_s=>//=; apply: free_free_neq_c. Qed.

Lemma lookup_free_eq_s (h : heap_s) (x : loc) : lookup_s x (free_s x h) = dyn tt. 
Proof. by move=>[h p][x] //=; apply: lookup_free_eq_c. Qed.

Lemma lookup_free_neq_s (h : heap_s) (x y : loc) : x != y -> lookup_s x (free_s y h) = lookup_s x h.
Proof. by move=>[h p][x][y]; apply: lookup_free_neq_c. Qed.

Lemma free_valid_neq_s (h : heap_s) (x : loc) : ~~ valid_s x h -> free_s x h = h.
Proof. by move=>[h p][x] nv; apply: heap_func_s; apply: free_valid_neq_c. Qed.

Lemma free_valid_eq_s (h : heap_s) (x : loc) : free_s x h = h -> ~~ valid_s x h.
Proof. by move=>[h p][x]eq;apply: free_valid_eq_c; inversion eq. Qed.

Lemma free_union2_s (h1 h2 : heap_s) (x : loc) : free_s x (union2_s h1 h2) = union2_s (free_s x h1) (free_s x h2).
Proof. by move=>[h1 p1][h2 p2] *; apply: heap_func_s=>//=; apply: free_union2_c. Qed.

Lemma disjoint2_free_left_eq_s (h1 h2 : heap_s) (x : loc) : ~~ valid_s x h2 -> 
  disjoint2_s (free_s x h1) h2 = disjoint2_s h1 h2.
Proof. by move=>[h1 p1][h2 p2][x] //= H1; rewrite /disjoint2_s /free_s disjoint2_free_left_eq_c. Qed.

Lemma disjoint2_free_right_eq_s (h1 h2 : heap_s) (x : loc) : ~~ valid_s x h1 -> 
  disjoint2_s h1 (free_s x h2) = disjoint2_s h1 h2.
Proof. by move=>h1 *; rewrite !(disjoint2_sym_s h1) disjoint2_free_left_eq_s. Qed.

Lemma disjoint2_free_left_s (h1 h2 : heap_s) (x : loc) : 
  disjoint2_s h1 h2 -> disjoint2_s (free_s x h1) h2.
Proof. by rewrite/disjoint2_s /free_s; move=> [h1 p1][h2 p2][x] //= dis; apply disjoint2_free_left_c. Qed.

Lemma disjoint2_free_right_s (h1 h2 : heap_s) (x : loc) : 
  disjoint2_s h1 h2 -> disjoint2_s h1 (free_s x h2).
Proof. by move => h1 h2 *; rewrite (disjoint2_sym_s h1) disjoint2_free_left_s // (disjoint2_sym_s h2). Qed.

Lemma free_union2_left_s (h1 h2 : heap_s) (x : loc) : ~~valid_s x h2 ->
  free_s x (union2_s h1 h2) = union2_s (free_s x h1) h2.
Proof. by move => h1 h2 x nv; rewrite free_union2_s (free_valid_neq_s nv). Qed.

Lemma free_union2_right_s (h1 h2 : heap_s) (x : loc) : ~~valid_s x h1 ->
  free_s x (union2_s h1 h2) = union2_s h1 (free_s x h2).
Proof. by move => h1 h2 x nv; rewrite free_union2_s (free_valid_neq_s nv). Qed.

(**************************)
(* properties of lookup_s *)
(**************************)

Lemma lookup_empty_s (x : loc) : lookup_s x empty_s = dyn tt. 
Proof. trivial. Qed.

(*
Lemma lookup_union2_none_intro_s (h1 h2 : heap_s) (x : locn) : 
        lookup_s x h1 = None -> lookup_s x h2 = None -> lookup_s x (union2_s h1 h2) = None.
Proof. by move=>[h1 p1][h2 p2] *; apply: lookup_union2_none_intro_c. Qed.

Lemma lookup_union2_none_elim_s (h1 h2 : heap_s) (x : locn) : 
        lookup_s x (union2_s h1 h2) = None -> lookup_s x h1 = None /\ lookup_s x h2 = None.
Proof. by move=>[h1 p1][h2 p2] *; apply: lookup_union2_none_elim_c. Qed.
*)

Lemma lookup_union2_some_intro1_s (h1 h2 : heap_s) (x : loc) : 
        ~~ valid_s x h2 -> lookup_s x (union2_s h1 h2) = lookup_s x h1. 
Proof. by move=>[h1 p1][h2 p2] *; apply: lookup_union2_some_intro1_c. Qed.

Lemma lookup_union2_some_intro2_s (h1 h2 : heap_s) (x : loc) : 
        valid_s x h2 -> lookup_s x (union2_s h1 h2) = lookup_s x h2.
Proof. by move=>[h1 p1][h2 p2] *; apply: lookup_union2_some_intro2_c. Qed.

Lemma lookup_union2_some_elim_s (h1 h2 : heap_s) (x : loc) (d : dynamic) : 
        valid_s x (union2_s h1 h2) -> lookup_s x (union2_s h1 h2) = d -> 
          (lookup_s x h1 = d /\ ~~ valid_s x h2) \/ lookup_s x h2 = d.
Proof. by move=>[h1 p1][h2 p2] *; apply: lookup_union2_some_elim_c. Qed.

(*
Lemma lookup_disjoint2_s (h1 h2 : heap_s) (x : locn) (d : dynamic) : 
        disjoint2_s h1 h2 -> lookup_s x h1 = Some d -> lookup_s x h2 = None.
Proof. move=>[h1 p1][h2 p2][x] d //= H1 H2; apply: (lookup_disjoint2_c H1 H2). Qed.

Lemma lookup_disjoint2_P1_s (h1 h2 : heap_s) : 
        reflect (forall x d, lookup_s x h1 = Some d -> lookup_s x h2 = None) (disjoint2_s h1 h2).
Proof.
move=>h1 h2; case: valid_disjoint2_P1_s=>H; constructor. 
- move=> x d H1; move: {H} (H x); case: valid_lookup_PN_s=>//.
  case: valid_lookup_P_s=>[_ _ H3|]; last by elim; exists d.
  by move: (H3 (refl_equal true)).
move=>H1; apply: H=> x.
case: valid_lookup_P_s=>//=; move=> [d H] _.
case: valid_lookup_PN_s=>//=; elim.
by apply: (H1 _ d).
Qed.

Lemma lookup_disjoint2_P2_s (h1 h2 : heap_s) :
        reflect (forall x d, lookup_s x h2 = Some d -> lookup_s x h1 = None) (disjoint2_s h1 h2).
Proof. move=>*; rewrite disjoint2_sym_s; apply: lookup_disjoint2_P1_s. Qed.
*)

(******************************************)
(* properties of union2_s and disjoint2_s *)
(******************************************)

(* the ring properties have already been proved above, *)
(* so there is only one more lemma left to do: *)

Lemma is_empty_union2_s (h1 h2 : heap_s) : is_empty_s (union2_s h1 h2) = is_empty_s h1 && is_empty_s h2.
Proof. by move=>*; apply: is_empty_union2_c. Qed.

(*********************)
(* Reflecting subdom *)
(*********************)

Lemma subdomP_s (h1 h2 : heap_s) : 
        reflect (forall x:loc, valid_s x h1 -> valid_s x h2) (subdom_s h1 h2).
Proof.
move=>[h1 p1][h2 p2]; rewrite /subdom_s /valid_s /=. 
case: subdomP_c=>// H; constructor.
- by case=>//=; case=>x /= _; apply: H.
move=>H1; elim: H; case=>[|n] H2.
- by move: (valid_null_c p1); rewrite H2. 
have Q : forall (n1 n2:nat), n1 != n2 -> loc_from_nat n1 != loc_from_nat n2.
- by move=>n1 n2; apply: contra; move/eqP=>[->]; rewrite eq_refl. 
by apply: (H1 (loc_from_nat n.+1)).
Qed.

Lemma subdomE_s (h1 h2 h : heap_s) : 
        disjoint2_s h1 h -> disjoint2_s h2 h -> 
        subdom_s h1 h2 -> subdom_s (union2_s h1 h) (union2_s h2 h).
Proof.
rewrite /disjoint2_s /subdom_s.
move=>[h1 p1][h2 p2][h p] /= H1 H2 H.
by apply: subdomE_c.
Qed.

Lemma subdom_disjoint2_s (h1 h2 h : heap_s) : 
        subdom_s h1 h2 -> disjoint2_s h2 h -> disjoint2_s h1 h.
Proof.
rewrite /subdom_s /disjoint2_s; move=>[h1 p1][h2 p2][h p].
by apply: subdom_disjoint2_c.
Qed. 

Lemma subdom_union2_s (h1 h2 h1' h2' : heap_s) : 
        subdom_s h1 h2 -> subdom_s h1' h2' -> 
          disjoint2_s h2 h2' -> subdom_s (union2_s h1 h1') (union2_s h2 h2').
Proof. by move=>[h1 p1][h2 p2][h1' p1'][h2' p2']; apply: subdom_union2_c. Qed.

Lemma subdom_emp_s (h : heap_s) : subdom_s empty_s h.
Proof. by case=>h p; apply: subdom_emp_c. Qed.

Lemma subdom_emp_inv_s (h : heap_s) : subdom_s h empty_s -> h = empty_s.
Proof.
case=>[h p]; rewrite /subdom_s /empty_s=>H1.
by apply: heap_func_s; apply: subdom_emp_inv_c.
Qed. 
 
(********)
(* Pick *)
(********)

Lemma emp_pick_s (h : heap_s) : (pick_s h == null) = (is_empty_s h).
Proof.
case=>h p; rewrite /pick_s /is_empty_s /=.
case: h p=>// n d h /=.
by case: n.
Qed.

Lemma pickP_s (h : heap_s) : ~~ is_empty_s h = valid_s (pick_s h) h.
Proof.
rewrite /valid_s /pick_s /is_empty_s=>h.
apply: pickP_c.
Qed.

(***************************************************************)
(* The Exported Interface.                                     *)
(*                                                             *)
(* A heap is an option heap_s. None represents an invalid heap *)
(***************************************************************)

(* eventually fold the proofs of sortedness into the definition *)
(* and get rid of heap_s *)

(* Undef is the top of the heap ordering; it's the result of unioning *)
(* non-disjoint heaps *)
Inductive heap : Type := Undef | Def of heap_s.

Definition def (h : heap) := nosimpl 
  if h is Def _ then true else false.

Definition empty : heap := nosimpl Def empty_s.

Definition upd (h : heap) (x : loc) (d : dynamic) : heap := nosimpl
  if h is Def hs then 
    if x != null then Def (update_s hs x d) else Undef 
  else Undef. 

Definition valid (x : loc) (h : heap) : bool := nosimpl
  if h is Def hs then valid_s x hs else false.

Definition free (x : loc) (h : heap) : heap := nosimpl
  if h is Def hs then Def (free_s x hs) else Undef. 

Definition look (x : loc) (h : heap) : dynamic := nosimpl
  if h is Def hs then lookup_s x hs else dyn tt. 

Definition union2 (h1 h2 : heap) : heap := nosimpl
  if (h1, h2) is (Def hs1, Def hs2) then  
     if (disjoint2_s hs1 hs2) then Def (union2_s hs1 hs2) else Undef
  else Undef.

Definition is_empty (h : heap) : bool := nosimpl
  if h is Def hs then is_empty_s hs else false.

Definition fresh (h : heap) : loc := nosimpl
  if h is Def hs then fresh_s hs else fresh_s empty_s.

Definition subdom (h1 h2 : heap) : bool := nosimpl 
  if (h1, h2) is (Def hs1, Def hs2) then subdom_s hs1 hs2
  else false.

Definition pick (h : heap) : loc := nosimpl
  if h is Def hs then pick_s hs else null.

Notation "h1 :+ h2" := (union2 h1 h2) (at level 43, left associativity).
Notation "x :-> v" := (upd empty x (dyn v)) (at level 30).

Lemma emptyD : def empty.
Proof. by []. Qed.

Lemma updD h x d : def (upd h x d) = (x != null) && (def h).
Proof. 
rewrite /upd=>[[|h]] x d; first by rewrite andbF.
by case: ifP=>->.
Qed.

Lemma validD (x : loc) (h : heap) : valid x h -> def h. 
Proof. by move=>x; case. Qed. 

Lemma freeD (x : loc) (h : heap) : def (free x h) = def h.
Proof. by move=>x; case. Qed. 

Lemma unDl (h1 h2 : heap) : def (h1 :+ h2) -> def h1. 
Proof. by case. Qed.

Lemma unDr (h1 h2 : heap) : def (h1 :+ h2) -> def h2. 
Proof. by case=>h1 //; case. Qed.

Lemma defPt (A : Type) (x : loc) (v : A) : def (x :-> v) = (x != null).
Proof. by move=>A x v; rewrite /def /upd /=; case: ifP=>->. Qed. 

(***********************)
(* Properties of heaps *)
(***********************)

Lemma is_emptyP (h : heap) : reflect (h = empty) (is_empty h).
Proof.
case=>[|h] /=; first by right.
rewrite /is_empty.
case: is_emptyP_s=>[->|H]; constructor=>//.
by case.
Qed. 

(****************************)
(* Ring properties of union *)
(****************************)

Lemma unC (h1 h2 : heap) : h1 :+ h2 = h2 :+ h1.
Proof.
case=>[|h1]; case=>[|h2] //. 
case H: (disjoint2_s h1 h2); rewrite /union2 H disjoint2_sym_s H //.
by rewrite union2_commute_s.
Qed.

Lemma unCA (h1 h2 h3 : heap) : h1 :+ (h2 :+ h3) = h2 :+ (h1 :+ h3).
Proof.
case=>[|h1]; case=>[|h2]; case=>[|h3] //; rewrite /union2. 
case: ifP; case: ifP=>//; rewrite disjoint2_union2_right_s andb_comm.
- move/andP=>[-> H1] H2.
  rewrite disjoint2_union2_right_s disjoint2_sym_s H1 H2 /=. 
  rewrite (union2_commute_s H2) -union2_assoc_s union2_commute_s //.
  by rewrite disjoint2_union2_left_s H1 disjoint2_sym_s.
- case: ifP=>//.
  move=>->/= H1 _.
  by rewrite disjoint2_union2_right_s disjoint2_sym_s H1 /=.
by move=>_ ->.
Qed.

Lemma unAC (h1 h2 h3 : heap) : h1 :+ h2 :+ h3 = h1 :+ h3 :+ h2.
Proof. by move=>h ??; rewrite (unC h) unC unCA unC (unC h). Qed.

Lemma unA (h1 h2 h3 : heap) : h1 :+ (h2 :+ h3) = h1 :+ h2 :+ h3.
Proof. by move=>? h ?; rewrite (unC h) unCA unC. Qed.
 
Lemma un0h (h : heap) : empty :+ h = h.
Proof. by case=>[|h] //=; rewrite /union2 /= disjoint2_empty_left_s union2_empty_left_s. Qed.

Lemma unh0 (h : heap) : h :+ empty = h.
Proof. by move=>h; rewrite unC un0h. Qed.

(*
(* we add a bunch of lemmas for automatic resolving of unions with empty *)

Lemma ung0h (h g : heap) : g = h -> g = empty :+ h.
Proof. by move=>h g ->; rewrite un0h. Qed.
Lemma ungh0 (h g : heap) : g = h -> g = h :+ empty.
Proof. by move=>h g ->; rewrite unh0. Qed.
Lemma un0hg (h g : heap) : g = h -> empty :+ h = g.
Proof. by move=>h g ->; rewrite un0h. Qed.
Lemma unh0g (h g : heap) : g = h -> h :+ empty = g.
Proof. by move=>h g ->; rewrite unh0. Qed.

Definition unh0_sym (h : heap) := ssrfun.esym (unh0 h).
Definition un0h_sym (h : heap) := ssrfun.esym (un0h h).

Hint Resolve unh0_sym un0h_sym ung0h ungh0 un0hg unh0g. *)

(************************)
(* Cancelation of union *)
(************************)
  
Lemma unKhl (h h1 h2 : heap) : def (h1 :+ h) -> h1 :+ h = h2 :+ h -> h1 = h2.
Proof.
case=>[|h]; case=>[|h1]; case=>[|h2] //=; rewrite /union2; case: ifP=>//.
case: ifP=>//= H1 H2 _ [H3].
have {H3} H3 : union2_s h1 h = union2_s h2 h by apply: heap_func_s.
by rewrite (union2_cancel_right_s H3).
Qed.

Lemma unhKl (h h1 h2 : heap) : def (h :+ h1) -> h :+ h1 = h :+ h2 -> h1 = h2.
Proof. by move=>h h1 h2; rewrite !(unC h); apply: unKhl. Qed.

Lemma unKhr (h h1 h2 : heap) : def (h2 :+ h) -> h1 :+ h = h2 :+ h -> h1 = h2.
Proof. by move=>h h1 h2 H1 H2; symmetry in H2; rewrite (unKhl H1 H2). Qed.

Lemma unhKr (h h1 h2 : heap) : def (h :+ h2) -> h :+ h1 = h :+ h2 -> h1 = h2.
Proof. by move=>h h1 h2; rewrite !(unC h); apply: unKhr. Qed.

(********)
(* Pick *)
(********)

Lemma emp_pick (h : heap) : (pick h == null) = (~~ def h || is_empty h).
Proof.
rewrite /pick /is_empty; case=>//. 
apply: emp_pick_s.
Qed.

Lemma pickP (h : heap) : def h && ~~ is_empty h = valid (pick h) h.
Proof.
rewrite /def /is_empty /valid /pick. case=>//= h.
apply: pickP_s.
Qed.   

(*********************)
(* Reflecting subdom *)
(*********************)

Lemma subdom_def (h1 h2 : heap) : subdom h1 h2 -> def h1 && def h2. 
Proof.  by case=>[|h1]; case=>[|h2]. Qed.

Lemma subdomP (h1 h2 : heap) : 
        def h1 -> ~~ is_empty h1 -> 
          reflect (forall x:loc, valid x h1 -> valid x h2) 
                  (subdom h1 h2). 
Proof.
rewrite /subdom /valid /is_empty.
case=>//= h1; case=>//= [|h2] _ H;
have T : pick_s h1 != null by [rewrite emp_pick_s];
rewrite pickP_s in H.
- by constructor; move/(_ _ H).
by case: subdomP_s=>H1; constructor.
Qed.

Lemma subdomD (h1 h2 h : heap) : 
        subdom h1 h2 -> def (h2 :+ h) -> def (h1 :+ h).
Proof.
case=>[|h1]; case=>[|h2]; case=>[|h] //=.
rewrite /subdom /def /union2=>H.
case E1: (disjoint2_s h1 h); case E2: (disjoint2_s h2 h)=>// _.
by move: (subdom_disjoint2_s H E2); rewrite E1.
Qed.

Lemma subdomE (h1 h2 h : heap) : 
        def (h2 :+ h) -> subdom h1 h2 -> subdom (h1 :+ h) (h2 :+ h).
Proof.
case=>[|h1]; case=>[|h2]; case=>[|h] //=.
rewrite /union2 /subdom /def /=.
case E1: (disjoint2_s h1 h); case E2: (disjoint2_s h2 h)=>// _ H; last first.
- by move: (subdom_disjoint2_s H E2); rewrite E1.
by apply: subdomE_s.
Qed.

Lemma subdomUE (h1 h2 h1' h2' : heap) : 
        def (h2 :+ h2') -> 
          subdom h1 h2 -> subdom h1' h2' ->
            subdom (h1 :+ h1') (h2 :+ h2').
Proof.
case=>[|h1]; case=>[|h2]; case=>[|h1']; case=>[|h2'] //.
rewrite /subdom /def /union2=>H H1 H2; move: H. 
case E2: (disjoint2_s h2 h2'); case E1: (disjoint2_s h1 h1')=>// _.
- by apply: subdom_union2_s.
move: (subdom_disjoint2_s H1 E2).
rewrite disjoint2_sym_s=>E3.
move: (subdom_disjoint2_s H2 E3).
by rewrite disjoint2_sym_s E1.
Qed.

Lemma subdom_emp (h : heap) : def h -> subdom empty h.
Proof.
case=>[|h] // _.
rewrite /subdom /empty.
apply: subdom_emp_s.
Qed.

Lemma subdom_emp_inv (h : heap) : subdom h empty -> h = empty.
Proof.
case=>[|h] //.
rewrite /subdom /empty.
by move/subdom_emp_inv_s=>->.
Qed.      

Lemma subdomPE (A B : Type) (x : loc) (v1 : A) (v2 : B) : 
        x != null -> subdom (x :-> v1) (x :-> v2).
Proof. 
move=>A B x v1 v2 H; rewrite /subdom /upd /= H /subdom_s /update_s.
by case: dec=>//= H2; rewrite ltnn eq_refl.
Qed.

Hint Resolve subdom_emp subdomPE. 

(*******************************************************)
(* McCarthy-style properties                           *)
(*                                                     *)
(* This list pairs up every primitive with every other *)
(* I have tried to make sure I didn't omit any pair    *)
(*******************************************************)

(************************)
(* properties of update *)
(************************)

Lemma upd_inj (h : heap) (x : loc) (d1 d2 : dynamic) :
        def h -> x != null -> upd h x d1 = upd h x d2 -> d1 = d2.
Proof. 
case=>[|h] x d1 d2 // _ H; rewrite /upd H /update_s.
by case: dec=>// H1 []; apply: update_inj_c.
Qed.

Lemma upd_eta (h : heap) (x : loc) : 
        valid x h -> h = upd h x (look x h).
Proof. 
rewrite /valid /upd; case=>// h x H.
by rewrite -update_eta_s // (valid_nonull_s H).
Qed.

Lemma updUE (h : heap) (x : loc) (d1 d2 : dynamic) : 
        upd (upd h x d1) x d2 = upd h x d2.
Proof. 
rewrite /upd=>[[|h]] x d1 d2 //. 
by case: ifP=>-> //; rewrite update_update_eq_s.
Qed.
 
Lemma updUN (h : heap) (x y : loc) (d1 d2 : dynamic) : 
        x != y -> upd (upd h x d1) y d2 = upd (upd h y d2) x d1.
Proof. 
rewrite /upd=>[[|h]] x y d1 d2 // H.
by case: ifP=>H1; case: ifP=>H2; rewrite ?H1 ?H2 // update_update_neq_s.
Qed.

Lemma validUE (h : heap) (x : loc) (d : dynamic) : 
        valid x (upd h x d) = def h && (x != null).
Proof.
rewrite /upd; case=>[|h] x d //=.
case: ifP=>H; rewrite H //.
by apply: valid_update_eq_s.
Qed.  

Lemma validUN (h : heap) (x : loc) (y : loc) (d : dynamic) : 
        x != y -> y != null -> valid x (upd h y d) = valid x h.
Proof. by case=>[|h] ???? H; rewrite /valid /upd // H -valid_update_neq_s. Qed.

Lemma freeUE (h : heap) (x : loc) (d : dynamic) : 
        x != null -> free x (upd h x d) = free x h.
Proof. by case=>[|h] x d H; rewrite /free /upd // H free_update_eq_s. Qed.

Lemma freeUN (h : heap) (x : loc) (y : loc) (d : dynamic) : 
        x != y -> free x (upd h y d) = upd (free x h) y d.
Proof.
case=>[|h] x y d H; rewrite /free /upd //.
by case: ifP=>H1; rewrite H1 // free_update_neq_s.
Qed.

Lemma updFE (h : heap) (x : loc) (d : dynamic) : 
        upd (free x h) x d = upd h x d.
Proof.
case=>[|h] x d; rewrite /free /upd //.
by case: ifP=>H; rewrite H // update_free_eq_s.
Qed.

Lemma updFN (h : heap) (x : loc) (y : loc) (d : dynamic) : 
        x != y -> upd (free x h) y d = free x (upd h y d).
Proof. by move=>*; rewrite freeUN. Qed.

Lemma lookUE (h : heap) (x : loc) (d : dynamic) : 
        def h -> x != null -> look x (upd h x d) = d.
Proof. by case=>[|h] x d H1 H2; rewrite /look /upd // H2 lookup_update_eq_s. Qed.

Hint Resolve lookUE.

Lemma lookUN (h : heap) (x : loc) (y : loc) (d : dynamic) : 
        x != y -> y != null -> look x (upd h y d) = look x h.
Proof. by case=>[|h] x y d H1 H2; rewrite /look /upd // H2 lookup_update_neq_s. Qed.

Lemma unUh (h1 h2 : heap) (x : loc) (d : dynamic) : 
        ~~ valid x h2 -> upd h1 x d :+ h2 = upd (h1 :+ h2) x d.
Proof. 
case=>[|h1]; case=>[|h2] // x d H; rewrite /upd /union2; case: ifP=>H1 //.
- rewrite H1 disjoint2_update_left_s //. 
  case: (disjoint2_s _ _)=>//.
  by rewrite update_union2_left_s.
by rewrite H1; case: ifP.
Qed.

Lemma unhU (h1 h2 : heap) (x : loc) (d : dynamic) : 
        ~~ valid x h1 -> h1 :+ upd h2 x d = upd (h1 :+ h2) x d.
Proof. by move=>h1 *; rewrite !(unC h1); apply: unUh. Qed. 

Lemma empUN (h : heap) (x : loc) (d : dynamic) : ~~ is_empty (upd h x d).
Proof. 
case=>[|h] x d //; rewrite /upd.
by case: ifP=>H //; rewrite is_empty_update_s.
Qed.

(***************************)
(* properties of valid_loc *)
(***************************)

Lemma valid_emp (x : loc) : valid x empty = false.
Proof. by []. Qed.

Lemma valid_null (h : heap) : valid null h = false.
Proof. by case=>// h; apply: negbET; apply: valid_null_s. Qed.

Lemma validFN (h : heap) (x : loc) : valid x (free x h) = false.
Proof. by case=>// *; apply: negbET; apply: valid_free_eq_s. Qed.

Lemma validFE (h : heap) (x y : loc) : x != y -> valid x (free y h) = valid x h.
Proof. by case=>// *; rewrite /valid /free valid_free_neq_s. Qed.

Lemma validUn (h1 h2 : heap) (x : loc) : 
        valid x (h1 :+ h2) = def (h1 :+ h2) && (valid x h1 || valid x h2).
Proof.
case=>[|h1]; case=>[|h2] // x; rewrite /union2 /valid.
case: (disjoint2_s _ _)=>//.
by rewrite valid_union2_s.
Qed.

CoInductive valid_some_spec (h1 h2 : heap) : bool -> Type :=
| valid_some_true of def h1 & def h2 & 
    (forall x:loc, valid x h1 -> ~~ valid x h2) : valid_some_spec h1 h2 true
| valid_some_false of (exists x:loc, valid x h1 /\ valid x h2) : valid_some_spec h1 h2 false
| valid_some_false1 of ~~ def h1 : valid_some_spec h1 h2 false
| valid_some_false2 of ~~ def h2 : valid_some_spec h1 h2 false.

Lemma validSP (h1 h2 : heap) : valid_some_spec h1 h2 (def (h1 :+ h2)).
Proof.
rewrite /union2; case=>[|h1]; case=>[|h2] /=; 
  try by [apply: valid_some_false1 | apply: valid_some_false2].
case: ifP=>/= H.
- apply: valid_some_true=>// [x] /=.
  by case: valid_disjoint2_P1_s H x.
apply: valid_some_false.
by case: valid_disjoint2_PN_s (negbT H)=>// [[x][H1]H2 _].
Qed.

(**************************)
(* properties of free_loc *)
(**************************)

CoInductive free_valid_spec (h : heap) (x : loc) : bool -> Type :=
| free_valid_true of free x h <> h : free_valid_spec h x true
| free_valid_false of free x h = h : free_valid_spec h x false.

(* the lemma below is a misnomer; all this has to be rearranged *)
Lemma freeVP (h : heap) (x : loc) : free_valid_spec h x (valid x h).
Proof.
case=>[|h] x; rewrite /valid; first by apply: free_valid_false.
case E : (valid_s x h); [left | right]; rewrite /free.
- move=>[H]; rewrite -H in E. 
  by move: E (valid_free_eq_s h x)=>->.
congr Def. (* WARNING: This should be an independent lemma *)
move: h E => [h p] E; rewrite /valid_s /free_s /= in E *. 
apply: heap_func_s=>/=.
elim: h p E => [| y d h IH] //=.
move/andP=>[H1 H2].
case: eqP=>// H3 H4.
congr cons_c.
by apply: IH.
Qed.

Lemma free0E (x : loc) : free x empty = empty.
Proof. by move=>* //; rewrite /free /empty free_empty_s. Qed.
 
Lemma freeFE (h : heap) (x : loc) :  free x (free x h) = free x h.
Proof. by case=>// *; rewrite /free free_free_eq_s. Qed.

Lemma freeFN (h : heap) (x y : loc) :  x != y -> free x (free y h) = free y (free x h).
Proof. by case=>// *; rewrite /free free_free_neq_s. Qed.

Lemma lookFE (h : heap) (x : loc) : look x (free x h) = dyn tt.
Proof. by case=>// *; rewrite /look /free lookup_free_eq_s. Qed.

Lemma lookFN (h : heap) (x y : loc) : x != y -> look x (free y h) = look x h.
Proof. by case=>// *; rewrite /look /free lookup_free_neq_s. Qed.

Lemma unFF (h1 h2 : heap) (x : loc) : 
        def (h1 :+ h2) -> (free x h1) :+ free x h2 = free x (h1 :+ h2).
Proof.  
rewrite /def /union2 /free.
case=>[|h1]; case=>[|h2] // x; case: ifP=>// H _.
rewrite disjoint2_free_left_s; last by rewrite disjoint2_free_right_s.
by rewrite free_union2_s.
Qed.  

Lemma unFh (h1 h2 : heap) (x : loc) : ~~ valid x h2 -> free x h1 :+ h2 = free x (h1 :+ h2).
Proof.
rewrite /valid /free /union2; case=>[|h1]; case=>[|h2] // x H1. 
rewrite disjoint2_free_left_eq_s //.
case H2: (disjoint2_s h1 h2)=>//.
by rewrite free_union2_left_s.
Qed.

Lemma unhF (h1 h2 : heap) (x : loc) : ~~ valid x h1 -> h1 :+ free x h2 = free x (h1 :+ h2).
Proof. by move=>h1 h2 x H; rewrite !(unC h1) unFh. Qed.

Lemma unFhS (h1 h2 : heap) (x : loc) : def (h1 :+ h2) -> def (free x h1 :+ h2).
Proof.
rewrite /def /free /union2.
case=>[|h1]; case=>[|h2] // x; case: ifP=>// H.
by rewrite disjoint2_free_left_s.
Qed.

Lemma unhFS (h1 h2 : heap) (x : loc) : def (h1 :+ h2) -> def (h1 :+ free x h2).
Proof. by move=>h1 h2 x H; rewrite (unC h1) unFhS // unC. Qed. 

(************************)
(* properties of lookup *)
(************************)

Lemma look0 (x : loc) : look x empty = dyn tt.
Proof. by []. Qed.

Lemma lookV (h : heap) (x : loc) : ~~ valid x h -> look x h = dyn tt.
Proof.
rewrite /valid /look; case=>// h x H. 
by apply: valid_lookup_s.
Qed.

Lemma lookhUn (h1 h2 : heap) (x : loc) : def (h1 :+ h2) -> valid x h1 -> look x (h1 :+ h2) = look x h1.
Proof.
rewrite /def /union2 /valid /look.
case=>// h1; case=>// h2 x.
case E: (disjoint2_s _ _)=>// _ H.
rewrite lookup_union2_some_intro1_s //.
case: valid_disjoint2_P1_s E=>// H1 _.
by apply: H1.
Qed.

Lemma lookUnh (h1 h2 : heap) (x : loc) : def (h1 :+ h2) -> valid x h2 -> look x (h1 :+ h2) = look x h2.
Proof. by move=>h1 h2 x; rewrite unC; apply: lookhUn. Qed.

Lemma lookhUnN (h1 h2 : heap) (x : loc) : def (h1 :+ h2) -> ~~ valid x h1 -> look x (h1 :+ h2) = look x h2.
Proof.
rewrite /def /union2 /valid /look.
case=>// h1; case=>// h2 x.
case E1 : (disjoint2_s _ _)=>// _ H.
case E2 : (valid_s x h2).
- by rewrite lookup_union2_some_intro2_s.
rewrite valid_lookup_s.
- by rewrite valid_lookup_s ?E2.
by rewrite valid_union2_s (negbET H) E2.
Qed. 

Lemma lookUnhN (h1 h2 : heap) (x : loc) : def (h1 :+ h2) -> ~~ valid x h2 -> look x (h1 :+ h2) = look x h1.
Proof. by move=>h1 h2 x; rewrite unC; apply: lookhUnN. Qed.
  
(**************************************)
(* properties of union2 and disjoint2 *)
(**************************************)

(* the ring properties have already been proved above, *)
(* so there is only one more lemma left to do: *)

Lemma empUn (h1 h2 : heap) : is_empty (h1 :+ h2) = is_empty h1 && is_empty h2.
Proof.
rewrite /is_empty /union2.
case=>[|h1]; case=>[|h2] // *; first by rewrite andbF.
case H: (disjoint2_s _ _); first by rewrite -is_empty_union2_s.
apply: sym_eq; apply: negbET; apply/negP; move/andP=>[H1 H2].
by rewrite (is_emptyP_s _ H1) (is_emptyP_s _ H2) disjoint2_empty_left_s in H. 
Qed.

(* some transformation lemmas *)

Lemma emphT (h : heap) : h = empty -> is_empty h.
Proof. by move=>h ->. Qed.

Lemma empTh (h : heap) : empty = h -> is_empty h.
Proof. by move=>h <-. Qed.  

Lemma unhh0E (h1 h2 : heap) : h1 :+ h2 = empty -> h1 = empty /\ h2 = empty.
Proof. by move=>h1 h2; move/emphT; rewrite empUn; do 2![case: is_emptyP=>//=]. Qed. 

Lemma emp_pt (A : Type) (x : loc) (v : A) : is_empty (x :-> v) = false.
Proof. by move=>*; rewrite (negbET (empUN _ _ _)). Qed. 
  
(****************************************************)
(* Plus an extra function and lemma about freshness *)
(****************************************************)

Lemma freshness_c (h : heap_c) (n : nat) : sorted_c h -> ~~ valid_c ((last_loc h).+1 + n) h.
Proof. by move=>h n H; rewrite last_valid // addSn ltnS leq_addr. Qed.  

Lemma freshness_s (h : heap_s) (n : nat) : ~~ valid_s (fresh_s h .+ n) h.
Proof. by move=>[h pf] n; apply: freshness_c. Qed.

Lemma freshness (h : heap) (n : nat) : ~~ valid (fresh h .+ n) h.
Proof. by case=>// h; apply: freshness_s. Qed.

Lemma fresh_null h : fresh h != null.
Proof. by case. Qed.  

Opaque fresh.

(******************************)
(* Some derived useful lemmas *)
(******************************)

Lemma validPE (A : Type) (v : A) (x : loc) (y : loc) : 
        valid x (y :-> v) = (x == y) && (y != null).
Proof.
move=>A v x y.
case E: (x == y); first by rewrite (eqP E) validUE.
case H: (y == null); first by rewrite (eqP H).
by rewrite validUN // ?E ?H.
Qed.

Lemma validPx (A : Type) (v : A) (x : loc) : valid x (x :-> v) = (x != null).
Proof. by move=>*; rewrite validPE eq_refl. Qed.

Lemma noalias (h1 h2 : heap) (x1 x2 : loc) : 
        valid x1 h1 -> valid x2 h2 -> def (h1 :+ h2) -> x1 != x2.
Proof.
move=>h1 h2 x1 x2 H1 H2.
case: validSP=>// _ _ H _.
by apply: contra (H _ H1); move/eqP=>->.
Qed. 

Lemma defPUn (A : Type) (h : heap) (x : loc) (v : A) : 
        def (x :-> v :+ h) = [&& x != null, def h & ~~ valid x h].
Proof. 
move=>A h x v. 
case E : (valid x h)=>/=; rewrite ?andbF ?andbT. 
- case: validSP=>// H1 H2; move/(_ x).
  by rewrite validPx -(defPt _ v) E; move/(_ H1).
case: validSP=>//.
- by rewrite defPt=>->->.
- by move=>[y []]; rewrite validPE; case/andP; move/eqP=>->; rewrite E.
- by rewrite defPt negbK=>->.
by move/negbET=>->; rewrite andbF.
Qed. 

Lemma defUnP (A : Type) (h : heap) (x : loc) (v : A) : 
        def (h :+ x :-> v) = [&& x != null, def h & ~~ valid x h].
Proof. by move=>*; rewrite unC defPUn. Qed.  

Lemma defNf A (v : A) h : def (null :-> v :+ h) = false.
Proof. by move=>*; rewrite defPUn. Qed.

Lemma defPN A p (v : A) h : def (p :-> v :+ h) -> p != null.
Proof. by move=>A p v h; rewrite defPUn; case/and3P. Qed.    

Lemma validPUn (A : Type) (h : heap) (x : loc) (v : A) : 
        valid x (x :-> v :+ h) = [&& x != null, def h & ~~ valid x h].
Proof.
move=>A h x v; rewrite validUn defPUn validPx.
case H: {2}(valid x h)=>/=; first by rewrite orbT andbT. 
by rewrite orbF andb_comm !andbA andbb.
Qed.
 
Lemma validUnP (A : Type) (h : heap) (x : loc) (v : A) : 
        valid x (h :+ x :-> v) = [&& x != null, def h & ~~ valid x h].
Proof. by move=>*; rewrite unC validPUn. Qed.

Lemma validUD A h x (v : A) : valid x (x :-> v :+ h) = def (x :-> v :+ h).
Proof. by move=>*; rewrite validPUn defPUn. Qed.

Lemma validDU A h x (v : A) : valid x (h :+ x :-> v) = def (x :-> v :+ h).
Proof. by move=>*; rewrite unC validUD. Qed.    

Lemma lookPUn (A : Type) (h : heap) (x : loc) (v : A) : 
        def (x :-> v :+ h) -> look x (x :-> v :+ h) = dyn v.
Proof.
move=>A h x v; rewrite defPUn; case/and3P=>H1 H2 H3. 
by rewrite lookhUn // ?lookUE // ?defPUn ?validPx ?H1 ?H2 ?H3.
Qed.

Lemma lookUnP (A : Type) (h : heap) (x : loc) (v : A) : 
        def (h :+ x :-> v) -> look x (h :+ x :-> v) = dyn v.
Proof. by move=>*; rewrite unC; apply: lookPUn; rewrite unC. Qed.

Lemma splitV (h : heap) (x : loc) : 
        valid x h -> h = (x :-> (val (look x h))) :+ free x h.
Proof.
move=>h x H; rewrite unUh ?validFN //.
by rewrite un0h updFE -dyn_eta -upd_eta.
Qed.

Lemma splitVL (A : Type) (h : heap) (x : loc) (v : A) : 
        valid x h -> look x h = dyn v -> h = (x :-> v) :+ free x h.
Proof.
move=>A h x v H1 H2; rewrite unUh ?validFN //.
by rewrite un0h updFE -H2 -upd_eta.
Qed.

Lemma pts_injT (A1 A2 : Type) (x : loc) (v1 : A1) (v2 : A2) : 
        x != null -> x :-> v1 = x :-> v2 -> A1 = A2.
Proof. 
by move=>A1 A2 x v1 v2 H; rewrite /upd H /= /update_s; case; case: dec=>// i; case.
Qed.  

Lemma pts_inj (A : Type) (x : loc) (v1 v2 : A) : 
        x != null -> x :-> v1 = x :-> v2 -> v1 = v2.
Proof. 
move=>A x v1 v2 H; rewrite /upd H /= /update_s; case; case: dec=>// i.
by case; move/(inj_pairT2 _ ssrfun.id _ _ _).
Qed. 

Lemma pts_injSTl (A1 A2 : Type) (h1 h2 : heap) (x : loc) (v1 : A1) (v2 : A2) : 
        def (x :-> v1 :+ h1) -> x :-> v1 :+ h1 = x :-> v2 :+ h2 -> A1 = A2.
Proof.
move=>A1 A2 h1 h2 x v1 v2 H1 H2; move: (H1); rewrite H2=>H3. 
have : look x (x :-> v1 :+ h1) = look x (x :-> v2 :+ h2) by rewrite H2.
do 2!rewrite lookPUn //.
by apply: dyn_injT.
Qed.

Lemma pts_injSTr (A1 A2 : Type) (h1 h2 : heap) (x : loc) (v1 : A1) (v2 : A2) : 
        def (x :-> v2 :+ h2) -> x :-> v1 :+ h1 = x :-> v2 :+ h2 -> A1 = A2.
Proof. by move=>A1 A2 h1 h2 x v1 v2 H1 H2; apply: pts_injSTl (H2); rewrite H2. Qed.

Lemma pts_injSl (A : Type) (h1 h2 : heap) (x : loc) (v1 v2 : A) : 
        def (x :-> v1 :+ h1) -> 
        x :-> v1 :+ h1 = x :-> v2 :+ h2 -> v1 = v2 /\ h1 = h2.
Proof.
move=>A h1 h2 x v1 v2 H1 H2; move: (H1); rewrite H2=>H3.
have : look x (x :-> v1 :+ h1) = look x (x :-> v2 :+ h2) by rewrite H2.
do 2![rewrite lookPUn //]=>H4.
move: (dyn_inj H4)=>{H4} H4.
rewrite H4 in H1 H2 H3 *.
by rewrite (unhKl H1 H2).
Qed.

Lemma cancel A h1 h2 x (v1 v2 : A) : 
        def (x :-> v1 :+ h1) -> 
        x :-> v1 :+ h1 = x :-> v2 :+ h2 -> [/\ v1 = v2, def h1 & h1 = h2].
Proof.
move=>A h1 h2 x v1 v2 D; case/(pts_injSl D)=>-><-; split=>//.
by apply: unDr D.
Qed.  

Lemma pts_injSr (A : Type) (h1 h2 : heap) (x : loc) (v1 v2 : A) : 
        def (x :-> v2 :+ h2) -> 
        x :-> v1 :+ h1 = x :-> v2 :+ h2 -> v1 = v2 /\ h1 = h2.
Proof.
move=>A h1 h2 x v1 v2 H T; symmetry in T.
by move: (pts_injSl H T)=>[<- <-].
Qed.

Lemma validP (A : Type) (v : A) (x : loc) (i : heap) : 
        def (x :-> v :+ i) -> ~~ valid x i.
Proof. by move=>A v x i; rewrite defPUn; case/and3P. Qed.
 
Lemma unP (A : Type) (v : A) (x : loc) (i : heap) : 
        def (x :-> v :+ i) -> x :-> v :+ i = upd i x (dyn v).
Proof.
move=>A v x i H.
move: (validP H) => T. 
by rewrite unUh // un0h.
Qed. 

(******************************)
(* Properties of block update *)
(******************************)

Section BlockUpdate.
Variable (A : Type).

Fixpoint updi (x : loc) (vs : seq A) {struct vs} : heap := 
  if vs is v'::vs' then (x :-> v') :+ updi (x .+ 1) vs' else empty.

Lemma updiS x v vs : updi x (v :: vs) = x :-> v :+ updi (x .+ 1) vs.
Proof. by []. Qed.   

Lemma updi_last x v vs : 
        updi x (add_last vs v) = updi x vs :+ (x .+ size vs) :-> v.
Proof.
move=>x v vs; elim: vs x v=>[|w vs IH] x v /=; first by rewrite offset0 unh0 un0h.
by rewrite -(addn1 (size vs)) addnC -offsetA IH unA.
Qed.

Lemma updi_cat x vs1 vs2 : updi x (vs1 ++ vs2) = updi x vs1 :+ updi (x .+ size vs1) vs2.
Proof.
move=>x vs1; elim: vs1 x=>[|v vs1 IH] x vs2 /=; first by rewrite offset0 un0h.
by rewrite -(addn1 (size vs1)) addnC -offsetA IH unA. 
Qed.

Lemma updi_catI x y vs1 vs2 : 
        y = x .+ size vs1 -> updi x vs1 :+ updi y vs2 = updi x (vs1 ++ vs2).
Proof. by move=>x y vs1 vs2 ->; rewrite updi_cat. Qed.

(* Helper lemma *)
Lemma updiVm' (x : loc) m xs : m > 0 -> valid x (updi (x .+ m) xs) = false.
Proof. 
move=>x m xs; elim: xs x m=>[|v vs IH] x m //= H.
rewrite offsetA validUn validPE -{3}(offset0 x) offsetK. 
case: eqP H=>[->|_ H /=]; first by rewrite ltnn. 
by rewrite IH // ?addn1 ?andbF.
Qed.

Lemma updiD x xs : def (updi x xs) = (x != null) || (size xs == 0).
Proof.
move=>x xs; elim: xs x=>[|v xs IH] x //=; first by rewrite orbC.
by rewrite defPUn updiVm' // andbT orbF IH offset_null /= andbT.
Qed.

Lemma updiVm (x : loc) m xs : 
        valid x (updi (x .+ m) xs) = [&& x != null, m == 0 & size xs > 0].
Proof.
move=>x m xs; case: m=>[|m] /=; last by rewrite andbF updiVm'.
case: xs=>[|v xs]; rewrite offset0 ?andbF ?andbT //=.
by rewrite validUn validPE eq_refl /= -updiS updiD updiVm' // orbF andbb.
Qed.

Lemma updimV x m xs : valid (x .+ m) (updi x xs) = (x != null) && (m < size xs).
Proof.
move=>x m vs; case H: (x == null)=>/=.
- by case: vs=>// a s; rewrite (eqP H).
elim: vs x m H=>[|v vs IH] x m H //; case: m=>[|m].
- by rewrite offset0 /= validUn validPE eq_refl H /= andbT -updiS updiD // H.
rewrite -addn1 addnC -offsetA updiS validUn validPE IH; last first.
- by rewrite loc_inv /= addn1. 
by rewrite H /= andbT -updiS updiD ?H //= -{2}(offset0 x) offsetA offsetK.
Qed.

Lemma updiP x y xs : reflect (y != null /\ exists m, x = y .+ m /\ m < size xs) 
                     (valid x (updi y xs)).
Proof.
move=>x y xs; case H: (y == null)=>/=.
- rewrite (eqP H); elim: xs=>[|z xs IH] //=.
  - by rewrite valid_emp; constructor; case.
  by constructor; case.
case E: (valid _ _); constructor; last first.
- by move=>[_][m][H1] H2; rewrite H1 updimV H2 H in E.
case: (offset_total x y) E=>[m]; case/orP; move/eqP=>->.
- by rewrite updimV H /= => H1; split=>//; exists m.
rewrite updiVm; case/and3P=>H1; move/eqP=>-> H2.
by split=>//; exists 0; rewrite offsetA addn0 offset0.
Qed.

(* Invertibility *)
Lemma updi_inv x xs1 xs2 : def (updi x xs1) -> updi x xs1 = updi x xs2 -> xs1 = xs2.
Proof.
move=>x xs1; elim: xs1 x =>[|v1 xs1 IH] x /=; case=>[|v2 xs2] //= H; 
  try by (move/empTh || move/emphT); rewrite empUn emp_pt.
have L: def (x :-> v2 :+ updi (x .+ 1) xs2).
- rewrite defPUn updiD /= ?offset_null // updiVm.
  by rewrite defPUn in H; case/and3P: H=>->.
case/(pts_injSr L)=>->; move/IH=>-> //; apply: unDr H.
Qed.

Lemma updi_iinv x xs1 xs2 h1 h2 : 
        size xs1 = size xs2 -> def (updi x xs1 :+ h1) ->        
        updi x xs1 :+ h1 = updi x xs2 :+ h2 -> xs1 = xs2 /\ h1 = h2.
Proof.
move=>x xs1; elim: xs1 x=>[|v1 xs1 IH] x /=; case=>[|v2 xs2] //= h1 h2.
- by rewrite !un0h.
move=>[E]; rewrite -!unA=>H; case/(pts_injSl H)=>->; rewrite defPUn in H.
by case/and3P: H=>_ H _; case/(IH _ _ _ _ E H)=>->->.
Qed.

End BlockUpdate. 

(**************************************************************************)
(* Several tactics for canceling common terms in disjoint unions          *)
(* Currently, they don't deal with weak pointers. I.e.  they only if they *)
(* see iterms like x :-> v1 and x :-> v2, they will reduce to v1 = v2     *)
(* only if v1, v2 are of the same type A more general tactic would emit   *)
(* obligation dyn v1 = dyn v2, but I don't bother with this now.          *)
(**************************************************************************)

(* First cancelation in hypotheses *)

Lemma injUh (A : Type) (h1 h2 : heap) (x : loc) (v1 v2 : A) : 
        def (h1 :+ (x :-> v1)) -> 
        h1 :+ (x :-> v1) = h2 :+ (x :-> v2) -> 
          def h1 /\ h1 = h2 /\ v1 = v2. 
Proof. 
move=>A h1 h2 x v1 v2 H1 H2.
rewrite unC in H1; rewrite -!(unC (x :-> _)) in H2; symmetry in H2.
case: (pts_injSr H1 H2) (H1)=>[->->]. 
by rewrite defPUn; case/and3P.
Qed.  

Lemma eqUh (h1 h2 h : heap) : 
        def (h1 :+ h) -> h1 :+ h = h2 :+ h -> def h1 /\ h1 = h2. 
Proof. 
move=>h1 h2 h H1 H2.
split; first by apply: unDl H1.
by apply: unKhl H2.
Qed.

Lemma exit1 (h1 h2 h : heap) : 
        def (h1 :+ h) -> h1 :+ h = h :+ h2 -> def h1 /\ h1 = h2.
Proof.
move=>h1 h2 h H1 H2.
split; first by apply: unDl H1.
by rewrite -unC in H1 H2; apply: unhKl H2.
Qed.

Lemma exit2 (h1 h : heap) : def (h1 :+ h) -> h1 :+ h = h -> def h1 /\ h1 = empty.
Proof. by move=>h1 h H1; rewrite -{2}(unh0 h)=>H2; apply: exit1 H2. Qed.

Lemma exit3 (h1 h : heap) : def h -> h = h :+ h1 -> def empty /\ empty = h1.
Proof.
move=>h1 h H1 H2. split=>//. 
rewrite -{1}(unh0 h) in H2.
apply: unhKl H2. by rewrite unh0.
Qed.

Lemma exit4 (h : heap) : def h -> h = h -> def empty /\ empty = empty.
Proof. by []. Qed. 

Ltac cancelator t H :=
  match goal with 
  (* we exit when we hit the terminator on the left *)
  | |- ?h1 :+ t = ?h2 -> _ => 
     let j := fresh "j" in
     set j := {1}(h1 :+ t); 
     rewrite -1?unA /j {j};
     (move/(exit1 H)=>{H} [H] || move/(exit2 H)=>{H} [H]) 
  | |- t = ?h2 -> _ => 
     rewrite -?unA;
     (move/(exit3 H)=>{H} [H] || move/(exit4 H)=>{H} [H])
  | |- (?h1 :+ (?x :-> ?v) = ?h2) -> _ => 
    let j := fresh "j" in 
    set j := {1}(h1 :+ (x :-> v));
    (* if x appears in the second union, first bring it to the back *)
    rewrite 1?(unC (x :-> _)) -?(unAC _ _ (x :-> _)) /j {j}; 
    (* then one of the following must apply *)
    (* if x is in the second union then cancel *)
    (move/(injUh H)=>{H} [H []] ||
    (* if not, rotate x in the first union *)
     rewrite (unC h1) ?unA in H * );
    (* and proceed *)
    cancelator t H
  (* if the heap is not a points-to relation, also try to cancel *)
  | |- (?h1 :+ ?h = ?h2) -> _ => 
    let j := fresh "j" in 
    set j := {1}(h1 :+ h);
    (* if h appears in the second union, first bring it to the back *)
    rewrite 1?(unC h) -?(unAC _ _ h) /j {j}; 
    (* then one of the following must apply *)
    (* if h is in the second union then cancel *)
    (move/(eqUh H)=>{H} [H []] ||
    (* if not, rotate h in the first union *)
    rewrite (unC h1) ?unA in H * );
    (* and proceed *)
    cancelator t H
  | |- _ => idtac 
  end.

(*
Ltac cancelator t H :=
  match goal with 
  (* we exit when we hit the terminator on the left *)
  | |- ?h1 :+ t = ?h2 -> _ => 
     let j := fresh "j" in
     set j := {1}(h1 :+ t); 
     rewrite -1?unA /j {j};
     (move/(exit1 H)=>{H} [H] || move/(exit2 H)=>{H} [H]) 
  | |- t = ?h2 -> _ => 
     rewrite -1?unA;
     (move/(exit3 H)=>{H} [H] || move/(exit4 H)=>{H} [H])
  | |- (?h1 :+ (?x :-> ?v) = ?h2) -> _ => 
    let j := fresh "j" in 
    set j := {1}(h1 :+ (x :-> v));
    (* if x appears in the second union, first bring it to the back *)
    rewrite 1?(unC (x :-> _)) -?(unAC _ _ (x :-> _)) /j {j}; 
    (* then one of the following must apply *)
    (* if x is in the second union then cancel *)
    (move/(injUh H)=>{H} [H []] ||
    (* if not, rotate x in the first union *)
     rewrite (unC h1) ?unA in H * );
    (* and proceed *)
    cancelator t H
  (* if the heap is not a points-to relation, also try to cancel *)
  | |- (?h1 :+ ?h = ?h2) -> _ => 
    let j := fresh "j" in 
    set j := {1}(h1 :+ h);
    (* if h appears in the second union, first bring it to the back *)
    rewrite 1?(unC h) -?(unAC _ _ h) /j {j}; 
    (* then one of the following must apply *)
    (* if h is in the second union then cancel *)
    (move/(eqUh H)=>{H} [H []] ||
    (* if not, rotate h in the first union *)
    rewrite (unC h1) ?unA in H * );
    (* and proceed *)
    cancelator t H
  | |- _ => idtac 
  end.
*)

Ltac heap_cancel := 
  match goal with 
  | |- ?h1 = ?h2 -> ?GG =>
    let t1 := fresh "t1" in
    let t2 := fresh "t2" in 
    let t := fresh "t" in
    let H := fresh "H" in
    let G := fresh "hidden goal" in
      (* generate the obligation to prove that the left heap is defined *)
      suff : def h1; first (
       (* make sure no sharing of expressions in the goal *)
       set t1 := {1 2}h1; set t2 := {1}h2; set G := GG;
       (* introduce terminators *)
       rewrite -(un0h t1) -(un0h t2) [empty]lock;
       set t := locked empty; rewrite /t1 /t2 {t1 t2}; 
       move=>H; 
       (* flatten the goal *)
       rewrite ?unA in H *; 
       (* call the cancelation routine *)
       cancelator t H; 
       (* remove the terminator and push H onto the goal *)
       move: H {t}; rewrite /G {G})
  | |- _ => idtac
  end.

(* Then cancelation in conclusions *)

Lemma cexit1 (h1 h2 h : heap) : h1 = h2 -> h1 :+ h = h :+ h2.
Proof. by move=>h1 h2 h ->; rewrite unC. Qed.

Lemma cexit2 (h1 h : heap) : h1 = empty -> h1 :+ h = h.
Proof. by move=>h1 h ->; rewrite un0h. Qed.

Lemma cexit3 (h1 h : heap) : empty = h1 -> h = h :+ h1.
Proof. by move=>h1 h <-; rewrite unh0. Qed.

Lemma congUh (A : Type) (h1 h2 : heap) (x : loc) (v1 v2 : A) : 
        h1 = h2 -> v1 = v2 -> h1 :+ (x :-> v1) = h2 :+ (x :-> v2).
Proof. by move=>A h1 h2 x v1 v2 -> ->. Qed.

Lemma congeqUh h1 h2 h : h1 = h2 -> h1 :+ h = h2 :+ h.
Proof. by move=>h1 h2 h ->. Qed. 

Ltac congruencer t :=
  match goal with 
  | |- ?h1 :+ t = ?h2 => 
     let j := fresh "j" in
     set j := {1}(h1 :+ t); 
     rewrite -1?unA /j {j};
     (apply: cexit1 || apply: cexit2)
  | |- t = ?h2 => 
     rewrite -1?unA;
     (apply: cexit3 || apply: refl_equal)
  | |- (?h1 :+ (?x :-> ?v) = ?h2) => 
    let j := fresh "j" in 
    set j := {1}(h1 :+ (x :-> v));
    (* if x appears in the second union, first bring it to the back *)
    rewrite 1?(unC (x :-> _)) -?(unAC _ _ (x :-> _)) /j {j}; 
    (* then one of the following must apply *)
    (* if x is in the second union then cancel *)
    ((apply: congUh; [congruencer t | idtac]) || 
    (* if not, rotate x in the first union *)
     (rewrite (unC h1) ?unA; congruencer t))
  (* if the heap is not a points-to relation, also try to cancel *)
  | |- (?h1 :+ ?h = ?h2) => 
    let j := fresh "j" in 
    set j := {1}(h1 :+ h);
    (* if h appears in the second union, first bring it to the back *)
    rewrite 1?(unC h) -?(unAC _ _ h) /j {j}; 
    (* then one of the following must apply *)
    (* if h is in the second union then cancel *)
    (apply: congeqUh || 
    (* if not, rotate h in the first union *)
    rewrite (unC h1) ?unA);
    (* and proceed *)
    congruencer t
  | |- _ => idtac 
  end.

Ltac heap_congr := 
  match goal with 
  | |- ?h1 = ?h2 =>
    let t1 := fresh "t1" in
    let t2 := fresh "t2" in 
    let t := fresh "t" in
      set t1 := {1}h1; set t2 := {1}h2; 
      (* introduce terminators *)
      rewrite -(un0h t1) -(un0h t2) [empty]lock;
      set t := locked empty; rewrite /t1 /t2 {t1 t2}; 
      (* flatten the goal *)
      rewrite ?unA; 
      (* call the congruence routine and remove the terminator *)
      congruencer t=>{t}
  | |- _ => idtac
  end.

Lemma test (h1 h2 h3 : heap) (x : loc) (v1 v2 : nat) : 
        h3 = h2 -> v1 = v2 -> 
        h1 :+ (x :-> v1) :+ h3= h2 :+ h1 :+ (x :-> v2).
Proof. by move=>h1 h2 h3 x v1 v2 H1 H2; heap_congr. Qed.

(* and a tactic for computing the subdom relation *)

Definition supdom h2 h1 := subdom h1 h2.

Lemma sexit1 (h1 h2 h : heap) : 
        def (h2 :+ h) -> 
          (def h2 -> supdom h2 h1) -> supdom (h2 :+ h) (h :+ h1).
Proof.
rewrite /supdom. 
move=>h1 h2 h H1 H2.
rewrite (unC h).
apply: subdomE=>//.
by apply: H2; apply: (unDl H1).
Qed.

Lemma sexit2 (h1 h : heap) : 
        def (h1 :+ h) -> (def h1 -> supdom h1 empty) -> 
          supdom (h1 :+ h) h.
Proof.
rewrite /supdom.
move=>h1 h2 H1 H2.
rewrite -{1}(un0h h2).
apply: subdomE=>//.
apply: H2. 
apply: unDl H1.
Qed.

Lemma sexit3 (h1 h : heap) : 
        def h -> (def empty -> supdom empty h1) -> 
          supdom h (h :+ h1).
Proof.
rewrite /supdom.
move=>h1 h2 H1 H2; rewrite unC. 
rewrite -{2}(un0h h2).
apply: subdomE; first by rewrite un0h.
by apply: H2.
Qed.

Lemma sexit4 (h : heap) : 
        def h -> (def empty -> empty = empty) -> 
          supdom h h.
Proof.
move=>h H1 H2; rewrite -(un0h h).
by apply: subdomE=>//; rewrite un0h.
Qed.

Lemma supdomUh (A B : Type) (h1 h2 : heap) (x : loc) (v1 : A) (v2 : B) : 
        def (h2 :+ (x :-> v2)) -> 
          (def h2 -> supdom h2 h1) -> 
            supdom (h2 :+ (x :-> v2)) (h1 :+ (x :-> v1)).
Proof.
rewrite /supdom=>A B h1 h2 x v1 v2 H1 H2.
apply: subdomUE=>//; first by apply: H2; apply: unDl H1.
by apply: subdomPE; rewrite defUnP in H1; case/and3P: H1.
Qed.

Lemma supdomeqUh (h1 h2 h : heap) : 
        def (h2 :+ h) -> 
          (def h2 -> supdom h2 h1) -> supdom (h2 :+ h) (h1 :+ h).
Proof. by move=>h1 h2 h; rewrite (unC h1); apply: sexit1. Qed.    

Lemma sup_defdef h1 h2 : def h2 -> supdom h2 h1 -> def h1.
Proof. by move=>h1 h2 H1; rewrite /supdom; move/subdom_def; rewrite H1 andbT. Qed.

Ltac supdom_checker t H :=
  match goal with 
  | |- is_true (supdom (?h1 :+ t) ?h2) => 
     let j := fresh "j" in
     set j := {1}(h1 :+ t); 
     rewrite -1?unA /j {j};
     (apply: (sexit1 H)=>{H} H || apply: (sexit2 H)=>{H} H)
  | |- is_true (supdom t ?h1) => 
     rewrite -1?unA;
     (apply: (sexit3 H)=>{H} H || apply: (sexit4 H)=>{H} H)
  | |- is_true (supdom (?h1 :+ (?x :-> ?v)) ?h2) => 
    let j := fresh "j" in 
    set j := {1}(h1 :+ (x :-> v));
    (* if x appears in the second union, first bring it to the back *)
    rewrite 1?(unC (x :-> _)) -?(unAC _ _ (x :-> _)) /j {j}; 
    (* then one of the following must apply *)
    (* if x is in the second union then cancel *)
    (apply: (supdomUh _ H)=>{H} H || 
    (* if not, rotate x in the first union *)
     (rewrite (unC h1) ?unA in H * )); supdom_checker t H
  (* if the heap is not a points-to relation, also try to cancel *)
  | |- is_true (supdom (?h1 :+ ?h) ?h2) => 
    let j := fresh "j" in 
    set j := {1}(h1 :+ h);
    (* if h appears in the second union, first bring it to the back *)
    rewrite 1?(unC h) -?(unAC _ _ h) /j {j}; 
    (* then one of the following must apply *)
    (* if h is in the second union then cancel *)
    (apply: (supdomeqUh H)=>{H} H || 
    (* if not, rotate h in the first union *)
    (rewrite (unC h1) ?unA in H * )); 
    (* and proceed *)
    supdom_checker t H
  | |- _ => idtac 
  end.

Ltac defcheck := 
  match goal with 
  | |- is_true (def ?h2) -> is_true (def ?h1) =>
    let t1 := fresh "t1" in
    let t2 := fresh "t2" in 
    let t := fresh "t" in
    let H := fresh "H" in
      set t2 := {1}h2; set t1 := {1}h1; 
      (* introduce terminators *)
      rewrite -(un0h t1) -(un0h t2) [empty]lock;
      set t := locked empty; rewrite /t1 /t2 {t1 t2}; 
      (* flatten the goal *)
      rewrite ?unA; 
      move=>H; 
      apply: (sup_defdef H); 
      (* call the subdom_cheker routine and remove the terminator *)
      supdom_checker t H; move: H {t}; rewrite /supdom
  | |- _ => idtac
  end.

Lemma test2 h1 h2 x (v1 v2 : nat) : subdom h1 h2 ->
        def (h2 :+ (x :-> v2)) -> def (h1 :+ (x :-> v1)).
Proof.
move=>h1 h2 x v1 v2 H.
by defcheck. 
Qed.







