(* A small theory of ordered types *)

Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq paths fintype.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Module OrdType. 
Section OrdMixin.
Variable T : Type.
(* Keep the pivot type separate from the ordering-related data. *)
(* This is better for inheritance. *)
Section Ord. 
Variable eq_mix : EqType.mixin T.
Structure ord_mix : Type := 
  mk_ord_mix {ord : rel T; 
              irr : irreflexive ord;
              trans : transitive ord;
              total : forall x y, [|| ord x y, EqType.eqd eq_mix x y | ord y x]}.
End Ord. 

CoInductive mixin : Type := Mixin eq_mix of ord_mix eq_mix.
Coercion eq_of_mix m := let: Mixin eqm _ := m in eqm.
Coercion ord_of_mix m := let: Mixin _ m' := m return ord_mix m in m'.
End OrdMixin.

Structure class : Type := Class {sort :> Type; _ : mixin sort}.
Coercion mixin_of T := let: Class _ m := T return mixin T in m.
Definition mkMixin (T : class) := Mixin T.
End OrdType.

Notation ordType := OrdType.class.
Notation OrdMixin := OrdType.Mixin.
Notation OrdClass := OrdType.Class.

Coercion ord_eqType (T : ordType) := EqClass T.
Canonical Structure ord_eqType.

Definition eqMixin_for_ord := nosimpl EqType.mixin_of.

Notation "{ 'ordMixin' T }" := (@OrdType.ord_mix T (eqMixin_for_ord _))
  (at level 0, format "{ 'ordMixin'  T }") : type_scope.

Notation "[ 'ordType' 'of' T 'for' oT ]" :=
  (match oT as s in ordType return {type of OrdClass for s} -> _ with
  | OrdClass _ m => fun k => k m end
  (@OrdClass T)) (at level 0, only parsing) : form_scope.

Notation "[ 'ordType' 'of' T ]" := [ordType of T for [is T <: ordType]]
  (at level 0, only parsing) : form_scope.

Definition mkOrdMix (eT : eqType) := @OrdType.mk_ord_mix _ eT. 

Definition mkOrdType (eT : eqType) := 
  let: EqClass T em := eT return OrdType.ord_mix eT -> ordType in
  fun m => OrdClass (@OrdMixin T em m).

Section Lemmas.
Variable T : ordType. 
Definition ord := OrdType.ord T.
Lemma irr : irreflexive ord. Proof. exact: OrdType.irr. Qed.  
Lemma trans : transitive ord. Proof. exact: OrdType.trans. Qed.
Lemma total : forall x y, [|| ord x y, x == y | ord y x]. 
Proof. exact: OrdType.total. Qed.
Lemma nsym : forall x y, ord x y -> ord y x -> False.
Proof. by move=>x y E1 E2; move: (trans E1 E2); rewrite irr. Qed. 
End Lemmas. 

Prenex Implicits ord.

Section Totality.
Variable K : ordType.  
 
CoInductive total_spec (x y : K) : bool -> bool -> bool -> Type :=
| total_spec_lt of ord x y : total_spec x y true false false
| total_spec_eq of x == y : total_spec x y false true false
| total_spec_gt of ord y x : total_spec x y false false true.

Lemma totalP x y : total_spec x y (ord x y) (x == y) (ord y x).
Proof.
move=>x y; case H1: (x == y).
- by rewrite (eqP H1) irr; apply: total_spec_eq.
case H2: (ord x y); case H3: (ord y x). 
- by elim: (nsym H2 H3). 
- by apply: total_spec_lt H2.
- by apply: total_spec_gt H3.
by move: (total x y); rewrite H1 H2 H3.
Qed.
End Totality. 

Section NatOrd.
Lemma irr_ltn_nat : irreflexive ltn. Proof. by move=>x; rewrite /= ltnn. Qed.
Lemma trans_ltn_nat : transitive ltn. Proof. by apply: ltn_trans. Qed.
Lemma total_ltn_nat : forall x y, [|| x < y, x == y | y < x]. 
Proof. by move=>*; case: ltngtP. Qed.

(* always define mixin first, then type *)
(* if the definition is inlined, canonical structures don't unify correctly *)
Definition nat_ordMixin := mkOrdMix irr_ltn_nat trans_ltn_nat total_ltn_nat.
Canonical Structure nat_ordType := mkOrdType nat_ordMixin.
End NatOrd.

Section ProdOrd.
Variables K T : ordType.

(* lexicographic ordering *)
Definition lex : rel (K * T) := 
  fun x y => if x.1 == y.1 then ord x.2 y.2 else ord x.1 y.1.

Lemma irr_lex : irreflexive lex.
Proof. by move=>x; rewrite /lex eq_refl irr. Qed.

Lemma trans_lex : transitive lex.
Proof.
move=>[x1 x2][y1 y2][z1 z2]; rewrite /lex /=.
case: ifP=>H1; first by rewrite (eqP H1); case: eqP=>// _; apply: trans.
case: ifP=>H2; first by rewrite (eqP H2) in H1 *; rewrite H1.
case: ifP=>H3; last by apply: trans. 
by rewrite (eqP H3)=>R1; move/(nsym R1).
Qed.

Lemma total_lex : forall x y, [|| lex x y, x == y | lex y x].
Proof.
move=>[x1 x2][y1 y2]; rewrite /lex /=.
case: ifP=>H1.
- rewrite (eqP H1) eq_refl -pair_eqE /= eq_refl /=; exact: total.
rewrite (eq_sym y1) -pair_eqE /= H1 /=.
by move: (total x1 y1); rewrite H1.
Qed.

Definition prod_ordMixin := mkOrdMix irr_lex trans_lex total_lex.
Canonical Structure prod_ordType := mkOrdType prod_ordMixin.
End ProdOrd.

Section FinTypeOrd.
Variable T : finType.

Definition ordf : rel (FinType.sort T) := 
  fun x y => index x (enum T) < index y (enum T). 

Lemma irr_ordf : irreflexive ordf.
Proof. by move=>x; rewrite /ordf ltnn. Qed.

Lemma trans_ordf : transitive ordf.
Proof. by move=>x y z; rewrite /ordf; apply: ltn_trans. Qed.

Lemma total_ordf : forall x y, [|| ordf x y, x == y | ordf y x].
Proof.
move=>x y; rewrite /ordf; case: ltngtP=>//= H; rewrite ?orbT ?orbF //.
have [H1 H2]: x \in enum T /\ y \in enum T by rewrite !mem_enum.
by rewrite -(sub_index x H1) -(sub_index x H2) H eq_refl.
Qed.

Definition fin_ordMixin := mkOrdMix irr_ordf trans_ordf total_ordf.
End FinTypeOrd.

Notation "[ 'ordMixin' 'of' T 'by' 'fin' ]" :=
  (fin_ordMixin _ : @OrdType.ord_mix T _) (at level 0).

Definition ordinal_ordMixin n := [ordMixin of 'I_n by fin].
Canonical Structure ordinal_ordType n := mkOrdType (ordinal_ordMixin n).




