(**************************************************************)
(* Verified typechecker for the simply-typed lambda calculus. *)
(**************************************************************)

Require Import ssreflect ssrbool ssrnat eqtype heaps. 
Require Import stmodel stsep vcgen.
Require Import List.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(* abstract syntax with deBruijn indices *)

Inductive tp : Type := Unit | Prod of tp & tp | Arr of tp & tp.

Inductive exp : Type := 
  Var of nat | Emp | Pair of exp & exp | Fst of exp | 
  Snd of exp | Lam of tp & exp | App of exp & exp.

(* Boolean type equality *)
Fixpoint eqtp (t1 t2 : tp) {struct t1} : bool :=
  match t1, t2 with 
      Unit, Unit => true
    | Prod p1 p2, Prod q1 q2 => (eqtp p1 q1) && (eqtp p2 q2)
    | Arr p1 p2, Arr q1 q2 => (eqtp p1 q1) && (eqtp p2 q2)
    | _, _ => false
  end.

(* Reflect into propositional type equality *)
Lemma eqtpP : reflect_eq eqtp.
Proof.
elim=>[|u1 IHu1 v1 IHv1|u1 IHu1 v1 IHv1]; case=>[|u2 v2|u2 v2] /=; 
try (case: (IHu1 u2); case: (IHv1 v2)=>H1 H2); constructor=>//;
by (move=>[] || rewrite H1 H2).
Qed.
  
(* Create the equality types *)
Canonical Structure tp_eqMixin := EqMixin eqtpP. 
Canonical Structure tp_eqType := EqClass tp_eqMixin.

(* Hoare construct for matching over tp *)
Program 
Definition Match_tp (A : Type) (t : tp) (p1 : spec A) (p2 p3 : tp -> tp -> spec A)
                    (c1 : STsep A p1) (c2 : forall u v, STsep A (p2 u v))
                    (c3 : forall u v, STsep A (p3 u v)) : 
             STsep A (match t with Unit => p1 | Prod u v => p2 u v | Arr u v => p3 u v end) :=
  match t with Unit => st.do c1 _ | Prod u v => st.do (c2 u v) _ | Arr u v => st.do (c3 u v) _ end.

(****************************************)
(* Specification of the typing judgment *)
(****************************************)

Inductive HasType (G : list tp) (e : exp) (t : tp) : Prop :=
| zvar G' of G = t::G' & e = Var 0 
| svar G' t' n of G = t'::G' & e = Var n.+1 & HasType G' (Var n) t 
| unitI of e = Emp & t = Unit 
| prodI e1 e2 t1 t2 of e = Pair e1 e2 & t = Prod t1 t2 & HasType G e1 t1 & HasType G e2 t2 
| prodE1 u s of e = Fst u & HasType G u (Prod t s) 
| prodE2 u s of e = Snd u & HasType G u (Prod s t) 
| arrI u s r of e = Lam s u & t = Arr s r & HasType (s::G) u r 
| arrE u v s of e = App u v & HasType G u (Arr s t) & HasType G v s. 

Hint Constructors HasType.

(****************************)
(* Terms have unique types. *)
(****************************)

Lemma unicity (e : exp) (G : list tp) (t1 t2 : tp) : 
        HasType G e t1 -> HasType G e t2 -> t1 = t2.
Proof.
elim=>[n||u IHu v IHv|e IH|e IH|s e IH|u IH v _] G t1 t2.
- elim: n G=>[|n IH] G; first by case=>// G1 -> _; case=>// G2 [<-].
  case=>// G1 s1 n1 -> [<-] H1; case=>// _ _ _ [_ <-] [<-].
  by apply: IH.
- by case=>// _ ->; case.
- case=>//_ _ t3 t4 [<- <-] -> H1 H2. 
  case=>//_ _ t5 t6 [<- <-] -> H3 H4.
  by rewrite (IHu _ _ _ H1 H3) (IHv _ _ _ H2 H4).
- case=>//_ s1 [<-] H1; case=>//_ s2 [<-] H2.
  by case: (IH _ _ _ H1 H2).
- case=>//_ s1 [<-] H1; case=>//_ s2 [<-] H2.
  by case: (IH _ _ _ H1 H2).
- case=>//_ _ r1 [<- <-] -> H1; case=>//_ _ r2 [<- <-] -> H2.
  by rewrite (IH _ _ _ H1 H2).
case=>//_ _ s1 [<- <-] H1 _; case=>//_ _ s2 [<- <-] H2 _. 
by case: (IH _ _ _ H1 H2).
Qed.
 
(*****************************************)
(* Implementation of the typing judgment *)
(* using monadic functional programming  *)
(*****************************************) 

Definition TypeError : exn := exn_from_nat 1. 

Definition spec_of_infer (G : list tp) (e : exp) : Type :=
  STsep tp (emp, 
            fun y i m => i = m /\ if y is Val t then HasType G e t else
                         y = Exn TypeError /\ forall t, ~HasType G e t).
Program
Definition pre_infer (G : list tp) (e : exp) 
                     (infer : forall (G : list tp) (e : exp), spec_of_infer G e) : 
           spec_of_infer G e :=
  match e with 
    | Emp => Do (ret Unit)
    | Var 0 => 
        match G with 
          | t::_ => Do (ret t)
          | _ => Do (Throw _ TypeError)
        end
    | Var (S n) => 
        match G with 
          | _::G1 => Do (infer G1 (Var n))
          | _ => Do (Throw _ TypeError)
        end
    | Pair e1 e2 => 
        Do (t1 <-- infer G e1;
            t2 <-- infer G e2;
            ret (Prod t1 t2))
    | Fst e1 => 
        Do (t <-- infer G e1;
            Match_tp t 
              (Throw _ TypeError)
              (fun t1 _ => ret t1)
              (fun _ _ => Throw _ TypeError))
    | Snd e2 => 
        Do (t <-- infer G e2;
            Match_tp t 
              (Throw _ TypeError)
              (fun _ t2 => ret t2)
              (fun _ _ => Throw _ TypeError))
    | Lam t1 e1 =>
        Do (t2 <-- infer (t1::G) e1; 
            ret (Arr t1 t2))
    | App e1 e2 => 
        Do (t1 <-- infer G e1; 
            t2 <-- infer G e2; 
            Match_tp t1 
              (Throw _ TypeError)
              (fun _ _ => Throw _ TypeError)
              (fun p q => If (p == t2) then ret q else Throw _ TypeError))
  end. 
Next Obligation. by case: H=><-; eauto. Qed.
Next Obligation. by case: H=><-; eauto. Qed.
Next Obligation.
case: H0=><-; apply: val_throw; do 2!split=>//.
by move=>t; case=>// G' H1; elim: (H t G').
Qed.
Next Obligation.
case: H=><-; apply: cons0=>//; case=>e _ [<-]; first by eauto.
move=>[[->]] H; do !split=>//.
by move=>t; case=>//_ _ _ [_ <-] [<-]; move/H. 
Qed.
Next Obligation.
case: H0=><-; apply: val_throw; do !split=>//.
by move=>t; case=>// G' t' n' H1; elim: (H t' G').
Qed.
Next Obligation.
case: H=><- {i}; rewrite -(unh0 empty). 
apply: bnd_do=>//; last first.
- move=>_ _ [<- [-> T]]; do !split=>//.
  by move=>t; case=>//_ _ t1 t2 [<- <-] _ H2; elim: (T t1).
move=>t1 _ [<- H1].
apply: bnd_do=>//; last first.
- move=>_ _ [<- [-> T]]; do ![split; auto]. 
  by move=>t; case=>//_ _ _ t2 [<- <-] _ _ H3; elim: (T t2).
move=>t2 _ [<- H2]; eauto. 
Qed. 
Next Obligation. 
case: H=><-; rewrite -(unh0 empty).
apply: bnd_do=>//; last first.
- move=>_ _ [<- [-> T]]; do !split=>//.
  by move=>t; case=>//_ s [<-]; move/T.
move=>t _ [<- H1]; apply: cons0=>//; first by case: t H1; rewrite unh0.
case: t H1=>[H1|t1 t2 H1|t1 t2 H1] y m; last 2 [by move=>[-> ->]; eauto]; 
move=>[-> ->]; do !split=>//; 
move=>t; case=>//_ s [<-]; 
by move/(unicity H1).
Qed.
Next Obligation. 
case: H=><-; rewrite -(unh0 empty). 
apply: bnd_do=>//; last first.
- move=>_ _ [<- [-> T]]; do !split=>//.
  by move=>t; case=>//_ s [<-]; move/T.
move=>t _ [<- H1]; apply: cons0=>//; first by case: t H1; rewrite unh0.
case: t H1=>[H1|t1 t2 H1|t1 t2 H1] y m; last 2 [by move=>[-> ->]; eauto]; 
move=>[-> ->]; do !split=>//; 
move=>t; case=>//_ s [<-]; 
by move/(unicity H1).
Qed. 
Next Obligation.
case: H=><-; rewrite -(unh0 empty).
apply: bnd_do=>//; last first.
- move=>_ _ [<- [-> T]]; do !split=>//.
  by move=>t; case=>//_ _ r [<- <-] _; move/T.
move=>t _ [<- H1]; apply: cons0=>//; first by rewrite unh0.
case=>_ _ [-> //= [->]]; eauto.
Qed.
Next Obligation.
case: H=><-; rewrite -(unh0 empty).
apply: bnd_do=>//; last first.
- move=>_ _ [<- [-> T]]; do !split=>//.
  by move=>t1; case=>//_ _ s [<- <-]; move/T.
move=>t1 _ [<- H1]; apply: bnd_do=>//; last first.
- move=>_ _ [<- [-> T]]; do !split=>//.
  by move=>t2; case=>//_ _ s [<- <-] _; move/T.
move=>t2 _ [<- H2]; apply: cons0=>//; rewrite unh0.
- case: t1 H1=>//= t t1 H1.
  by case: eqP.
case: t1 H1=>//= [H1|t t1 H1|t t1 H1] y m;
try by move=>[-> ->]; do !split=>//; 
       move=>t3; case=>//_ _ s [<- <-]; move/(unicity H1).
case: eqP=>//= H3 [-> ->]; do !split=>//. 
- by rewrite H3 in H1; eauto.
move=>t3; case=>//_ _ s [<- <-].
by move/(unicity H1)=>[<- _]; move/(unicity H2); auto.
Qed.

(* Main function. Some currying is necessary to get the arguments *)
(* intro right places.                                            *)

Definition infer (G : list tp) (e : exp) : spec_of_infer G e := 
  Fix (fun infer x => 
             pre_infer (fst x) (snd x) (fun G e => infer (G, e)))
  (G,e).




