(*****************************************)
(* Signatures for the monadic operations *)
(* in the small footprint style          *)
(*****************************************)

Require Import ssreflect eqtype ssrbool ssrnat ssrfun seq.
Require Import heaps stmodel.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Delimit Scope stsep_scope with stsep. 
Open Scope stsep_scope.

(***********************************************)
(* Definitions of separation logic connectives *)
(* and associated lemmas                       *)
(* In ssreflect                                *)
(***********************************************)

Section STsep.
 
Inductive star1 (p1 p2 : heap -> Prop) (h : heap) : Prop :=
| starI h1 h2 of def (h1 :+ h2) & h1 :+ h2 = h & p1 h1 & p2 h2.
Implicit Arguments starI [p1 p2 h].

Inductive star2 (q : heap -> heap -> Prop) (m i h : heap) : Prop :=
| star2I i1 h1 of 
    i = i1 :+ m & h = h1 :+ m & q i1 h1 & def (i1 :+ m) & def (h1 :+ m).
Implicit Arguments star2I [q i h].

Inductive wand (p1 p2 : heap -> Prop) (h : heap) : Prop :=
| wandI of forall h1, def (h1 :+ h) -> p1 h -> p2 (h1 :+ h). 

Definition seteq (p1 p2 : (heap -> Prop) -> Prop) : Prop := 
  forall f, p1 f <-> p2 f. 
 
Inductive diff2 (p : heap -> Prop) (q : heap -> heap -> Prop) (i h : heap) : Prop :=
| diff2I of (forall i1 m, def (i1 :+ m) -> p i1 -> i1 :+ m = i -> 
              exists h1, h = h1 :+ m /\ q i1 h1 /\ def (h1 :+ m)).

Inductive pts (A : Type) (x : loc) (v : A) (h : heap) : Prop :=
| ptsI of x :-> v = h /\ def h.

Inductive emp (i : heap) : Prop := empI of empty = i.
Inductive this (i h : heap) : Prop := thisI of h = i. 
Inductive top (h : heap) : Prop := topI.

Hint Constructors star1 star2 wand diff2 pts emp this top.

Notation "p1 '#' p2"   := (star1 p1 p2) (at level 50) : sep_scope.
Notation "p1 '--#' p2" := (wand p1 p2) (at level 80) : sep_scope.
Notation "q '##' h"  := (star2 q h) (at level 50) : sep_scope.
Notation "l '-->' v"   := (pts l v) (at level 50) : sep_scope.
Notation "p '--o' q"   := (diff2 p q) (at level 75) : sep_scope.

Open Local Scope sep_scope.

(* this is the antiframe lemma *)   
Lemma diff2_frame (h i1 m1 : heap) (p : heap -> Prop) (q : heap -> heap -> Prop) : 
        def (i1 :+ h) -> (p --o q) (i1 :+ h) (m1 :+ h) -> (p --o q) i1 m1.
Proof.
move=>h i1 m1 p q H1; case=>T.
apply: diff2I=>i2 m _ H2 H3; rewrite -H3 in H1 T * => {H3 i1}.
rewrite -unA in H1 T.
move: (T _ _ H1 H2 (refl_equal _))=>{T} [h2 [H3 [H4 H5]]].
rewrite unA in H3 H5.
move: (unKhr H5 H3)=>->.
exists h2; do !split=>//. 
by apply: unDl H5.
Qed.

(* this is the lemma stating that p --o q is local, in the *)
(* calcagno, ohearn, yang sense *)
(* this must be true, as p --o q is the best local action for p and q *)
Lemma diff2_frame_ex (m h i1 : heap) (p : heap -> Prop) (q : heap -> heap -> Prop) : 
        def (i1 :+ h) -> (p --o q) (i1 :+ h) m -> (p # top) i1 -> 
          exists m1, def (m1 :+ h) /\ m1 :+ h = m /\ (p --o q) i1 m1.
Proof.
move=>m h i1 p q H1 T.
case: (T)=>T1 [h1 h2 _ H3 H2 _]; rewrite -H3 in H1 T T1 * => {i1 H3}.
rewrite -unA in H1 T1.
move: (T1 _ _ H1 H2 (refl_equal _))=>[h3 [H3 [H4 H5]]]; rewrite H3 in T * => {H3}.
exists (h3 :+ h2).
rewrite -unA; do 2 split=>//.
by rewrite !unA in H1 T; apply: diff2_frame T.
Qed.

(* this is the lemma stating that the rule of frame holds *)
(* for p --o q *)
Lemma diff2_frame2 (i j : heap) (p : heap -> Prop) (q : heap -> heap -> Prop) (r : heap -> Prop) : 
        p i -> 
        (forall h, (p --o q) i h -> r h) -> 
        (forall h, def (i :+ j) -> (p --o q) (i :+ j) h -> 
                    exists h', h = j :+ h' /\ def h /\ r h').  
Proof.
move=>i j p q r H3 H1 h D1 H2.
have H: (p # top) i by apply: (starI i empty); rewrite ?unh0 //; apply: (unDl D1).
move: (diff2_frame_ex D1 H2 H)=>[h1][E][Dh] H4.
exists h1; rewrite unC; do !split=>//; first by rewrite -Dh .
by apply: H1.
Qed.

Lemma frame_pre (h1 h2 : heap) (p1 : heap -> Prop) : 
        def (h1 :+ h2) -> (p1 # top) h1 -> (p1 # top) (h1 :+ h2). 
Proof.
move=>h1 h2 p1 H [i1 i2] H1 E H2 _; rewrite -E in H *=>{h1 E}.
rewrite -unA in H.
apply: (starI _ _ H)=>//.
by rewrite unA.
Qed. 

Lemma frame_weaken (p q : heap -> Prop) (h : heap) : (p # q) h -> (p # top) h.
Proof. by move=>p q h [h1 h2 D <- H1 H2]; apply: (starI _ _ D). Qed.

Lemma frame_lolli (p : heap -> Prop) (q : heap -> heap -> Prop) (h i1 i2 : heap) : 
       (p # this h) i1 -> (p --o q) i1 i2 -> (q ## h) i1 i2.
Proof.
move=>p q _ i1 i2 [h1 h2 H1 <- H2 [<-]] [].
move/(_ h1 h2 H1 H2 (refl_equal _))=>[h3 [-> [H4 H3]]]. 
by apply: star2I.
Qed.

(********************)
(* Separation monad *)
(********************)

Definition spec (A : Type) := (pre * post A)%type.

Definition STsep (A : Type) (s : spec A) : Type :=
  st.ST (s.1 # top) A (fun x => s.1 --o s.2 x).

Definition fr (A : Type) (s : spec A) : spec A :=
  (s.1 # top, fun x => s.1 --o s.2 x).

Section MonadicUnit.
Variable A : Type.
 
Definition ret_pre : pre := emp.
Definition ret_post x : post A := fun y i m => m = i /\ y = Val x.
Definition ret_s x : spec A := (ret_pre, ret_post x). 

Program
Definition Ret x : STsep (ret_s x) := st.do (st.ret x) _. 
Next Obligation.
case: H=>h2 h1 H1 <- [H2] _; rewrite -H2 un0h in H1 *=>{i h2 H2}.
split=>//; case=>[a m [-> [<-]]|e m []] //.
apply: diff2I=>i1 m1 H2 [H3] H4. 
rewrite -H3 in H2 H4 *; rewrite -H4 un0h in H1 *=>{i1 h1 H2 H3 H4}.
by exists empty; rewrite un0h.
Qed.
End MonadicUnit. 

Section MonadicBind. 
Variables A B : Type. 

Definition bind_pre (s1 : spec A) (s2 : A -> spec B) : pre :=
  fun i => (s1.1 # top) i /\ forall x h, (s1.1 --o s1.2 (Val x)) i h -> 
           ((s2 x).1 # top) h.
Definition bind_post (s1 : spec A) (s2 : A -> spec B) : post B :=
  fun y i m => 
    (exists x, exists h, (s1.1 --o s1.2 (Val x)) i h /\ 
                         ((s2 x).1 --o (s2 x).2 y) h m) \/
    (exists e, y = Exn e /\ (s1.1 --o s1.2 (Exn e)) i m).
Definition bind_s (s1 : spec A) (s2 : A -> spec B) : spec B := 
  (bind_pre s1 s2, bind_post s1 s2). 

Program 
Definition Bind (s1 : spec A) (s2 : A -> spec B) 
                (c1 : STsep s1) (c2 : forall x, STsep (s2 x)) 
                : STsep (bind_s s1 s2) :=
           st.do (st.bind c1 c2) _. 
Next Obligation.
case: H=>h1 h2 D1 <- {i} [P1 Q1] _; split.
- split=>[|x m]; first by apply: frame_pre P1.
  move/diff2_frame_ex=>T; move: {T} (T D1 P1)=>[m1 [D2 [<- Q2]]] {m}.
  by apply: frame_pre=>//; apply: Q1.
move=>y m [[x [h [MAIN1 MAIN2]]]|].
- move/diff2_frame_ex: MAIN1 (MAIN1) MAIN2=>T; 
    move: (T D1 P1)=>[h3 [D2 [<- Q2]]] MAIN1 MAIN2 {h T D1 P1}.
  move: (Q1 _ _ Q2)=>P2 {Q1 Q2}.
  move/diff2_frame_ex: MAIN2 (MAIN2)=>T; 
    move: (T D2 P2)=>[h4 [D3 [<- Q3]]] MAIN2 {m T D2 P2}.
  apply: diff2I=>i1 i2 D1 [P1 Q1] E1; rewrite -E1 in MAIN1.
  move/diff2_frame_ex: MAIN1 (MAIN1) MAIN2=>T; 
    move: (T D1 P1)=>[i3 [D2 [<- Q2]]] MAIN1 MAIN2 {T E1 P1}.
  move: (Q1 _ _ Q2)=>P2 {Q1 Q2}.
  move/diff2_frame_ex: MAIN2 (MAIN2)=>T; 
    move: {D3 Q3} (T D2 P2)=>[i4 [D3 [<- Q3]]] MAIN2 {T D2 P2}.
  exists i4; do 2 split=>//.
  left; exists x; exists i3; split=>//.
  by apply: diff2_frame MAIN1.
case: y=>[_ [_ [//=]]|e [_ [[<-] MAIN1]] {D1 P1 Q1}].
apply: diff2I=>i1 i2 D1 [P1 Q1] E1; rewrite -E1 in MAIN1.
move/diff2_frame_ex: MAIN1 (MAIN1)=>T; 
  move: (T D1 P1)=>[i3 [D2 [<- Q2]]] MAIN1 {T E1 P1}.
exists i3; do 2 split=>//.
by right; exists e.
Qed.
End MonadicBind.

Section ConsequenceFrame.
Variable A : Type. 

Definition verify (i : heap) (s : spec A) (q : ans A -> heap -> Prop) : Prop :=
  s.1 i /\ forall y m, s.2 y i m -> q y m.

Lemma frame (s : spec A) i j (r : ans A -> heap -> Prop) : 
        (def i -> verify i (fr s) (fun y m => def (m :+ j) -> r y (m :+ j)))
         -> def (i :+ j) -> verify (i :+ j) (fr s) r.
Proof.
move=>s i j r H1 D1; move: {H1} (H1 (unDl D1))=>[H1] H2.
case: (H1)=>h1 h2 D Eh H3 _; rewrite -{i}Eh /= in H1 D1 H2 *. 
split=>[|y m H]; first by apply: (starI h1 (h2 :+ j)); rewrite ?unA.
case: (H); move/(_ h1 (h2 :+ j)); rewrite unA; move/(_ D1 H3 (refl_equal _)).
move=>[h3][Em][H4] D4; rewrite {m}Em in H *; rewrite unA in D4 *.
apply: H2=>//; rewrite unA /= in H. 
move: (diff2_frame_ex D1 H H1)=>[m1][Dm1][].
by heap_cancel=>// _ ->.
Qed.

Lemma frame0 (s : spec A) i r : 
        verify i s r -> 
         def i -> verify i (fr s) r.
Proof.
move=>s i r [H1] H2 D1.
split=>[|y m]; first by apply: (starI i empty); rewrite ?unh0.
case.
move/(_ i empty); rewrite unh0; move/(_ D1 H1 (refl_equal _))=>[h1][Em][H3] _.
rewrite Em unh0.
by apply: H2.
Qed.

Lemma frame1 (s : spec A) i (r : ans A -> heap -> Prop) :
        verify empty s (fun y m => def (m :+ i) -> r y (m :+ i)) -> 
          def i -> verify i (fr s) r.
Proof.
move=>s i r H.
rewrite -(un0h i).
apply: frame=> _.
by apply: frame0.
Qed.

Lemma verif_diff s i (p : heap -> Prop) (q : ans A -> heap -> heap -> Prop) : 
           verify i s (fun y m => forall i1 h, i = i1 :+ h -> def i -> p i1 -> 
                                   exists m1, m = m1 :+ h /\ def m /\ q y i1 m1) -> 
           verify i s (fun y m => (p --o q y) i m).
Proof.
move=>s i p q [H1 H2].
split=>// y m; move/H2=>H3.
split=>i1 m1 D1 H4 H5.
move: (H3 i1 m1); rewrite -H5; move/(_ (refl_equal _) D1 H4).
move=>[m2][Em][Dm'] Hm.
by exists m2; rewrite -Em.
Qed.

Lemma pull_verif (i : heap) (s : spec A) (p : heap -> Prop) (q : ans A -> heap -> heap -> Prop) : 
         (forall i1 h, i = i1 :+ h -> def i -> p i1 -> 
            verify i s (fun y m => exists m1, m = m1 :+ h /\ def m /\ q y i1 m1)) -> 
         (exists i1, exists i2, i = i1 :+ i2 /\ def i /\ p i1) -> 
         verify i s (fun y m => forall i1 h, i = i1 :+ h -> def i -> p i1 -> 
                                  exists m1, m = m1 :+ h /\ def m /\ q y i1 m1).
Proof.
move=>i s p q H1 [i1][i2][E][D]H2.
move: (H1 i1 i2 E D H2)=>[H3 H4].
split=>//.
move=>y m H5. 
move: (H4 _ _ H5)=>[h6][H7][_] H8. 
move=>// i3 h H9 D6 H10.
move: (H1 i3 h H9 D6 H10)=>[H11].
move/(_ _ _ H5)=>[h12][H12][H13] H14. 
by exists h12.
Qed.

Program 
Definition Do (s1 s2 : spec A) 
              (c : STsep s1 | forall i, s2.1 i -> 
                                verify i (fr s1) (fun y m => s2.2 y i m)) : STsep s2 :=
           st.do c _. 
Next Obligation.
suff H1: verify i (fr s1) (fun y m => (fr s2).2 y i m) by []. 
apply: verif_diff.
apply: pull_verif; last first.
- case: H=>h1 h2 Dh Eh Hh _.
  by exists h1; exists h2; rewrite -Eh.
move=>i1 h E1 D1 H1; rewrite E1 in D1 *.
apply: frame=>// D2.
move: (H0 _ H1)=>[H3] H4.
split=>// y m; move/H4=>H5 Dm.
by exists m.
Qed. 
End ConsequenceFrame. 

Section Read.
Variable A : Type. 

Definition read_pre (x : loc) : pre := fun i => exists v : A, (x --> v) i.
Definition read_post (x : loc) : post A :=
  fun y i m => i = m /\ forall v : A, (x --> v) i -> y = Val v.
Definition read_s x : spec A := (read_pre x, read_post x).

Program 
Definition Read (x : loc) : STsep (read_s x) := st.do (st.read x) _.
Next Obligation.
case: H=>i1 i2 D1 <- {i} [v [[E _]]] _; rewrite -E in D1 * => {E i1}.
move: (D1); rewrite defPUn; move/andP=>[D1' D2].
split=>[|y m [-> H]].
- rewrite /st.read_pre validPUn // D1' D2; split=>//.
  by exists v; rewrite lookPUn.
apply: diff2I=>h1 h2 D [v1 [[H1 _]]]; rewrite -H1 in D *=>{h1 H1} H1.
move: (pts_injSl D H1)=>[E1 E2]; subst.
exists (x :-> v); do 3![split=>//]=>v1 [[H2] H3].
apply: H; rewrite lookPUn //; rewrite defPt in H3.
by move: (pts_inj H3 H2)=>->.
Qed.
End Read.
 
Implicit Arguments Read [A].

Section Alloc.
Variable A : Type.
  
Definition alloc_pre : pre := emp.
Definition alloc_post (v : A) : post loc :=
  fun y i m => exists x, y = Val x /\ (x --> v) m.
Definition alloc_s v : spec loc := (alloc_pre, alloc_post v). 

Program Definition Alloc (v : A) : STsep (alloc_s v) := st.do (st.alloc v) _. 
Next Obligation.
case: H=>h1 h2 D1 <- {i} [E]; rewrite -E un0h in D1 *=>{E h1} _.
split=>// x m [y][H][->{x}][H1][<-{m}].
split=>i1 i2 D2 [E]; subst.
rewrite un0h in D2 *=>E; subst i2.
exists (y :-> v).
rewrite {1}unUh // un0h /alloc_post defPUn D1 H1 ?H; do !split=>//.
by exists y; split=>//; apply: ptsI; rewrite defPt. 
Qed.
End Alloc.

Section BlockAlloc.  
Variable A : Type. 

Definition allocb_pre : pre := emp.
Definition allocb_post (v : A) (n : nat) : post loc :=
  fun y i m => exists x, y = Val x /\ m = updi x (seqn n v) /\ def m.
Definition allocb_s v n : spec loc := (allocb_pre, allocb_post v n).

Program
Definition Allocb (v : A) (n : nat) : STsep (allocb_s v n) :=
           st.do (st.allocb v n) _.
Next Obligation.
case: H=>h1 h2 D1 <- {i} [E]; rewrite -E un0h in D1 *=>{E h1} _.
split=>// x m [y [-> {x} [H1 H2]]].
split=>i1 i2 D2 [E]; subst.
rewrite un0h in D2 *=>E; subst i2.
exists (updi y (seqn n v)); do !split; try by rewrite unC.
by exists y; do !split=>//; apply: unDr H2. 
Qed.
End BlockAlloc.

Section Deallocation.
Definition dealloc_pre (x : loc) : pre :=
  fun i => exists A, exists v:A, (x --> v) i. 
Definition dealloc_post (x : loc) : post unit :=
  fun y i m => y = Val tt /\ emp m.
Definition dealloc_s x := (dealloc_pre x, dealloc_post x).

Program 
Definition Dealloc (x : loc) : STsep (dealloc_s x) := st.do (st.dealloc x) _. 
Next Obligation.
case: H=>h1 h2 D1 <- [A [v [[E _]]]] _; subst; split. 
- suff : valid x (x :-> v :+ h2) by [].
  move: D1; rewrite defPUn; move/andP=>[D1 D2].
  by rewrite validPUn D1.
move=>y m [->{y}][<-{m}]; split=>i1 i2 D [B][v1][[<- _]] H.
move: D1; rewrite defPUn; case/and3P=>D1 D2 D3.
have H2: def (x :-> v :+ h2) by rewrite defPUn D1 D2 D3. 
move: (pts_injSTr H2 H)=>E; subst.
move: (pts_injSr H2 H)=>[E1 E2]; subst.
exists empty.
by rewrite un0h -unFh // freeUE // free0E un0h.
Qed.
End Deallocation.

Section Update. 
Variable A : Type.
 
Definition write_pre (x : loc) : pre :=
  fun i => exists b, exists y : b, (x --> y) i.
Definition write_post (x : loc) (v : A) : post unit :=
  fun y i m => (x --> v) m /\ y = Val tt.
Definition write_s x v := (write_pre x, write_post x v). 

Program 
Definition Write (x : loc) (v : A) : STsep (write_s x v) :=
           st.do (st.write x v) _. 
Next Obligation.
case: H=>h1 h2 D <- {i} [B [w [[E _]]]] _; subst; split.
- suff : valid x (x :-> w :+ h2) by [].
  move: D; rewrite defPUn; move/andP=>[D1 D2].
  by rewrite validPUn D1.
move=>y m [-> {y} <- {m}]; split; move=>i1 i2 _ [C [v1 [[<- _]]]] H.
move: D; rewrite defPUn; case/and3P=>D1 D2 D3.
have H2: def (x :-> w :+ h2) by rewrite defPUn D1 D2 D3.  
move: (pts_injSTr H2 H)=>E; subst.
move: (pts_injSr H2 H)=>[E1 E2]; subst.
exists (x :-> v).
rewrite defPUn D1 D2 -unUh //= updUE; do !split=>//.
by rewrite defPt.
Qed.
End Update. 

Section Throw.
Variable A : Type.
 
Definition throw_pre : pre := emp.
Definition throw_post (e : exn) : post A :=
  fun y i m => m = i /\ y = Exn e.
Definition throw_s e := (throw_pre, throw_post e).

Program 
Definition Throw (e : exn) : STsep (throw_s e) :=
           st.do (st.throw A e) _.
Next Obligation.
case: H=>h1 h2 D <- {i} [E] _; subst; rewrite un0h in D *; split=>//.
move=>x m [-> {m} -> {x}].
split=>i1 i2 _ [E1 E2]; subst; rewrite un0h in D *.
by exists empty; rewrite un0h.
Qed.
End Throw.

Section Try.
Variables (A B : Type) (s : spec A) (s1 : A -> spec B) (s2 : exn -> spec B).
 
Definition try_pre : pre :=
  fun i => (s.1 # top) i /\ 
           (forall x h, (s.1 --o s.2 (Val x)) i h -> ((s1 x).1 # top) h) /\ 
           (forall e h, (s.1 --o s.2 (Exn e)) i h -> ((s2 e).1 # top) h).
Definition try_post : post B :=
  fun y i m => (exists x, exists h, (s.1 --o s.2 (Val x)) i h /\ 
                                    ((s1 x).1 --o (s1 x).2 y) h m) \/
               (exists e, exists h, (s.1 --o s.2 (Exn e)) i h /\ 
                                    ((s2 e).1 --o (s2 e).2 y) h m).
Definition try_s := (try_pre, try_post).
   
Program 
Definition Try (c : STsep s) 
               (c1 : forall x, STsep (s1 x))
               (c2 : forall e, STsep (s2 e)) : STsep try_s :=
           st.do (st.try c c1 c2) _. 
Next Obligation.
case: H=>h1 h2 D1 <- {i} [P1] [Q1 Q1'] _; split.
- split; first by apply: frame_pre P1.
  split=>[x m|e m]; move/(diff2_frame_ex D1); 
  move/(_ P1)=>[h3 [D2 [<- Q2]]].
  - by move: (Q1 _ _ Q2)=>P3; apply: frame_pre P3.
  by move: (Q1' _ _ Q2)=>P3; apply: frame_pre P3.
move=>y m [[x [h [MAIN1 MAIN2]]]|[e [h [MAIN1 MAIN2]]]].
- move/diff2_frame_ex: MAIN1 (MAIN1) MAIN2=>T; 
    move: (T D1 P1)=>[h3 [D2 [<- Q2]]] MAIN1 MAIN2 {h T D1 P1}.
  move: (Q1 _ _ Q2)=>P2 {Q1 Q2}.
  move/diff2_frame_ex: MAIN2 (MAIN2)=>T; 
    move: (T D2 P2)=>[h4 [D3 [<- Q3]]] MAIN2 {m T D2 P2}.
  apply: diff2I=>i1 i2 D1 [P1 [Q1 _]] E1; rewrite -E1 in MAIN1.
  move/diff2_frame_ex: MAIN1 (MAIN1) MAIN2=>T; 
    move: (T D1 P1)=>[i3 [D2 [<- Q2]]] MAIN1 MAIN2 {T E1 P1}.
  move: (Q1 _ _ Q2)=>P2 {Q1 Q2}.
  move/diff2_frame_ex: MAIN2 (MAIN2)=>T; 
    move: {D3 Q3} (T D2 P2)=>[i4 [D3 [<- Q3]]] MAIN2 {T D2 P2}.
  exists i4; do 2 split=>//.
  left; exists x; exists i3; split=>//.
  by apply: diff2_frame MAIN1.
move/diff2_frame_ex: MAIN1 (MAIN1) MAIN2=>T; 
  move: (T D1 P1)=>[h3 [D2 [<- Q2]]] MAIN1 MAIN2 {h T D1 P1}.
move: (Q1' _ _ Q2)=>P2 {Q1' Q1 Q2}.
move/diff2_frame_ex: MAIN2 (MAIN2)=>T; 
  move: (T D2 P2)=>[h4 [D3 [<- Q3]]] MAIN2 {m T D2 P2}.
apply: diff2I=>i1 i2 D1 [P1 [_ Q1']] E1; rewrite -E1 in MAIN1.
move/diff2_frame_ex: MAIN1 (MAIN1) MAIN2=>T; 
  move: (T D1 P1)=>[i3 [D2 [<- Q2]]] MAIN1 MAIN2 {T E1 P1}.
move: (Q1' _ _ Q2)=>P2 {Q1' Q2}.
move/diff2_frame_ex: MAIN2 (MAIN2)=>T; 
  move: {D3 Q3} (T D2 P2)=>[i4 [D3 [<- Q3]]] MAIN2 {T P2}.
exists i4; do 2 split=>//.
right; exists e; exists i3; split.
- by apply: diff2_frame MAIN1.
by apply: diff2_frame MAIN2.
Qed.
End Try.

Section Recursion.
Variables (A : Type) (B : A -> Type) (s : forall x, spec (B x)).

Definition Fix (f : (forall x, STsep (s x)) -> (forall x, STsep (s x))) : 
                forall x, STsep (s x) := st.ffix f.
End Recursion.

(* Conditionals for various types *)


Section CondBool.
Variables (A : Type) (b : bool) (s1 s2 : spec A). 

Program
Definition If (c1 : STsep s1) (c2 : STsep s2) : STsep (if b then s1 else s2) :=
  match b with 
    true => st.do c1 _ 
  | false => st.do c2 _ 
  end.
End CondBool.

Section CondOption.
Variable (A B : Type) (x : option A) (s1 : spec B) (s2 : A -> spec B).

Program
Definition Match_opt (c1 : STsep s1) (c2 : forall v, STsep (s2 v)) :
           STsep (match x with Some v => s2 v | None => s1 end) :=
  match x with Some v => st.do (c2 v) _ | None => st.do c1 _ end.
End CondOption. 

Section CondDecide.
Variable (A : Type) (p1 p2 : Prop) (b : {p1} + {p2}) 
         (s1 : p1 -> spec A) (s2 : p2 -> spec A).

Program
Definition Match_dec (c1 : forall x, STsep (s1 x))
                     (c2 : forall x, STsep (s2 x)) : 
                 STsep (match b with left x => s1 x | right x => s2 x end) :=
  match b with left x => st.do (c1 x) _ | right x => st.do (c2 x) _ end.
End CondDecide. 

Section CondNat.
Variable (A : Type) (n : nat) (s1 : spec A) (s2 : nat -> spec A).

Program 
Definition Match_nat (c1 : STsep s1) (c2 : forall n, STsep (s2 n)) : 
             STsep (match n with 0 => s1 | m.+1 => s2 m end) :=
  match n with 0 => st.do c1 _ | m.+1 => st.do (c2 m) _ end.
End CondNat.

Section CondSeq.
Variable (A B : Type) (s : seq A) (s1 : spec B) (s2 : A -> seq A -> spec B).

Program
Definition Match_seq (c1 : STsep s1) (c2 : forall hd tl, STsep (s2 hd tl)) :
             STsep (match s with [::] => s1 | hd::tl => s2 hd tl end) :=
  match s with [::] => st.do c1 _ | hd::tl => st.do (c2 hd tl) _ end.
End CondSeq. 
End STsep.

Implicit Arguments starI [p1 p2 h].
Implicit Arguments star2I [q i h].
Implicit Arguments STsep [].
Implicit Arguments Read [A].

Prenex Implicits Do Read. 

(* for some reason, exist is not declared with correct implicits *)
(* which results in awful display *)

Implicit Arguments exist [A P].
Prenex Implicits exist.

Notation "x '<--' c1 ';' c2" := (Bind c1 (fun x => c2)) 
  (at level 78, right associativity) : stsep_scope.
Notation "c1 ';;' c2" := (Bind c1 (fun _ => c2)) 
  (at level 78, right associativity) : stsep_scope.
Notation "'!' x" := (Read x) (at level 50) : stsep_scope.
Notation "e1 '::=' e2" := (Write e1 e2) (at level 60) : stsep_scope.
Notation "'ret' E" := (Ret E) (at level 70, no associativity) : stsep_scope.
Notation "'alloc' E" := (Alloc E) (at level 70, no associativity) : stsep_scope.
Notation "'dealloc' E" := (Dealloc E) (at level 70, no associativity) : stsep_scope.
Notation "'throw' [ t ] E" := (Throw t E) (at level 70, no associativity) : stsep_scope.
Notation "'ttry' E 'then' [ r ] E1 'else' [ x ] E2" := (Try E (fun r => E1) (fun x => E2)) 
  (at level 80) : stsep_scope.
Notation "'ttry' E 'then' [ r ] E1 'else' E2" := (Try E (fun r => E1) (fun _ => E2)) 
  (at level 80) : stsep_scope.
Notation "'ttry' E 'then' E1 'else' [ x ] E2" := (Try E (fun _ => E1) (fun x => E2)) 
  (at level 80) : stsep_scope.
Notation "'ttry' E 'then' E1 'else' E2" := (Try E (fun _ => E1) (fun _ => E2)) 
  (at level 80) : stsep_scope.
Notation "'match_opt' E 'then' E1 'else' [ x ] E2" := (Match_opt E E1 (fun x => E2)) 
  (at level 80) : stsep_scope.
Notation "'match_opt' E 'then' E1 'else' [ x ] E2" := (Match_opt E E1 (fun x => E2)) 
  (at level 80) : stsep_scope.
Notation "'If' E 'then' E1 'else' E2" := (If E E1 E2) (at level 80) : stsep_scope.

(* divergence *)
Definition diverge (A : Type) : STsep A (emp, fun y i m => False).
Proof. move=>A; move: tt; apply: Fix=>f; apply: f. Defined.




