(******************) 
(* Key-Value maps *)
(******************)

Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq fintype tuple finfun. 
Require Import finset rels heaps stmodel stsep hprop vcgen array fmap ordtype.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(********************)
(* stateful KV maps *)
(********************)

Module KVmap. 
Record Sig (K : ordType) (V : Type) : Type := 
  make {tp :> Type; 
        default : tp; 
        shape : tp -> {fmap K -> V} -> Rel heap; 
        shapeD : forall x s, proper (shape x s);
        shape_invert : forall x, invertible (shape x);
      
        new : STsep tp (fun i => emp i, fun y i m => exists x, y = Val x /\ 
                                                     m \In shape x (nil K V));

        free : forall x, STsep unit (fun i => exists s, i \In shape x s, 
                                     fun y i m => y = Val tt /\ m = empty); 

        insert : forall x k v, 
                   STsep unit (fun i => exists s, i \In shape x s, 
                               fun y i m => forall s, i \In shape x s -> 
                                            m \In shape x (ins k v s) /\ y = Val tt);

        remove : forall x k, 
                   STsep unit (fun i => exists s, i \In shape x s, 
                               fun y i m => forall s, i \In shape x s -> 
                                            m \In shape x (rem k s) /\ y = Val tt); 

        lookup : forall x k, 
                   STsep (option V) (fun i => exists s, i \In shape x s, 
                                     fun y i m => forall s, i \In shape x s ->
                                       m \In shape x s /\ y = Val (fnd k s))}.
End KVmap.  

(*******************************************************************)
(* KVmap implemented as a pointer to a finitely-supported function *)
(*******************************************************************)

Module AssocList. Section AssocList.
Variable (K : ordType) (V : Type).
Notation fmap := (fmap K V).
Notation nil := (nil K V).
Definition kvmap := loc.

Definition shape (x : kvmap) (s : fmap) := x --> s.

Lemma shapeD x s : proper (shape x s).
Proof. by move=>x s h; case/swp=>->. Qed.

Lemma shape_invert x (s1 s2 : fmap) h1 h2 :
        h1 \In shape x s1 -> h2 \In shape x s2 -> agree h1 h2 -> s1 = s2 /\ h1 = h2.
Proof.
by move=>?????; rewrite /shape !opn=>[[_]]<-[_]<-[]?[?][D]; heap_cancel=>// _ _ ->.
Qed.

Program Definition new : STsep kvmap (emp, fun y i m => exists x, y = Val x /\ 
                                                        m \In shape x nil) :=
  Do (Alloc nil).
Next Obligation. by apply: val_alloc0=>// x; exists x; split=>//; apply/swp. Qed. 

Program Definition free x : STsep unit (fun i => exists s, i \In shape x s, 
                                        fun y i m => y = Val tt /\ m = empty) :=
  Do (Dealloc x).
Next Obligation. by move: H H0=>s1; case/swp=>->; apply: val_dealloc0. Qed. 

Program Definition insert x k v : 
  STsep unit (fun i => exists s, i \In shape x s, 
              fun y i m => forall s, i \In shape x s -> 
                           m \In shape x (ins k v s) /\ y = Val tt) :=
  Do (s <-- !x; x ::= ins k v s).
Next Obligation. 
by apply: (ghost H) H0=>s; case/swp=>->; heval; split=>//; apply/swp.
Qed.

Program Definition remove x k : 
  STsep unit (fun i => exists s, i \In shape x s, 
              fun y i m => forall s, i \In shape x s -> 
                             m \In shape x (rem k s) /\ y = Val tt) :=
  Do (s <-- !x; x ::= @rem K V k s). 
Next Obligation.
by apply: (ghost H) H0=>s; case/swp=>->; heval; split=>//; apply/swp.
Qed.

Program Definition lookup x k : 
  STsep (option V) (fun i => exists s, i \In shape x s, 
                    fun y i m => forall s, i \In shape x s -> 
                                   m \In shape x s /\ y = Val (fnd k s)) :=
  Do (s <-- !x; ret (fnd k s)).
Next Obligation.
by apply: (ghost H) H0=>s; case/swp=>->; heval; split=>//; apply/swp.
Qed.

Definition AssocList := KVmap.make null shapeD shape_invert new free insert remove lookup. 

End AssocList. End AssocList.

    

