(* A small theory of ordered types *)

Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq paths fintype.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Module Ordered. 
Section OrdMixin.

(* The pivot type of the structure *)
Variable T : Type.

Section Ord.
(*
Ordered types are equality types with ordering relations, so we
require two structures.  Another possibility is to take T : eqType,
but that confuses unification. Thus, it's better to keep T naked
*)

Variable eq_mixin : Equality.mixin_of T.
Structure ord_mixin : Type := 
  Mixin {ord : rel T; 
         irr : irreflexive ord;
         trans : transitive ord;
         total : forall x y, [|| ord x y, Equality.op eq_mixin x y | ord y x]}.
End Ord.

Structure class_of := Class eq_mixin of ord_mixin eq_mixin.
Coercion eq_of m := let: Class eqc _ := m in eqc.
Coercion ord_of m := let: Class _ ordc := m return ord_mixin m in ordc.
End OrdMixin.

Structure type : Type := Pack {sort :> Type; _ : class_of sort}.
Definition class cT := let: Pack _ c := cT return class_of cT in c.
Definition unpack K (k : forall T (c : class_of T), K T c) cT := 
  let: Pack T c := cT return K _ (class cT) in k _ c.
Definition repack cT : _ -> type := let k T c p := p c in unpack k cT.
End Ordered.

Notation ordType := Ordered.type.

Coercion eqType_of_ordType (T : ordType) := EqType (Ordered.eq_of (Ordered.class T)). 
Canonical Structure eqType_of_ordType.

Notation "[ 'ordType' 'of' T 'for' C ]" := 
  (@Ordered.repack C (@Ordered.Pack T))
  (at level 0, format "[ 'ordType' 'of' T 'for' C ]") : form_scope.

Notation "[ 'ordType' 'of' T ]" :=
  (Ordered.repack (fun C => @Ordered.Pack T C))
  (at level 0, format "[ 'ordType' 'of' T ]") : form_scope.

Definition OrdMixin T := @Ordered.Mixin _ (@Equality.class T).

Definition OrdType T ord_mix : ordType := 
  Equality.unpack (K := fun _ _ => _)
  (fun T eqm ordm => Ordered.Pack (@Ordered.Class T eqm ordm)) T ord_mix.

Section Lemmas.
Variable T : ordType. 
Definition ord := Ordered.ord (Ordered.ord_of (Ordered.class T)). 
Lemma irr : irreflexive ord. Proof. exact: Ordered.irr. Qed.  
Lemma trans : transitive ord. Proof. exact: Ordered.trans. Qed.
Lemma total : forall x y, [|| ord x y, x == y | ord y x]. 
Proof. exact: Ordered.total. Qed.
Lemma nsym : forall x y, ord x y -> ord y x -> False.
Proof. by move=>x y E1 E2; move: (trans E1 E2); rewrite irr. Qed. 
End Lemmas. 

Prenex Implicits ord.

Section Totality.
Variable K : ordType.  
 
CoInductive total_spec (x y : K) : bool -> bool -> bool -> Type :=
| total_spec_lt of ord x y : total_spec x y true false false
| total_spec_eq of x == y : total_spec x y false true false
| total_spec_gt of ord y x : total_spec x y false false true.

Lemma totalP x y : total_spec x y (ord x y) (x == y) (ord y x).
Proof.
move=>x y; case H1: (x == y).
- by rewrite (eqP H1) irr; apply: total_spec_eq.
case H2: (ord x y); case H3: (ord y x). 
- by elim: (nsym H2 H3). 
- by apply: total_spec_lt H2.
- by apply: total_spec_gt H3.
by move: (total x y); rewrite H1 H2 H3.
Qed.
End Totality. 

Section NatOrd.
Lemma irr_ltn_nat : irreflexive ltn. Proof. by move=>x; rewrite /= ltnn. Qed.
Lemma trans_ltn_nat : transitive ltn. Proof. by apply: ltn_trans. Qed.
Lemma total_ltn_nat : forall x y, [|| x < y, x == y | y < x]. 
Proof. by move=>*; case: ltngtP. Qed.

(* always define mixin first, then type *)
(* if the definition is inlined, canonical structures don't unify correctly *)
Definition nat_ordMixin := OrdMixin irr_ltn_nat trans_ltn_nat total_ltn_nat.
Canonical Structure nat_ordType := OrdType nat_ordMixin.
End NatOrd.

Section ProdOrd.
Variables K T : ordType.

(* lexicographic ordering *)
Definition lex : rel (K * T) := 
  fun x y => if x.1 == y.1 then ord x.2 y.2 else ord x.1 y.1.

Lemma irr_lex : irreflexive lex.
Proof. by move=>x; rewrite /lex eq_refl irr. Qed.

Lemma trans_lex : transitive lex.
Proof.
move=>[x1 x2][y1 y2][z1 z2]; rewrite /lex /=.
case: ifP=>H1; first by rewrite (eqP H1); case: eqP=>// _; apply: trans.
case: ifP=>H2; first by rewrite (eqP H2) in H1 *; rewrite H1.
case: ifP=>H3; last by apply: trans. 
by rewrite (eqP H3)=>R1; move/(nsym R1).
Qed.

Lemma total_lex : forall x y, [|| lex x y, x == y | lex y x].
Proof.
move=>[x1 x2][y1 y2]; rewrite /lex /=.
case: ifP=>H1.
- rewrite (eqP H1) eq_refl -pair_eqE /= eq_refl /=; exact: total.
rewrite (eq_sym y1) -pair_eqE /= H1 /=.
by move: (total x1 y1); rewrite H1.
Qed.

Definition prod_ordMixin := OrdMixin irr_lex trans_lex total_lex.
Canonical Structure prod_ordType := OrdType prod_ordMixin.
End ProdOrd.

Section FinTypeOrd.
Variable T : finType.

Definition ordf : rel T :=
  fun x y => index x (enum T) < index y (enum T). 

Lemma irr_ordf : irreflexive ordf.
Proof. by move=>x; rewrite /ordf ltnn. Qed.

Lemma trans_ordf : transitive ordf.
Proof. by move=>x y z; rewrite /ordf; apply: ltn_trans. Qed.

Lemma total_ordf : forall x y, [|| ordf x y, x == y | ordf y x].
Proof.
move=>x y; rewrite /ordf; case: ltngtP=>//= H; rewrite ?orbT ?orbF //.
have [H1 H2]: x \in enum T /\ y \in enum T by rewrite !mem_enum.
by rewrite -(nth_index x H1) -(nth_index x H2) H eq_refl.
Qed.

Definition fin_ordMixin := OrdMixin irr_ordf trans_ordf total_ordf.
End FinTypeOrd.

(* notation to let us write I_n instead of (ordinal_finType n) *)
Notation "[ 'fin_ordMixin' 'of' T ]" :=
  (fin_ordMixin _ : @Ordered.ord_mixin T _) (at level 0).

Definition ordinal_ordMixin n := [fin_ordMixin of 'I_n].
Canonical Structure ordinal_ordType n := OrdType (ordinal_ordMixin n).





