(* Module for relations *)

Require Import ssreflect ssrfun eqtype Setoid Morphisms.
Set Implicit Arguments.
Unset Strict Implicit. 
Import Prenex Implicits.

Delimit Scope rel_scope with rel.
Open Scope rel_scope.

Structure Rel (T : Type) : Type := RelI {rel_of :> T -> Prop}.

Definition mem (T : Type) (r : Rel T) (x : T) : Prop := nosimpl (r x).

Notation "x \In r" := (mem r x) (at level 70) : rel_scope. 
Notation "x \Notin r" := (~ mem r x) (at level 70) : rel_scope. 
Notation "[ 'Rel' x | E ]" := (RelI (fun x => E)) 
 (at level 0, x ident, format "[ 'Rel' x | E ]") : rel_scope.
Notation "[ 'Rel' x : T | E ]" := (RelI (fun x : T => E)) 
 (at level 0, x ident, only parsing) : rel_scope.
Notation "[ 'Rel' x y | E ]" := (RelI (fun t => let: (x, y) := t in E))
 (at level 0, x ident, y ident, format "[ 'Rel' x y | E ]") : rel_scope.
Notation "[ 'Rel' x y : T | E ]" := (RelI (fun (t : T*T) => let: (x, y) := t in E))
 (at level 0, x ident, y ident, only parsing) : rel_scope. 
Notation "[ 'Rel' x : T1 y : T2 | E ]" := 
  (RelI (fun (t : T1*T2) => let: (x, y) := t in E))
 (at level 0, x ident, y ident, only parsing) : rel_scope.

(* rewrite rule to remove \In from defined relations *)
Lemma in_rel T x (f : T -> Prop) : x \In (RelI f) = f x.
Proof. by []. Qed.

(* rewrite rule to remove \In from constructed or variable relations *)
Lemma in_simpl (T : Type) (r : Rel T) (x : T) : x \In r = r x. 
Proof. by []. Qed.

(* combined together, the rules should completely remove \In from the expression *)
Definition InE := (in_simpl, in_rel). 

Definition subrel (T : Type) (r1 r2 : Rel T) : Prop := 
  forall x, x \In r1 -> x \In r2.

Notation "r1 <=r r2" := (subrel r1 r2) (at level 80) : rel_scope.  

Inductive releq (T : Type) (r1 r2 : Rel T) : Prop := 
  releqI of r1 <=r r2 & r2 <=r r1.

Notation "r1 =r r2" := (releq r1 r2) (at level 80, no associativity) : rel_scope.

Section Subrel.
Variable T : Type. 

Lemma subrel_refl (r : Rel T) : r <=r r. Proof. by move=>r. Qed.

Lemma subrel_asym (r1 r2 : Rel T) : r1 <=r r2 -> r2 <=r r1 -> r1 =r r2.
Proof. by move=>r1 r2 H1 H2; split. Qed.

Lemma subrel_trans' (r1 r2 r3 : Rel T) : r1 <=r r2 -> r2 <=r r3 -> r1 <=r r3.
Proof. by move=>??? H1 H2 x; move/H1; move/H2. Qed.

Lemma subrel_trans (r2 r1 r3 : Rel T) : r1 <=r r2 -> r2 <=r r3 -> r1 <=r r3.
Proof. by move=>??? H1 H2 x; move/H1; move/H2. Qed.   
End Subrel. 
     
Add Parametric Relation T : (Rel T) (@subrel T) 
  reflexivity proved by (@subrel_refl T)
  transitivity proved by (@subrel_trans' T) as subrel_rel. 

Section Eqrel. 
Variable T : Type. 

Lemma releq_refl (r : Rel T) : r =r r. Proof. by move=>r; split. Qed.

Lemma releq_sym (r1 r2 : Rel T) : r1 =r r2 -> r2 =r r1.
Proof. by move=>r1 r2; case=>H1 H2; apply: releqI. Qed.

Lemma releq_trans' (r1 r2 r3 : Rel T) : r1 =r r2 -> r2 =r r3 -> r1 =r r3.
Proof.
move=>r1 r2 r3 [H1 H2][H3 H4]; apply: releqI;
by [apply: subrel_trans H3 | apply: subrel_trans H2].
Qed.

Definition releq_trans r2 r1 r3 := @releq_trans' r1 r2 r3.
End Eqrel. 

Hint Resolve releq_refl releq_sym. 

Add Parametric Relation T : (Rel T) (@releq T)
  reflexivity proved by (@releq_refl T)
  symmetry proved by (@releq_sym T)
  transitivity proved by (@releq_trans' T) as releq_rel. 

Add Parametric Morphism T : (@subrel T) with 
  signature (@releq T) ==> (@releq T) ==> iff as subrel_morph.
Proof.
by move=>r1 s1 [H1 H2] r2 s2 [H3 H4]; split=>H5; 
   [apply: subrel_trans H3 | apply: subrel_trans H4]; apply: subrel_trans H5.
Qed. 

Add Parametric Morphism T : (@mem T) 
  with signature (@releq T) ==> (@eq T) ==> iff as mem_morph.
Proof. by move=>r1 r2 [H1 H2] x; split; [move/H1 | move/H2]. Qed.

(* we have the other direction as well *)
Lemma mem_rel (T : Type) (r1 r2 : Rel T) : r1 =r r2 <-> forall x, x \In r1 <-> x \In r2.
Proof.
move=>T r1 r2; split=>[H x|H]; first by rewrite H.
by split=>x; rewrite H.
Qed.

Section Properties.
Variables (T : Type) (r : Rel (T * T)).
Definition Reflexive := forall x, (x, x) \In r. 
Definition Symmetric := forall x y, (x, y) \In r -> (y, x) \In r.
Definition Transitive := forall y x z, (x, y) \In r -> (y, z) \In r -> (x, z) \In r.
Definition Equivalence := Reflexive /\ Symmetric /\ Transitive.
Definition Antisymmetric := forall x y, (x, y) \In r -> (y, x) \In r -> x = y.
End Properties.

Section Operations.
Variables (T : Type) (r1 r2 : Rel T).
Definition bot := [Rel x : T | False].
Definition top := [Rel x : T | True]. 
Definition join := [Rel x | x \In r1 \/ x \In r2].
Definition meet := [Rel x | x \In r1 /\ x \In r2].
Definition diag := [Rel x y : T | x = y].
Definition prod := [Rel x y : T | x \In r1 /\ y \In r2].
Definition compl := [Rel x | x \Notin r1].
End Operations.

Implicit Arguments bot [T].
Implicit Arguments top [T]. 

Notation "r1 +r r2" := (join r1 r2) (at level 65, right associativity) : rel_scope.
Notation "r1 *r r2" := (meet r1 r2) (at level 63, right associativity) : rel_scope.
Notation "r1 ** r2" := (prod r1 r2) (at level 62, right associativity) : rel_scope.

Add Parametric Morphism T : (@join T) 
  with signature (@releq T) ==> (@releq T) ==> (@releq T) as join_morph.
Proof.
move=>r1 s1 [H1 H2] r2 s2 [H3 H4]; split=>x.
- by case; [move/H1 | move/H3]; rewrite !InE /=; auto.
by case; [move/H2 | move/H4]; rewrite !InE /=; auto.
Qed.

Add Parametric Morphism T : (@meet T) with 
  signature (@releq T) ==> (@releq T) ==> (@releq T) as meet_morph.
Proof.
move=>r1 s1 [H1 H2] r2 s2 [H3 H4]; split=>x [T1 T2];
by [move: (H1 _ T1) (H3 _ T2) | move: (H2 _ T1) (H4 _ T2)].
Qed.

Add Parametric Morphism T : (@prod T) with 
  signature (@releq T) ==> (@releq T) ==> (@releq (T * T)) as prod_morph.
Proof.
move=>r1 s1 [H1 H2] r2 s2 [H3 H4]; split; move=>[x y] [T1 T2];
by [move: (H1 _ T1) (H3 _ T2) | move: (H2 _ T1) (H4 _ T2)].
Qed.

Add Parametric Morphism T : (@compl T) with 
  signature (@releq T) ==> (@releq T) as compl_morph.
Proof. 
by move=>r1 r2 [H1 H2]; split=>x T1 T2; elim: T1; [apply: H2 | apply: H1].
Qed.

Section Operators.
Variables (T : Type) (op : Rel T -> Rel T -> Rel T).

Definition Monotone := forall r1 s1 r2 s2, r1 <=r r2 -> s1 <=r s2 -> op r1 s1 <=r op r2 s2.
Definition Unitop b := (forall r, op b r =r r) /\ (forall r, op r b =r r). 
Definition Associative := forall r1 r2 r3, op (op r1 r2) r3 =r op r1 (op r2 r3). 
End Operators. 
    
Section Laws.
Variable (T : Type).
 
Lemma orrr (r : Rel T) : r +r r =r r.
Proof. by move=>r; split=>x; [case | left]. Qed.

Lemma orrC (r1 r2 : Rel T) : r1 +r r2 =r r2 +r r1.
Proof. move=>r1 r2; split=>x; rewrite !InE /=; tauto. Qed.

Lemma orr0 (r : Rel T) : r +r bot =r r.
Proof. by move=>r; split=>x; [case | left]. Qed.

Lemma or0r (r : Rel T) : bot +r r =r r.
Proof. by move=>r; rewrite orrC orr0. Qed.

Lemma orrCA (r1 r2 r3 : Rel T) : r1 +r r2 +r r3 =r r2 +r r1 +r r3.
Proof. by move=>r1 r2 r3; split=>x; rewrite !in_rel !in_simpl; intuition. Qed. 

Lemma orrAC (r1 r2 r3 : Rel T) : (r1 +r r2) +r r3 =r (r1 +r r3) +r r2.
Proof. by move=>???; split=>x; rewrite !in_rel !in_simpl; intuition. Qed.

Lemma orrA (r1 r2 r3 : Rel T) : (r1 +r r2) +r r3 =r r1 +r r2 +r r3.
Proof. by move=>r1 r2 r3; rewrite (orrC r2) orrCA orrC. Qed.

(* absorption *)
Lemma orrAb (r1 a : Rel T) : r1 =r r1 +r a <-> a <=r r1. 
Proof.
move=>r1 a; split; first by move=>-> x; rewrite in_rel; auto.
move=>H; split=>x; first by rewrite in_rel; auto.
by case=>//; move/H.
Qed.

Lemma sub_orl (r1 r2 : Rel T) : r1 <=r r1 +r r2. Proof. by left. Qed.
Lemma sub_orr (r1 r2 : Rel T) : r2 <=r r1 +r r2. Proof. by right. Qed.   

(* idempotence *) 
Lemma orI (r1 : Rel T) : r1 +r r1 =r r1.
Proof. by move=>r1; split=>x; [case | move=>H; left]. Qed.

(* a helper for rewriting *)
Definition pull (r : Rel T) := (orrC r, orrCA r).

End Laws.
