(* Predicate transformer model for heaps with undefined element *)

Require Import ssreflect ssrbool ssrnat eqtype seq heaps. 
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits. 

(* Exceptions are an equality type *)
Inductive exn : Type := exn_from_nat : nat -> exn.

Definition exn_to_nat (e : exn) : nat :=
  let: exn_from_nat y := e in y.

Definition eqexn (e1 e2 : exn) : bool :=
  match e1, e2 with exn_from_nat m, exn_from_nat n => m == n end.

Lemma eqexnP : Equality.axiom eqexn.
Proof.
move=>[x][y]//=; case: eqP=>[->|*];constructor=>//.
by move=>[*].
Qed.

Canonical Structure exn_eqMixin := EqMixin eqexnP.
Canonical Structure exn_eqType := EqType exn_eqMixin.

(* Answer type *)
Inductive ans (A : Type) : Type := Val of A | Exn of exn.
Implicit Arguments Exn [A].

Definition pre := heap -> Prop.
Definition post (A : Type) := ans A -> heap -> heap -> Prop.

Module st.

Definition hpleq (x y : heap -> Prop) := forall i, x i -> y i.

Notation "x << y" := (hpleq x y) (at level 70).

Lemma hprefl (x : heap -> Prop) : x << x.
Proof. by move=>*. Qed.
Hint Resolve hprefl.

Lemma hptrans (x y z : heap -> Prop) : x << y -> y << z -> x << z.
Proof. by move=>x y z H1 H2 h H3; apply: H2; apply: H1. Qed.
Implicit Arguments hptrans [x z].  

(* ideal P consists of all heap predicates that "satisfy" P (i.e., are *)
(* logically below P).                                                 *)
Record ideal (P : pre) : Type := Ideal {pred : heap -> Prop; _ : pred << P}.
Implicit Arguments Ideal [P].

(* ideal P1 can be relaxed into ideal P2 if P1 -> P2 *)
Lemma relax_bound P1 P2 (p : ideal P1) : P1 << P2 -> pred p << P2.
Proof. by move=>P1 P2 [p] H1 H2; apply: hptrans H2. Qed.

Definition relax (P1 P2 : pre) (p : ideal P1) (pf : forall i, P1 i -> P2 i) : ideal P2 := 
  Ideal (pred p) (relax_bound pf).

(* Intuitively, the predicate transformer model axiomatizes pushing sp's. *)
(* Model of a computation ST P A Q takes a predicate satisfying the       *)
(* precondition P and produces the strongest postcondition for the        *)
(* computation with respect to P. This postcondition must also depend on  *)
(* the return result of the computation. The model function must be       *)
(* monotone, and it must be bounded from above by Q.                      *)

Definition model (P : pre) (A : Type) : Type := ideal P -> ans A -> heap -> Prop. 

Definition monotoneX (P : pre) (A : Type) (sp : model P A) :=
  forall (r1 r2 : ideal P), pred r1 << pred r2 -> forall x, sp r1 x << sp r2 x.

Definition boundedX (P : pre) (A : Type) (sp : model P A) (Q : post A) := 
  forall r x, sp r x << fun h => exists i, pred r i /\ Q x i h.  

Record ST  (P : pre) (A : Type) (Q : post A) : Type :=
  build {sp : model P A; _ : monotoneX sp; _ : boundedX sp Q}. 
Implicit Arguments ST []. 
Implicit Arguments build [P A Q].

(**********************************)
(* Lattice structure of the model *)
(**********************************)

Definition leq P A Q (e1 e2 : ST P A Q) := forall r x, sp e1 r x << sp e2 r x.
Implicit Arguments leq [P A Q]. 

Notation "e1 <= e2" := (leq e1 e2) (at level 70). 
Notation "e1 =1 e2" := (e1 <= e2 /\ e2 <= e1) (at level 70). 

Lemma leq_refl P A Q (e : ST P A Q) : e <= e. 
Proof. by move=>P A Q e p x. Qed.
Hint Resolve leq_refl. 

Lemma leq_trans P A Q (e1 e2 e3 : ST P A Q) : e1 <= e2 -> e2 <= e3 -> e1 <= e3.
Proof. by move=>P A Q e1 e2 e3 H1 H2 p x; apply: hptrans (H2 p x). Qed.
Implicit Arguments leq_trans [P A Q e1 e3].  

Section Lattice. 
Variables (P : pre) (A : Type) (Q : post A). 

(* diverging computation *)
Definition bot_sp : model P A := fun p x m => False.
Lemma bot_mono : monotoneX bot_sp. Proof. by []. Qed.
Lemma bot_bound : boundedX bot_sp Q. Proof. by move=>*. Qed.

Definition bot : ST P A Q := build bot_sp bot_mono bot_bound.

Lemma botP : forall e, bot <= e. Proof. by move=>[p] H1 H2 r x. Qed.

(* trivial computation -- most general one, given P and Q *)
Definition top_sp : model P A := fun p x m => exists i, pred p i /\ Q x i m.

Lemma top_mono : monotoneX top_sp. 
Proof. by move=>r1 r2 H x m [i][H1] H2; exists i; split; [apply: H H1|]. Qed.

Lemma top_bound : boundedX top_sp Q. Proof. by []. Qed.

Definition top : ST P A Q := build top_sp top_mono top_bound.

Lemma topP : forall e, e <= top. Proof. by move=>[]. Qed.
End Lattice.

(* call-by-name fixpoints *)
Section CBNFix.  
Variables (P : pre) (A : Type) (Q : post A) (f : ST P A Q -> ST P A Q). 

Definition monotoneN := forall e1 e2, e1 <= e2 -> f e1 <= f e2. 

Definition lfpn_sp : model P A := 
  fun p x m => forall (e : ST P A Q), f e <= e -> sp e p x m.

Lemma lfpn_mono : monotoneX lfpn_sp.
Proof. by move=>r1 r2 H x m H1 e; move/H1; move: e=>[p M B]; apply: M. Qed.

Lemma lfpn_bound : boundedX lfpn_sp Q.
Proof. by move=>r x m; move/(_ (top P Q) (topP (e := f (top P Q)))). Qed.

Definition lfpn := build lfpn_sp lfpn_mono lfpn_bound. 

(* Knaster-Tarski *)
Lemma lfpn_fixed : monotoneN -> f lfpn =1 lfpn. 
Proof.
move=>M.
suff: f lfpn <= lfpn by split=>// r y m; apply; apply: M.
by move=>r y m H1 x H2; apply: (leq_trans (f x)) H1=>//; apply: M=>r1 z r2; apply.
Qed.

Lemma lfpn_least : forall x, f x =1 x -> lfpn <= x.
Proof. by move=>x [H1 _] r y m; apply. Qed.
End CBNFix.

(* call-by-value fixpoints -- done separately to allow range types to depend on args *)
Section CBVFix.  
Variables (A : Type) (B : A -> Type) (p : A -> pre) (q : forall x, post (B x)).
Implicit Arguments q [].
Variable F : (forall x, ST (p x) (B x) (q x)) -> (forall x, ST (p x) (B x) (q x)). 

Definition monotoneV := 
  forall e1 e2, (forall x, e1 x <= e2 x) -> forall x, F e1 x <= F e2 x.

Definition lfpv_sp : forall x, model (p x) (B x) :=
  fun x p y m => forall e, (forall x, F e x <= e x) -> sp (e x) p y m.
Implicit Arguments lfpv_sp []. 

Lemma lfpv_mono : forall x, monotoneX (lfpv_sp x).
Proof. by move=>x r1 r2 H y m H1 e; move/H1; move: e=>[r H2 H3]; apply: H2. Qed.

Lemma lfpv_bound : forall x, boundedX (lfpv_sp x) (q x).
Proof.
move=>x r y m.
set T := fun x => top (p x) (q x). 
have L: forall x, F T x <= T x by move=>*; apply: topP.
by move/(_ T L).
Qed.

Definition lfpv : forall x, ST (p x) (B x) (q x) := 
  fun x => build (lfpv_sp x) (@lfpv_mono x) (@lfpv_bound x).

Lemma lfpv_fixed : monotoneV -> forall x, F lfpv x =1 lfpv x.
Proof.
move=>M.
suff: forall x, F (lfpv) x <= lfpv x by move=>x; split=>// r y m; apply; apply: M.
by move=>x r y m H1 X H2; apply: (leq_trans (F X x)) H1=>//; apply: M=>t r1 z r2; apply.
Qed.

Lemma lfpv_least : forall f, (forall x, F f x =1 f x) -> forall x, lfpv x <= f x.
Proof. by move=>f H x; move: (H x)=>[H1 _] r y m; apply=>t; apply: (proj1 (H t)). Qed.

End CBVFix.

(* When taking fixpoints of a function, we don't really do it for the    *)
(* given function, but for a "completion" of that function to a monotone *)
(* one. Functions which are already monotone complete to themselves, so  *)
(* this is just a trick to avoid proving monotonicity of functions when  *)
(* taking fixpoints. In other words, if we were not careful, and managed *)
(* to write a function which is not monotone, we may get a result which  *)
(* is a bit surprising, but still sound wrt. the semantics, and still a  *)
(* fixed point. That said, I expect that this kind of tricky functions   *)
(* can only be written if one writes types whose pre/posts use equations *)
(* over effectful code, following the counter-examples of Smith and      *)
(* Crary. *)

Section Completion.
Variables (A : Type) (B : A -> Type) (p : A -> pre) (q : forall x, post (B x)).
Implicit Arguments q [].
Variable F : (forall x, ST (p x) (B x) (q x)) -> (forall x, ST (p x) (B x) (q x)). 

(* at a point e, we take the value of the completed function to be the *)
(* union of the values of F at all points below e. *)
Definition lift_sp (e : forall x, ST (p x) (B x) (q x)) : 
                    (forall x, model (p x) (B x)) :=
   fun x p y m => exists e', (forall x, e' x <= e x) /\ sp (F e' x) p y m.
Implicit Arguments lift_sp [].

Lemma lift_mono e : forall x, monotoneX (lift_sp e x).
Proof.
move=>e x r1 r2 H y h [e'][H'] T.
have M: monotoneX (sp (F e' x)) by move: (F e' x)=>[sp1] M ?; apply: M. 
by exists e'; split=>//; apply: M T.
Qed.

Lemma lift_bound e : forall x, boundedX (lift_sp e x) (q x).
Proof.  
move=>e x r y h [e'][H'] T.
have H: boundedX (sp (F e' x)) (q x) by move: (F e' x)=>[sp1] ? B1; apply: B1.
by apply: H T.
Qed.

Definition lift e : forall x, ST (p x) (B x) (q x) := 
  fun x => build (lift_sp e x) (@lift_mono e x) (@lift_bound e x).

Definition lfp : forall x, ST (p x) (B x) (q x) := lfpv lift.

Lemma lfp_fixed : forall x, lift lfp x =1 lfp x.
Proof.
apply: lfpv_fixed=>e1 e2 H x r y h [e'][H'] T.
by exists e'; split=>// t; apply: leq_trans (H' t) (H t).
Qed.

Lemma lfp_least : forall f, (forall x, lift f x =1 f x) -> forall x, lfp x <= f x.
Proof. by apply: lfpv_least. Qed.

(* monotone functions complete to themselves *)
Lemma lift_fix : monotoneV F -> forall e x, lift e x =1 F e x.
Proof.
move=>H e x; split=>h y i.
- by move=>[e'][H'] T; apply: H T.
by move=>T; exists e.
Qed.

End Completion. 

(******************************)
(* Modeling language commands *)
(******************************)

(* monadic unit *)
Definition ret_pre : pre := def.
Definition ret_post A (x : A) : post A := fun y i m => m = i /\ y = Val x.
Definition ret_sp A (x : A) : model ret_pre A := fun p y m => pred p m /\ y = Val x.

Lemma ret_mono A (x : A) : monotoneX (ret_sp x). 
Proof. by move=>A x r1 r2 H y i [H1 H2]; split; [apply: H H1|]. Qed.

Lemma ret_bound A (x : A) : boundedX (ret_sp x) (ret_post x).
Proof. by move=>A x r y i [H1 H2]; exists i. Qed.

Definition ret A (x : A) : ST ret_pre A (ret_post x) := 
  build (ret_sp x) (@ret_mono A x) (@ret_bound A x).

(* throw *)
Definition throw_pre : pre := def.
Definition throw_post A (x : exn) : post A := fun y i m => m = i /\ y = Exn x.
Definition throw_sp A (x : exn) : model throw_pre A := fun p y m => pred p m /\ y = Exn x.

Lemma throw_mono A (x : exn) : monotoneX (@throw_sp A x).
Proof. by move=>A x r1 r2 H y i [H1 H2]; split; [apply: H H1|]. Qed.

Lemma throw_bound A (x : exn) : boundedX (@throw_sp A x) (@throw_post A x).
Proof. by move=>A x r y i [H1 H2]; exists i. Qed.

Definition throw A (x : exn) : ST throw_pre A (throw_post x) := 
  build (@throw_sp A x) (@throw_mono A x) (@throw_bound A x).

(* monadic bind *)
Section Bind. 
Variables (A B : Type) (p1 : pre) (q1 : post A) (p2 : A ->pre) (q2 : A -> post B). 
Variables (c1 : ST p1 A q1) (c2 : forall x, ST (p2 x) B (q2 x)). 

Definition bind_pre : pre := fun i => p1 i /\ forall x m, q1 (Val x) i m -> p2 x m.
Definition bind_post : post B :=
  fun y i m => (exists x, exists h, q1 (Val x) i h /\ q2 x y h m) \/
               (exists e, y = Exn e /\ q1 (Exn e) i m).

Lemma bind_preP : forall i, bind_pre i -> p1 i. 
Proof. by move=>i; case. Qed.

Definition c1_sp (p : ideal bind_pre) (x : A) := sp c1 (relax p bind_preP) (Val x).

Lemma c1_leq (p : ideal bind_pre) (x : A) : c1_sp p x << p2 x. 
Proof. 
move=>[p H] x m; rewrite /c1_sp; case: c1=>e1 _ B1 /=.
by case/B1=>i /= []; case/(H i)=>_; apply.
Qed.

Definition bind_sp : model bind_pre B := 
  fun p y m => exists x, sp (c2 x) (Ideal (c1_sp p x) (@c1_leq p x)) y m.

Lemma bind_mono : monotoneX bind_sp.
Proof.
move=>r1 r2 H y i [x] T; exists x. 
case: (c2 x) T=>e2 M2 B2; apply: M2=>/=.
by rewrite /c1_sp; case: c1=>e1 M1 B1; apply: M1.
Qed.

Lemma bind_bound : boundedX bind_sp bind_post.
Proof.
move=>r y m [x]; case: (c2 x)=>e2 M2 B2 /=; case/B2=>h /= [].
rewrite /c1_sp; case: c1=>e1 M1 B1 /=; case/B1=>i /= [H1 H2] H3.
by exists i; split=>[//|]; left; exists x; exists h.
Qed.

Definition bind := build bind_sp bind_mono bind_bound. 
End Bind.

(* associative try, Benton/Kennedy style *)
Section Try.
Variables (A B : Type) (p1 : pre) (q1 : post A) (p2 : A -> pre) (q2 : A -> post B)
          (p3 : exn -> pre) (q3 : exn -> post B).
Variables (c1 : ST p1 A q1) (c2 : forall x, ST (p2 x) B (q2 x))
          (c3 : forall e, ST (p3 e) B (q3 e)).

Definition try_pre : pre := 
  fun i => p1 i /\ (forall x m, q1 (Val x) i m -> p2 x m) /\ 
                    forall e m, q1 (Exn e) i m -> p3 e m.
Definition try_post : post B := 
  fun y i m => (exists x, exists h, q1 (Val x) i h /\ q2 x y h m) \/
               (exists e, exists h, q1 (Exn e) i h /\ q3 e y h m).

Lemma try_preP : forall i, try_pre i -> p1 i.
Proof. by move=>i; case. Qed.

Definition c1_val (p : ideal try_pre) (x : A) := sp c1 (relax p try_preP) (Val x).
Definition c1_exn (p : ideal try_pre) (e : exn) := sp c1 (relax p try_preP) (Exn e).  

Lemma val_leq (p : ideal try_pre) (x : A) : c1_val p x << p2 x.
Proof.
move=>[p H] x m; rewrite /c1_val; case: c1=>e1 _ B1 /=.
by case/B1=>i /= []; case/(H i)=>_ [H1 _]; apply: H1.
Qed.

Lemma exn_leq (p : ideal try_pre) (e : exn) : c1_exn p e << p3 e.
Proof. 
move=>[p H] x m; rewrite /c1_exn; case: c1=>e1 _ B1 /=.
by case/B1=>i /= []; case/(H i)=>_ [_ H2]; apply: H2.
Qed.

Definition try_sp : model try_pre B := 
  fun p y m => exists x, 
                 match x with 
                   Val x' => sp (c2 x') (Ideal (c1_val p x') (@val_leq p x')) y m
                 | Exn e' => sp (c3 e') (Ideal (c1_exn p e') (@exn_leq p e')) y m
                 end.

Lemma try_mono : monotoneX try_sp.
Proof.
move=>r1 r2 H y i [x] T; exists x; case: x T=>[x'|e'];
[case: (c2 x') | case: (c3 e')]; move=>e2 M2 B2; apply: M2=>/=;
by rewrite /c1_val /c1_exn; case: c1=>e1 M1 B1; apply: M1.
Qed.

Lemma try_bound : boundedX try_sp try_post.
Proof.
move=>r y m []; case=>[x'|e']; [case: (c2 x') | case: (c3 e')]; move=>e2 M2 B2 /=; 
case/B2=>h /= []; rewrite /c1_val /c1_exn; case: c1=>e1 M1 B1 /=; case/B1=> i /= [H1 H2] H3;
exists i; split=>//; [left; exists x' | right; exists e']; 
by exists h. 
Qed.

Definition try := build try_sp try_mono try_bound.
End Try.

(* call-by-value fixpoint *)
Definition ffix (A : Type) (B : A -> Type) (p : A -> pre) (q : forall x, post (B x)) 
                (F : (forall x, ST (p x) (B x) (q x)) -> (forall x, ST (p x) (B x) (q x))) :
           forall x, ST (p x) (B x) (q x) := lfp F.

Section Consequence.
Variables (A : Type) (p1 p2 : pre) (q1 q2 : post A) (c : ST p1 A q1)
          (pf : forall i, p2 i -> p1 i /\ forall x m, q1 x i m -> q2 x i m). 

Lemma doP : forall i, p2 i -> p1 i.
Proof. by move=>i; case/pf. Qed.

Definition do_sp : model p2 A := fun p => sp c (relax p doP).

Lemma do_mono : monotoneX do_sp.
Proof. by move=>r1 r2 H x; rewrite /do_sp; case: c=>e M B; apply: M. Qed.

Lemma do_bound : boundedX do_sp q2.
Proof.
move=>[r H] x h; rewrite /do_sp; case: c=>e M B; case/B=>i /= [H1 H2].
by exists i; split=>//; case: (pf (H i H1))=>_; apply.
Qed.

Definition do := build do_sp do_mono do_bound.
End Consequence.

Section Allocation.
Variables (A : Type) (v : A).  
Definition alloc_pre : pre := def.
Definition alloc_post : post loc :=
  fun y i m => exists r:loc, r != null /\ y = Val r /\ ~~ valid r i /\ 
                             upd i r (dyn v) = m.

(*
Definition alloc_sp : model alloc_pre loc := 
  fun p x m => if x is Val r then 
                  exists i, pred p i /\ ~~ valid r i /\ upd i r (dyn v) = m
               else False.

Lemma alloc_mono : monotoneX alloc_sp.
Proof.  by move=>r1 r2 H; case=>[r m [i][H1] H2|//]; exists i; split; [apply: H|]. Qed.

Lemma alloc_bound : boundedX alloc_sp alloc_post.
Proof. by move=>p x m; case: x=>[r [i][H1][H2] H3|//]; exists i; split; [|exists r]. Qed.
*)

Definition alloc_sp : model alloc_pre loc := 
  fun p x m => exists i, x = Val (fresh i) /\ pred p i /\ 
               upd i (fresh i) (dyn v) = m.

Lemma alloc_mono : monotoneX alloc_sp.
Proof. by move=>r1 r2 H x r [i][->][H2] <-; exists i; auto. Qed.

Lemma alloc_bound : boundedX alloc_sp alloc_post.
Proof. 
move=>p x m [i][->][H1] <-; exists i.
by split; [| exists (fresh i); rewrite fresh_null -{3}(offset0 (fresh i)) freshness].
Qed.
 
Definition alloc := build alloc_sp alloc_mono alloc_bound.
End Allocation.

Section BlockAllocation. 
Variables (A : Type) (v : A) (n : nat).

Definition allocb_pre : pre := def.
Definition allocb_post : post loc := 
  fun y i m => exists r, y = Val r /\ m = i :+ updi r (nseq n v) /\ def m.

Definition allocb_sp : model allocb_pre loc := 
  fun p x m => exists i, x = Val (fresh i) /\ pred p i /\ m = i :+ updi (fresh i) (nseq n v).

Lemma allocb_mono : monotoneX allocb_sp.
Proof. by move=>r1 r2 H x r [i][->][H1][H2]; exists i; auto. Qed.

Lemma allocb_bound : boundedX allocb_sp allocb_post.
Proof.
move=>p x m [i][->][H1] ->; exists i; split=>//.
exists (fresh i); do ![split]=>//.
case: validSP=>//; last first.
- by rewrite updiD // fresh_null. 
- by case: p H1=>p H1 /=; move/H1; rewrite /allocb_pre /def=>->.
case=>y []; case: updiP=>// [[H]][k][->].  
by rewrite (negbTE (freshness i k)).
Qed.

Definition allocb := build allocb_sp allocb_mono allocb_bound.
End BlockAllocation.        

Section Deallocation. 
Variable r : loc. 
Definition dealloc_pre : pre := valid r.
Definition dealloc_post : post unit := fun y i m => y = Val tt /\ free r i = m.

Definition dealloc_sp : model dealloc_pre unit :=
  fun p x m => if x is Val _ then exists i, pred p i /\ free r i = m else False. 

Lemma dealloc_mono : monotoneX dealloc_sp.
Proof. by move=>r1 r2 H; case=>[x m [i][H1] <-|//]; exists i; split; [apply: H|]. Qed.

Lemma dealloc_bound : boundedX dealloc_sp dealloc_post.
Proof. by move=>p x m; case: x=>[[][i][H1] H2|//]; exists i. Qed.

Definition dealloc := build dealloc_sp dealloc_mono dealloc_bound.
End Deallocation.

Section Read.
Variable (A : Type) (r : loc). 
Definition read_pre : pre := fun i => valid r i /\ exists v:A, look r i = dyn v.
Definition read_post : post A :=
  fun y i m => m = i /\ forall v, look r i = dyn v -> y = Val v.

Definition read_sp : model read_pre A := 
  fun p x m => if x is Val x' then exists i, pred p i /\ m = i /\ 
                                   forall v, look r i = dyn v -> x = Val v
               else False.

Lemma read_mono : monotoneX read_sp.
Proof. by move=>r1 r2 H; case=>[x' m [i][H1] H2|//]; exists i; split; [apply: H|]. Qed.

Lemma read_bound : boundedX read_sp read_post.
Proof. by move=>p; case=>[|*]. Qed.

Definition read := build read_sp read_mono read_bound.
End Read.

Implicit Arguments read [A].
  
Section Write.
Variable (A : Type) (r : loc) (v : A).
Definition write_pre : pre := valid r.
Definition write_post : post unit := fun y i m => y = Val tt /\ upd i r (dyn v) = m.

Definition write_sp : model write_pre unit := 
  fun p x m => if x is Val x' then 
                 exists i, pred p i /\ x = Val tt /\ upd i r (dyn v) = m
               else False.

Lemma write_mono : monotoneX write_sp.
Proof. by move=>r1 r2 H; case=>[[] m [i][H1] H2|//]; exists i; split; [apply: H|]. Qed.

Lemma write_bound : boundedX write_sp write_post.
Proof. by move=>p; case=>[|*]. Qed.

Definition write := build write_sp write_mono write_bound.
End Write.

End st. 




