(**********************************************)
(* Verification condition generator for STsep *)
(**********************************************)

Require Import ssreflect eqtype ssrbool ssrnat seq ssrfun.
Require Import rels heaps stmodel stsep hprop.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(* VC-gen consist of a set of lemmas whose application to the principal   *)
(* spec of a program extracts the proof obligation for the immediate      *)
(* command.  For each primitive command there are base case, bind case    *)
(* and try case. *)

(* First lemma shows that each bind is a special case of try with an      *)
(* exception continuation that just re-throws the exception.  We still    *)
(* keep bind as a primitive construct, as actually treating it via try in *)
(* practice incurrs obligations about the correctness of the re-raising   *)
(* continuation. *)

Lemma bind_is_try (A B : Type) (s1 : spec A) 
                  (s2 : A -> spec B) (i : heap) (r : ans B -> heap -> Prop) : 
        verify i (try_s s1 s2 (fun y => fr (throw_s B y))) r ->
        verify i (bind_s s1 s2) r.
Proof.
move=>A B s1 s2 i r [[H1]][H2] H3 H4. 
split=>// y m [[x]|[e]]; first by move=>[h][H5] H6; apply: H4; left; eauto.
case=>-> H5; apply: H4; right.
exists e; exists m; do !split=>//.
move=>i1 m1 D1 H6 Em; rewrite -{m}Em in H5 *.
exists i1; do !split=>//.
move=>i2 m2 D2 H7 Ei; rewrite -{i1}Ei in D1 H6 H5 *.
by exists i2.
Qed.

(**********)
(* return *)
(**********)

Lemma cval_ret (A : Type) (v : A) (r : ans A -> heap -> Prop) : 
        r (Val v) empty -> verify empty (ret_s v) r. 
Proof. by move=>A v r H; split=>// y m [->] ->. Qed.

Lemma val_ret (A : Type) (v : A) (i : heap) (r : ans A -> heap -> Prop) : 
        r (Val v) i -> def i -> verify i (fr (ret_s v)) r. 
Proof.
by move=>A v i r H D; apply: frame1=>//; apply: cval_ret; rewrite un0h.
Qed. 

Lemma ctry_ret (A B : Type) (v : A) (s1 : A -> spec B) 
              (s2 : exn -> spec B) (i : heap) (r : ans B -> heap -> Prop) : 
        verify i (fr (s1 v)) r ->
          def i -> verify i (try_s (ret_s v) s1 s2) r.
Proof.
move=>A B v s1 s2 i r [H1] H2 D. 
split.
- split; first by apply: (starI empty i); rewrite ?un0h.
  split=>x h [];
  move/(_ empty i); rewrite un0h; move/(_ D (empI (refl_equal _)) (refl_equal _));
  move=>[h1][->][[//]][->][->] _.
  by rewrite un0h.
move=>y m [[x]|[e]]; move=>[h]/= [H3] H4;
case: H3; move/(_ empty i); rewrite un0h; move/(_ D (empI (refl_equal _)) (refl_equal _));
move=>[h1][Eh][[]] // Eh1 [Ev] _.
apply: H2.
by rewrite {x}Ev in H4; rewrite {h1}Eh1 un0h in Eh *; rewrite {h}Eh in H4. 
Qed.

Lemma try_ret (A B : Type) (v : A) (s1 : A -> spec B) 
              (s2 : exn -> spec B) (i : heap) (r : ans B -> heap -> Prop) : 
        verify i (fr (s1 v)) r ->
          def i -> verify i (fr (try_s (ret_s v) s1 s2)) r.
Proof.
move=>A B v s1 s2 i r [H1] H2 D.
by apply: frame0=>//; apply: ctry_ret.
Qed. 

Lemma bnd_ret (A B : Type) (v : A) (s : A -> spec B) (i : heap) 
              (r : ans B -> heap -> Prop) : 
        verify i (fr (s v)) r -> 
          def i -> verify i (fr (bind_s (ret_s v) s)) r.
Proof. 
move=>A B v s i r H D; apply: frame0=>//; apply: bind_is_try.
by apply: ctry_ret.
Qed.  

(* testing *)
Program
Definition test_val_ret (x : nat) : STsep nat (def, fun y i m => i = m /\ y = Val (x+1)) :=
  Do (ret x+1).
Next Obligation. by apply: val_ret. Qed.

Program 
Definition test_bnd_ret (x : nat) : STsep nat (def, fun y i m => i = m /\ y = Val (x+3)) :=
  Do (y <-- ret x + 1; ret y + 2).
Next Obligation. by apply: bnd_ret=>//; apply: val_ret=>//; rewrite -addnA. Qed.

(********)
(* read *)
(********)

Lemma cval_read (A : Type) (v : A) (l : loc) (r : ans A -> heap -> Prop) :
        r (Val v) (l :-> v) -> 
          def (l :-> v) -> verify (l :-> v) (read_s A l) r.
Proof.
move=>A v l r H1 H.
split=>[|y m [<-]]; first by exists v.
have T: stsep.pts l v (l :-> v) by [].
by move/(_ v T)=>->.
Qed.

Lemma val_read (A : Type) (v : A) (l : loc) (i : heap) (r : ans A -> heap -> Prop) :
        r (Val v) (l :-> v :+ i) -> 
          def (l :-> v :+ i) -> verify (l :-> v :+ i) (fr (read_s A l)) r.
Proof.
by move=>*; apply: frame=>// D; apply: frame0=>//; apply: cval_read.
Qed. 

Lemma ctry_read (A B : Type) (v : A) (l : loc) (s1 : A -> spec B)
               (s2 : exn -> spec B) (i : heap) 
               (r : ans B -> heap -> Prop) : 
        verify (l :-> v :+ i) (fr (s1 v)) r -> 
        def (l :-> v :+ i) -> 
          verify (l :-> v :+ i) (try_s (read_s A l) s1 s2) r. 
Proof.
move=>A B v l s1 s2 i r [H1 H2] D.
have Q: (stsep.pts l v) (l :-> v) by apply: ptsI; rewrite (unDl D).
split.
- split; first by apply: (starI (l :-> v) i)=>//; exists v.
  split=>y h [];
  move/(_ (l :-> v) i D (ex_intro _ v _)); move/(_ Q (refl_equal _)); 
  by move=>[h1][->{h}][[<-{h1}]]; move/(_ v Q)=>// [->] _.
move=>y m [[x]|[e]][h][[]];
move/(_ (l :-> v) i D (ex_intro _ v _)); move/(_ Q (refl_equal _));
move=>[h1][->{h}][[<-{h1}]]; move/(_ v Q)=>// [->] _ H.
by apply: H2.
Qed.

Lemma try_read (A B : Type) (v : A) (l : loc) (s1 : A -> spec B)
               (s2 : exn -> spec B) (i : heap) 
               (r : ans B -> heap -> Prop) : 
        verify (l :-> v :+ i) (fr (s1 v)) r -> 
        def (l :-> v :+ i) -> 
          verify (l :-> v :+ i) (fr (try_s (read_s A l) s1 s2)) r. 
Proof. by move=>*; apply: frame0=>//; apply: ctry_read. Qed.

Lemma bnd_read (A B : Type) (v : A) (l : loc) (s : A -> spec B) 
               (i : heap) (r : ans B -> heap -> Prop) : 
        verify (l :-> v :+ i) (fr (s v)) r -> 
        def (l :-> v :+ i) -> 
          verify (l :-> v :+ i) (fr (bind_s (read_s A l) s)) r.
Proof. by move=>*; apply: frame0=>//; apply: bind_is_try; apply: ctry_read. Qed.

(* special instances, when i = empty *)

Lemma val_read0 (A : Type) (v : A) (l : loc) (r : ans A -> heap -> Prop) :
        r (Val v) (l :-> v) -> 
          def (l :-> v) -> verify (l :-> v) (fr (read_s A l)) r.
Proof. by move=>*; rewrite -(unh0 (_ :-> _)); apply: val_read; rewrite unh0. Qed.  

Lemma try_read0 (A B : Type) (v : A) (l : loc) (s1 : A -> spec B)
                (s2 : exn -> spec B) (r : ans B -> heap -> Prop) : 
        verify (l :-> v) (fr (s1 v)) r -> def (l :-> v) -> 
        verify (l :-> v) (fr (try_s (read_s A l) s1 s2)) r.
Proof. by move=>A B v l s1 s2 rl; rewrite -(unh0 (_ :-> _)); apply: try_read. Qed.

Lemma bnd_read0 (A B : Type) (v : A) (l : loc) (s : A -> spec B)
                (r : ans B -> heap -> Prop) : 
        verify (l :-> v) (fr (s v)) r -> def (l :-> v) -> 
        verify (l :-> v) (fr (bind_s (read_s A l) s)) r.
Proof. by move=>A B v l s r; rewrite -(unh0 (_ :-> _)); apply: bnd_read. Qed.

(*********)
(* write *)
(*********)

Lemma cval_write (A B : Type) (w : B) (l : loc) (v : A) (r : ans unit -> heap -> Prop) :
        r (Val tt) (l :-> v) ->
         def (l :-> w) -> verify (l :-> w) (write_s l v) r.
Proof.
move=>A B w l v r H2 H1.
by split=>[|y m [[[<-]] _ ->]]; first by exists B; exists w.
Qed.

Lemma val_write (A B : Type) (w : B) (l : loc) (v : A) (i : heap) (r : ans unit -> heap -> Prop) :
        (def (l :-> v :+ i) -> r (Val tt) (l :-> v :+ i)) -> 
         def (l :-> w :+ i) -> verify (l :-> w :+ i) (fr (write_s l v)) r.
Proof.
by move=>A B w l v i r H2 D; apply: frame=>// D1; apply: frame0=>//; apply: cval_write.
Qed.

Lemma ctry_write (A B C : Type) (w : C) (l : loc) (v : A) 
                (s1 : unit -> spec B)
                (s2 : exn -> spec B) (i : heap) 
                (r : ans B -> heap -> Prop) : 
        (def (l :-> v :+ i) -> verify (l :-> v :+ i) (fr (s1 tt)) r) -> 
         def (l :-> w :+ i) -> verify (l :-> w :+ i) (try_s (write_s l v) s1 s2) r. 
Proof.
move=>A B C w l v s1 s2 i r H1 H.
have Q : stsep.pts l w (l :-> w) by apply: ptsI; rewrite (unDl H).
have T : def (l :-> v :+ i) by move: H; rewrite !defPUn.
have R : write_pre l (l :-> w) by exists C; exists w. 
move: {H1} (H1 T)=>[H1 H2].
split.
- split; first by apply: (starI (l :-> w) i)=>//; by exists C; exists w.
  split=>[x|e] h []; move/(_ (l :-> w) i H R (refl_equal _)); 
  by move=>[h1][->][[]] // [[<-]] _ [->] _.
move=>y m [[x]|[e]] [h] [[]];
move/(_ _ _ H R (refl_equal _))=>[h1][->{h}][[]] // [[<-]] _ [->] _.
by apply: H2.
Qed.

Lemma try_write (A B C : Type) (w : C) (l : loc) (v : A) 
                (s1 : unit -> spec B)
                (s2 : exn -> spec B) (i : heap) 
                (r : ans B -> heap -> Prop) : 
        (def (l :-> v :+ i) -> verify (l :-> v :+ i) (fr (s1 tt)) r) -> 
         def (l :-> w :+ i) -> verify (l :-> w :+ i) (fr (try_s (write_s l v) s1 s2)) r. 
Proof. by move=>*; apply: frame0=>//; apply: ctry_write. Qed.   

Lemma bnd_write (A B C : Type) (w : C) (l : loc) (v : A) 
                (p1 : unit -> spec B)
                (i : heap) (r : ans B -> heap -> Prop) : 
        (def (l :-> v :+ i) -> verify (l :-> v :+ i) (fr (p1 tt)) r) -> 
         def (l :-> w :+ i) -> verify (l :-> w :+ i) (fr (bind_s (write_s l v) p1)) r. 
Proof. by move=>*; apply: frame0=>//; apply: bind_is_try; apply: ctry_write. Qed.

(* Special instances when i = empty *)

Lemma val_write0 (A B : Type) (w : B) (l : loc) (v : A) (r : ans unit -> heap -> Prop) :
        (def (l :-> v) -> r (Val tt) (l :-> v)) ->
         def (l :-> w) -> verify (l :-> w) (fr (write_s l v)) r.
Proof. by move=>*; rewrite -(unh0 (_ :-> _)); apply: val_write; rewrite unh0. Qed.   

Lemma try_write0 (A B C : Type) (w : C) (l : loc) (v : A) 
         (p1 : unit -> spec B)
         (p2 : exn -> spec B) (r : ans B -> heap -> Prop) : 
        (def (l :-> v) -> verify (l :-> v) (fr (p1 tt)) r) -> 
         def (l :-> w) -> verify (l :-> w) (fr (try_s (write_s l v) p1 p2)) r.
Proof.
move=>A B C w l v p1 p2 r H; rewrite -(unh0 (_ :-> _)); apply: try_write.
by rewrite unh0.
Qed.

Lemma bnd_write0 (A B C : Type) (w : C) (l : loc) (v : A) 
        (p1 : unit -> spec B) (r : ans B -> heap -> Prop) : 
        (def (l :-> v) -> verify (l :-> v) (fr (p1 tt)) r) -> 
         def (l :-> w) -> verify (l :-> w) (fr (bind_s (write_s l v) p1)) r.
Proof. 
move=>A B C w l v p1 p2 r H; rewrite -(unh0 (_ :-> _)).
by apply: bnd_write; rewrite unh0.
Qed.

(**************)
(* allocation *)
(**************)

Lemma cval_alloc (A : Type) (v : A) (r : ans loc -> heap -> Prop) : 
        (forall x : loc, def (x :-> v) -> r (Val x) (x :-> v)) -> 
         verify empty (alloc_s v) r.
Proof.
by move=>A v r H; split=>// y m [x][->][[<-]]; apply: H.
Qed.

Lemma val_alloc (A : Type) (v : A) (i : heap) (r : ans loc -> heap -> Prop) : 
        (forall x : loc, def (x :-> v :+ i) -> r (Val x) (x :-> v :+ i)) -> 
         def i -> verify i (fr (alloc_s v)) r.
Proof.
by move=>*; apply: frame1=>//; apply: cval_alloc; move=>x _; move: x.
Qed. 

Lemma ctry_alloc (A B : Type) (v : A) (p1 : loc -> spec B) 
                (p2 : exn -> spec B) (i : heap) 
                (r : ans B -> heap -> Prop) : 
        (forall x : loc, def (x :-> v :+ i) -> verify (x :-> v :+ i) (fr (p1 x)) r) ->
         def i -> verify i (try_s (alloc_s v) p1 p2) r.
Proof.
move=>A B v p1 p2 i r H1 H.
have T1 : alloc_pre empty by apply: empI.
split.
- split; first by apply: (starI empty i); rewrite ?un0h.
  split; move=>x h [];
  move/(_ empty i); rewrite un0h; move/(_ H T1 (refl_equal _));
  move=>[h1][->{h}][[l]][//][->{x}][[<-{h1}] _] H3.
  by move: (H1 _ H3)=>[H4 H5].
move=>y m [[x]|[e]] [h] [[]];
move/(_ empty i); rewrite un0h; move/(_ H T1 (refl_equal _));
move=>[h2 [->{h} [[l [// [<-{l}] [[<-{h2}]_]]] H3]]].
move: (H1 _ H3)=>[H4 H6].
by apply: H6.
Qed.

Lemma try_alloc (A B : Type) (v : A) (p1 : loc -> spec B) 
                (p2 : exn -> spec B) (i : heap) 
                (r : ans B -> heap -> Prop) : 
        (forall x : loc, def (x :-> v :+ i) -> verify (x :-> v :+ i) (fr (p1 x)) r) ->
         def i -> verify i (fr (try_s (alloc_s v) p1 p2)) r.
Proof. by move=>*; apply: frame0=>//; apply: ctry_alloc. Qed.   

Lemma bnd_alloc (A B : Type) (v : A) (p : loc -> spec B) 
                (i : heap) (r : ans B -> heap -> Prop) : 
        (forall x : loc, def (x :-> v :+ i) -> verify (x :-> v :+ i) (fr (p x)) r) ->
         def i -> verify i (fr (bind_s (alloc_s v) p)) r.
Proof. by move=>*; apply: frame0=>//; apply: bind_is_try; apply: ctry_alloc. Qed.

(* special instances when i = empty *)

Lemma val_alloc0 (A : Type) (v : A) (r : ans loc -> heap -> Prop) : 
        (forall x : loc, def (x :-> v) -> r (Val x) (x :-> v)) -> 
         def empty -> verify empty (fr (alloc_s v)) r.
Proof. by move=>*; apply: val_alloc=>// x; rewrite unh0; move: x. Qed.

Lemma try_alloc0 (A B : Type) (v : A) (p1 : loc -> spec B) 
                (p2 : exn -> spec B) (r : ans B -> heap -> Prop) : 
        (forall x : loc, def (x :-> v) -> verify (x :-> v) (fr (p1 x)) r) ->
        def empty -> verify empty (fr (try_s (alloc_s v) p1 p2)) r.
Proof. by move=>A B v p1 p2 r H; apply: try_alloc=>x; rewrite unh0; apply: H. Qed.

Lemma bnd_alloc0 (A B : Type) (v : A) (p : loc -> spec B) 
                 (r : ans B -> heap -> Prop) : 
        (forall x : loc, def (x :-> v) -> verify (x :-> v) (fr (p x)) r) ->
        def empty -> verify empty (fr (bind_s (alloc_s v) p)) r.
Proof. by move=>A B v p r H; apply: bnd_alloc=>// x; rewrite unh0; apply: H. Qed.

(********************)
(* block allocation *)
(********************)

Lemma cval_allocb (A : Type) (v : A) (n : nat) (r : ans loc -> heap -> Prop) : 
        (forall x : loc, def (updi x (nseq n v)) -> 
                         r (Val x) (updi x (nseq n v))) -> 
         verify empty (allocb_s v n) r.
Proof. by move=>A v n r H; split=>// y m [x][->][->]; apply: H. Qed.

Lemma val_allocb (A : Type) (v : A) (n : nat) (i : heap) (r : ans loc -> heap -> Prop) : 
        (forall x : loc, def (updi x (nseq n v) :+ i) -> 
                         r (Val x) (updi x (nseq n v) :+ i)) -> 
         def i -> verify i (fr (allocb_s v n)) r.
Proof. by move=>*; apply: frame1=>//; apply: cval_allocb; move=>x _; move: x. Qed.

Lemma ctry_allocb (A B : Type) (v : A) (n : nat) (p1 : loc -> spec B) 
                 (p2 : exn -> spec B) (i : heap) 
                 (r : ans B -> heap -> Prop) : 
        (forall x : loc, def (updi x (nseq n v) :+ i) -> 
           verify (updi x (nseq n v) :+ i) (fr (p1 x)) r) ->
         def i -> 
         verify i (try_s (allocb_s v n) p1 p2) r.
Proof.
move=>A B v n p1 p2 i r H1 H.
have T1 : alloc_pre empty by apply: empI.
split.
- split; first by apply: (starI empty i); rewrite ?un0h.
  split; move=>x h [];
  move/(_ empty i); rewrite un0h; move/(_ H T1 (refl_equal _));
  move=>[h1][->{h}][[l]][// [-> {x}]][->] H3 H4.
  by move: (H1 l H4)=>[H5 H6].
move=>y m [[x]|[e]] [h] [[]];
move/(_ empty i); rewrite un0h; move/(_ H T1 (refl_equal _));
move=>[h1][->{h}][[l]][// [->]][->] H3 H4.
by move: (H1 _ H4)=>[H5 H6]; apply: H6.
Qed.

Lemma try_allocb (A B : Type) (v : A) (n : nat) (p1 : loc -> spec B) 
                 (p2 : exn -> spec B) (i : heap) 
                 (r : ans B -> heap -> Prop) : 
        (forall x : loc, def (updi x (nseq n v) :+ i) -> 
           verify (updi x (nseq n v) :+ i) (fr (p1 x)) r) ->
         def i -> 
         verify i (fr (try_s (allocb_s v n) p1 p2)) r.
Proof. by move=>*; apply: frame0=>//; apply: ctry_allocb. Qed.    

Lemma bnd_allocb (A B : Type) (v : A) (n : nat) (p : loc -> spec B) 
                (i : heap) (r : ans B -> heap -> Prop) : 
        (forall x : loc, def (updi x (nseq n v) :+ i) -> 
           verify (updi x (nseq n v) :+ i) (fr (p x)) r) ->
         def i -> verify i (fr (bind_s (allocb_s v n) p)) r.
Proof. by move=>*; apply: frame0=>//; apply: bind_is_try; apply: ctry_allocb. Qed.

(* special instances when i = empty *)

Lemma val_allocb0 (A : Type) (v : A) (n : nat) (r : ans loc -> heap -> Prop) : 
        (forall x : loc, def (updi x (nseq n v)) -> 
                         r (Val x) (updi x (nseq n v))) -> 
         def empty -> verify empty (fr (allocb_s v n)) r.
Proof. by move=>*; apply: val_allocb=>// x; rewrite unh0; move: x. Qed.    

Lemma try_allocb0 (A B : Type) (v : A) (n : nat) (p1 : loc -> spec B) 
                  (p2 : exn -> spec B) (r : ans B -> heap -> Prop) : 
        (forall x : loc, def (updi x (nseq n v)) -> 
           verify (updi x (nseq n v)) (fr (p1 x)) r) ->
         def empty -> 
         verify empty (fr (try_s (allocb_s v n) p1 p2)) r.
Proof. by move=>A B v n p1 p2 r H; apply: try_allocb=>x; rewrite unh0; apply: H. Qed.

Lemma bnd_allocb0 (A B : Type) (v : A) (n : nat) (p : loc -> spec B) 
                  (r : ans B -> heap -> Prop) : 
        (forall x : loc, def (updi x (nseq n v)) -> 
          verify (updi x (nseq n v)) (fr (p x)) r) ->
        def empty -> verify empty (fr (bind_s (allocb_s v n) p)) r.
Proof. by move=>A B v n p r H; apply: bnd_allocb=>// x; rewrite unh0; apply: H. Qed.

(***********)
(* dealloc *)
(***********)

Lemma cval_dealloc (A : Type) (v : A) (l : loc) (r : ans unit -> heap -> Prop) :
        r (Val tt) empty -> 
         def (l :-> v) -> verify (l :-> v) (dealloc_s l) r.
Proof.
move=>A v l r H D.
split=>//; first by exists A; exists v.
by move=>y m [->][<-].
Qed.

Lemma val_dealloc (A : Type) (v : A) (l : loc) (i : heap) (r : ans unit -> heap -> Prop) :
        r (Val tt) i -> 
         def (l :-> v :+ i) -> verify (l :-> v :+ i) (fr (dealloc_s l)) r.
Proof.
move=>*; apply: frame=>// D; apply: frame0=>//; apply: cval_dealloc=>//.
by rewrite un0h.
Qed.

Lemma ctry_dealloc (A B : Type) (v : B) (l : loc) (p1 : unit -> spec A)
                  (p2 : exn -> spec A) (i : heap) (r : ans A -> heap -> Prop) : 
        (def i -> verify i (fr (p1 tt)) r) -> 
         def (l :-> v :+ i) -> verify (l :-> v :+ i) (try_s (dealloc_s l) p1 p2) r.
Proof.
move=>A B v l p1 p2 i r H1 H.
have T : def i by apply: unDr H.
have T1 : dealloc_pre l (l :-> v) by exists B; exists v; split; rewrite (unDl H).
move: (H1 T)=>[H2 H3].
split.
- split; first by apply: (starI (l:->v) i). 
  split=>x h [];
    move/(_ (l:->v) i H T1 (refl_equal _));
    move=>[_ [->{h} [[// [->{x}] [<-]] _]]].
  by rewrite un0h.
move=>y m [[x]|[e]] [h] [[]]; 
move/(_ (l:->v) i H T1 (refl_equal _));
move=>[h3 [->{h} [[// [->{x}] [<-]] _]]].
by rewrite un0h; apply: H3.
Qed.

Lemma try_dealloc (A B : Type) (v : B) (l : loc) (p1 : unit -> spec A)
                  (p2 : exn -> spec A) (i : heap) (r : ans A -> heap -> Prop) : 
        (def i -> verify i (fr (p1 tt)) r) -> 
         def (l :-> v :+ i) -> verify (l :-> v :+ i) (fr (try_s (dealloc_s l) p1 p2)) r.
Proof. by move=>*; apply: frame0=>//; apply: ctry_dealloc. Qed. 

Lemma bnd_dealloc (A B : Type) (v : B) (l : loc) (p : unit -> spec A)
                  (i : heap) (r : ans A -> heap -> Prop) : 
        (def i -> verify i (fr (p tt)) r) -> 
        def (l :-> v :+ i) -> verify (l :-> v :+ i) (fr (bind_s (dealloc_s l) p)) r.
Proof. by move=>*; apply: frame0=>//; apply: bind_is_try; apply: ctry_dealloc. Qed.

(* special instances *)

Lemma val_dealloc0 (A : Type) (v : A) (l : loc) (r : ans unit -> heap -> Prop) :
        r (Val tt) empty -> 
         def (l :-> v) -> verify (l :-> v) (fr (dealloc_s l)) r.
Proof. by move=>*; rewrite -(unh0 (_ :-> _)); apply: val_dealloc; rewrite ?unh0. Qed.    

Lemma try_dealloc0 (A B : Type) (v : B) (l : loc) 
                   (p1 : unit -> spec A)
                  (p2 : exn -> spec A) (r : ans A -> heap -> Prop) : 
        (def empty -> verify empty (fr (p1 tt)) r) -> 
         def (l :-> v) -> verify (l :-> v) (fr (try_s (dealloc_s l) p1 p2)) r.
Proof. by move=>A B v l p1 p2 r H; rewrite -(unh0 (_ :-> _)); apply: try_dealloc. Qed.

Lemma bnd_dealloc0 (A B : Type) (v : B) (l : loc) 
                   (p : unit -> spec A)
                   (r : ans A -> heap -> Prop) : 
        (def empty -> verify empty (fr (p tt)) r) -> 
         def (l :-> v) -> verify (l :-> v) (fr (bind_s (dealloc_s l) p)) r.
Proof. by move=>A B v l p r H; rewrite -(unh0 (_ :-> _)); apply: bnd_dealloc. Qed.

(*********)
(* throw *)
(*********)

Lemma cval_throw (A : Type) (e : exn) (r : ans A -> heap -> Prop) : 
        r (Exn e) empty -> verify empty (throw_s A e) r.
Proof.
by move=>A e r H; split=>// y m [->] ->.
Qed.

Lemma val_throw (A : Type) (e : exn) (i : heap) (r : ans A -> heap -> Prop) : 
        r (Exn e) i -> def i -> verify i (fr (throw_s A e)) r.
Proof. by move=>*; apply: frame1=>//; apply: cval_throw; rewrite un0h. Qed. 

Lemma ctry_throw (A B : Type) (e : exn) (p1 : A -> spec B) 
                (p2 : exn -> spec B) (i : heap) (r : ans B -> heap -> Prop) :
        verify i (fr (p2 e)) r -> def i -> verify i (try_s (throw_s A e) p1 p2) r.
Proof.
move=>A B v p1 p2 i r [H1] H2 D.
split. 
- split; first by apply: (starI empty i); rewrite ?un0h.
  split=>x h [];
  move/(_ empty i); rewrite un0h; move/(_ D (empI (refl_equal _)) (refl_equal _));
  move=>[h1][->][[]] // -> [->] _.
  by rewrite un0h.
move=>y m [[x]|[e]] [h] [[]];
move/(_ empty i); rewrite un0h; move/(_ D (empI (refl_equal _)) (refl_equal _));
move=>[h3 [->{h} [[-> {h3} // [->{x}]] _]]].
by rewrite un0h; apply: H2.
Qed.

Lemma try_throw (A B : Type) (e : exn) (p1 : A -> spec B) 
                (p2 : exn -> spec B) (i : heap) (r : ans B -> heap -> Prop) :
        verify i (fr (p2 e)) r -> def i -> verify i (fr (try_s (throw_s A e) p1 p2)) r.
Proof. by move=>*; apply: frame0=>//; apply: ctry_throw. Qed.  
 
Lemma bnd_throw (A B : Type) (e : exn) (p1 : A -> spec B) (i : heap)
                (r : ans B -> heap -> Prop) :
        r (Exn e) i -> 
        def i -> verify i (fr (bind_s (throw_s A e) p1)) r.
Proof.
move=>A B e p1 i r H D; apply: frame0=>//; apply: bind_is_try; apply: ctry_throw=>//.
by apply: frame0=>//; apply: val_throw.
Qed.

(******)
(* do *)
(******)

Lemma ctry_do (A B : Type) (p1 : spec A) (p2 : A -> spec B) 
             (p3 : exn -> spec B) (i1 i2 : heap) (r : ans B -> heap -> Prop) :
        p1.1 i1 -> 
        (forall x i1', p1.2 (Val x) i1 i1' -> 
           def (i1' :+ i2) -> verify (i1' :+ i2) (fr (p2 x)) r) ->
        (forall e i1', p1.2 (Exn e) i1 i1' -> 
           def (i1' :+ i2) -> verify (i1' :+ i2) (fr (p3 e)) r) ->
        def (i1 :+ i2) -> 
        verify (i1 :+ i2) (try_s p1 p2 p3) r.
Proof.
move=>A B p1 p2 p3 i1 i2 r H1 H2 H3 H.
split.
- split; first by apply: (starI i1 i2).
  split=>x h []; move/(_ i1 i2 H H1 (refl_equal _));
  move=>[i1'][->][T] D; [move: (H2 _ _ T D) | move: (H3 _ _ T D)];
  by move=>[H4 H5].
move=>y m [[x]|[e]] [h] [[]];
move/(_ i1 i2 H H1 (refl_equal _)); move=>[i1'][->][H4] D;
[move: (H2 _ _ H4 D) | move: (H3 _ _ H4 D)]; move=>[H5 H6];
by apply: H6.
Qed.

Lemma try_do (A B : Type) (p1 : spec A) (p2 : A -> spec B) 
             (p3 : exn -> spec B) (i1 i2 : heap) (r : ans B -> heap -> Prop) :
        p1.1 i1 -> 
        (forall x i1', p1.2 (Val x) i1 i1' -> 
           def (i1' :+ i2) -> verify (i1' :+ i2) (fr (p2 x)) r) ->
        (forall e i1', p1.2 (Exn e) i1 i1' -> 
           def (i1' :+ i2) -> verify (i1' :+ i2) (fr (p3 e)) r) ->
        def (i1 :+ i2) -> 
        verify (i1 :+ i2) (fr (try_s p1 p2 p3)) r.
Proof. by move=>*; apply: frame0=>//; apply: ctry_do. Qed.

Implicit Arguments try_do [A B p1 p2 p3 i1 i2 r].

Lemma bnd_do (A B : Type) (p1 : spec A) (p2 : A -> spec B) 
             (i1 i2 : heap) (r : ans B -> heap -> Prop) :
        p1.1 i1 -> 
        (forall x i1', p1.2 (Val x) i1 i1' ->
           def (i1' :+ i2) -> verify (i1' :+ i2) (fr (p2 x)) r) -> 
        (forall e i1', p1.2 (Exn e) i1 i1' -> def (i1' :+ i2) -> r (Exn e) (i1' :+ i2)) ->
        def (i1 :+ i2) -> 
        verify (i1 :+ i2) (fr (bind_s p1 p2)) r.
Proof.
move=>A B p1 p2 i1 i2 r H1 H2 H3 H4; apply: frame0=>//; apply: bind_is_try.
apply: ctry_do=>// e i1' H5 D5.
by apply: frame1=>//; apply: val_throw=>// _; rewrite un0h; apply: H3.
Qed.

Implicit Arguments bnd_do [A B p1 p2 i1 i2 r].

Lemma bnd_do0 (A B : Type) (p1 : spec A) (p2 : A -> spec B)
              (i2 : heap) (r : ans B -> heap -> Prop) : 
       p1.1 empty -> 
       (forall x i1, p1.2 (Val x) empty i1 -> 
          def (i1 :+ i2) -> verify (i1 :+ i2) (fr (p2 x)) r) -> 
       (forall e i1, p1.2 (Exn e) empty i1 -> def (i1 :+ i2) -> r (Exn e) (i1 :+ i2)) ->
       def i2 -> verify i2 (fr (bind_s p1 p2)) r.
Proof.
move=>A B p1 p2 i2 r H1 H2 H3.
rewrite -(un0h i2).
by apply: bnd_do.
Qed.

Implicit Arguments bnd_do0 [A B p1 p2 i2 r].

Lemma val_do (A : Type) (p1 : spec A) (i1 i2 : heap) (r : ans A -> heap -> Prop) : 
         p1.1 i1 -> 
         (forall x i1', p1.2 (Val x) i1 i1' -> def (i1' :+ i2) -> r (Val x) (i1' :+ i2)) -> 
         (forall e i1', p1.2 (Exn e) i1 i1' -> def (i1' :+ i2) -> r (Exn e) (i1' :+ i2)) ->            
         def (i1 :+ i2) -> verify (i1 :+ i2) (fr p1) r.
Proof.
move=>A p1 i1 i2 r H1 H2 H3 D.
apply: frame=>// D1; apply: frame0=>//; split=>//.
by move=>[v|e] m H4; [apply: H2 | apply: H3].
Qed.

Implicit Arguments val_do [A p1 i1 i2 r].

Program 
Definition test_bnd_do (x : nat) : STsep nat (stsep.emp, fun y i m => emp m /\ y = Val (x+2)) :=
  Do (t <-- test_val_ret x; 
      ret t + 1).
Next Obligation.
case: H=><-; rewrite -{1}(unh0 empty); apply: bnd_do=>// y i1'; last by move=>[].
by move=>[<-][->]; apply: val_ret; rewrite -addnA unh0.
Qed.

(* Specialized versions of do lemmas, in the case when q1 hides a ghost variable *)

Lemma try_gh (G : Type) (t : G) (A B : Type) (p1' : G -> Rel heap)
             (q1' : G -> post A) 
             (p2 : A -> spec B) 
             (p3 : exn -> spec B) 
             (i1 i2 : heap) (r : ans B -> heap -> Prop) :
        let: p1 := (fun i => exists x:G, i \In p1' x, 
                    fun y i m => forall x, i \In p1' x -> q1' x y i m) in 
        (forall x i1', q1' t (Val x) i1 i1' -> 
           def (i1' :+ i2) -> verify (i1' :+ i2) (fr (p2 x)) r) ->
        (forall e i1', q1' t (Exn e) i1 i1' -> 
           def (i1' :+ i2) -> verify (i1' :+ i2) (fr (p3 e)) r) ->
        i1 \In p1' t -> 
        def (i1 :+ i2) -> 
        verify (i1 :+ i2) (fr (try_s p1 p2 p3)) r.
Proof. by move=>*; apply: try_do=>/=; eauto. Qed.

Implicit Arguments try_gh [G A B p1' q1' p2 p3 i1 i2 r].

Lemma try_gh1 (G : Type) (t : G) (A B : Type) (p1' : G -> Rel heap) 
              (q1' : G -> post A)
             (p2 : A -> spec B) 
             (p3 : exn -> spec B) (p1 : pre)
             (i1 i2 : heap) (r : ans B -> heap -> Prop) :
        (i1 \In p1' t -> p1 i1) -> 
        let: q1 := (fun y i m => forall x, i \In p1' x -> q1' x y i m) in 
        (forall x i1', q1' t (Val x) i1 i1' -> 
           def (i1' :+ i2) -> verify (i1' :+ i2) (fr (p2 x)) r) ->
        (forall e i1', q1' t (Exn e) i1 i1' -> 
           def (i1' :+ i2) -> verify (i1' :+ i2) (fr (p3 e)) r) ->
        i1 \In p1' t -> 
        def (i1 :+ i2) -> 
        verify (i1 :+ i2) (fr (try_s (p1, q1) p2 p3)) r.
Proof. by move=>*; apply: try_do=>/=; eauto. Qed.

Implicit Arguments try_gh1 [G A B p1' q1' p2 p3 p1 i1 i2 r].

Lemma bnd_gh (G : Type) (t : G) (A B : Type) (p1' : G -> Rel heap) 
             (q1' : G -> post A) 
             (p2 : A -> spec B) 
             (i1 i2 : heap) (r : ans B -> heap -> Prop) :
        let: p1 := (fun i => exists x:G, i \In p1' x, 
                    fun y i m => forall x, i \In p1' x -> q1' x y i m) in 
        (forall x i1', q1' t (Val x) i1 i1' ->
           def (i1' :+ i2) -> verify (i1' :+ i2) (fr (p2 x)) r) -> 
        (forall e i1', q1' t (Exn e) i1 i1' -> 
           def (i1' :+ i2) -> r (Exn e) (i1' :+ i2)) ->
        i1 \In p1' t -> 
        def (i1 :+ i2) -> 
        verify (i1 :+ i2) (fr (bind_s p1 p2)) r.
Proof. by move=>*; apply: bnd_do=>/=; eauto. Qed.

Implicit Arguments bnd_gh [G A B p1' q1' p2 i1 i2 r].

Lemma bnd_gh1 (G : Type) (t : G) (A B : Type) (p1' : G -> Rel heap) 
              (q1' : G -> post A)
             (p2 : A -> spec B) (p1 : pre) 
             (i1 i2 : heap) (r : ans B -> heap -> Prop) :
        (i1 \In p1' t -> p1 i1) ->
        let: q1 := (fun y i m => forall x, i \In p1' x -> q1' x y i m) in 
        (forall x i1', q1' t (Val x) i1 i1' ->
           def (i1' :+ i2) -> verify (i1' :+ i2) (fr (p2 x)) r) -> 
        (forall e i1', q1' t (Exn e) i1 i1' -> def (i1' :+ i2) -> 
           r (Exn e) (i1' :+ i2)) ->
        i1 \In p1' t -> 
        def (i1 :+ i2) -> 
        verify (i1 :+ i2) (fr (bind_s (p1, q1) p2)) r.
Proof. by move=>*; apply: bnd_do=>/=; eauto. Qed.

Implicit Arguments bnd_gh1 [G A B p1' q1' p2 i1 i2 r].

Lemma val_gh (G : Type) (t : G) (A : Type) (p1' : G ->  Rel heap) (q1' : G -> post A) 
             (i1 i2 : heap) (r : ans A -> heap -> Prop) : 
        let: p1 := (fun i => exists x:G, i \In p1' x, 
                    fun y i m => forall x, i \In p1' x -> q1' x y i m) in 
         (forall x i1', q1' t (Val x) i1 i1' -> def (i1' :+ i2) -> r (Val x) (i1' :+ i2)) -> 
         (forall e i1', q1' t (Exn e) i1 i1' -> def (i1' :+ i2) -> r (Exn e) (i1' :+ i2)) ->            
        i1 \In p1' t -> 
         def (i1 :+ i2) -> verify (i1 :+ i2) (fr p1) r.
Proof. by move=>*; apply: val_do=>/=; eauto. Qed.

Implicit Arguments val_gh [G A p1' q1' i1 i2 r].

Lemma val_gh1 (G : Type) (t : G) (A : Type) (p1' : G ->  Rel heap) (q1' : G -> post A) 
              (p1 : pre) (i1 i2 : heap) (r : ans A -> heap -> Prop) : 
        (i1 \In p1' t -> p1 i1) -> 
        let: q1 := (fun y i m => forall x, i \In p1' x -> q1' x y i m) in 
         (forall x i1', q1' t (Val x) i1 i1' -> def (i1' :+ i2) -> r (Val x) (i1' :+ i2)) -> 
         (forall e i1', q1' t (Exn e) i1 i1' -> def (i1' :+ i2) -> r (Exn e) (i1' :+ i2)) ->            
         i1 \In p1' t ->
         def (i1 :+ i2) -> verify (i1 :+ i2) (fr (p1, q1)) r.
Proof. by move=>*; apply: val_do=>/=; eauto. Qed.

Implicit Arguments val_gh1 [G A p1' q1' i1 i2 r].

Lemma cons0 (A : Type) (i : heap) (p : spec A) (r : ans A -> heap -> Prop) : 
        p.1 i -> (forall y m, p.2 y i m -> def m -> r y m) -> def i -> verify i (fr p) r.
Proof.
move=>A i p r H1 H2; rewrite -(unh0 i).
by apply: val_do=>// x i1'; rewrite unh0; apply: H2.
Qed.

Implicit Arguments cons0 [A i p r]. 

Lemma cons_gh (G : Type) (t : G) (A : Type) (p1' : G -> Rel heap) (q1' : G -> post A)
                (i : heap) (r : ans A -> heap -> Prop) :  
        let: p1 := (fun i => exists x:G, i \In p1' x, 
                    fun y i m => forall x, i \In p1' x -> q1' x y i m) in 
         (forall x i', q1' t (Val x) i i' -> def i' -> r (Val x) i') -> 
         (forall e i', q1' t (Exn e) i i' -> def i' -> r (Exn e) i') ->            
        i \In p1' t -> 
         def i -> verify i (fr p1) r.
Proof.
move=>G t A p1' q1' i r H1 H2 H3; apply: cons0=>//=; eauto. 
by case=>[x|e] m H; [apply: H1 | apply: H2]; apply: H.
Qed.

Implicit Arguments cons_gh [G A p1' q1' i r].

Lemma cons_gh1 (G : Type) (t : G) (A : Type) (p1' : G -> Rel heap) (q1' : G -> post A)
                (i : heap) (r : ans A -> heap -> Prop) (p1 : pre) :  
        (i \In p1' t -> p1 i) ->
        let: q1 := (fun y i m => forall x, i \In p1' x -> q1' x y i m) in 
         (forall x i', q1' t (Val x) i i' -> def i' -> r (Val x) i') -> 
         (forall e i', q1' t (Exn e) i i' -> def i' -> r (Exn e) i') ->            
        i \In p1' t -> 
         def i -> verify i (fr (p1, q1)) r.
Proof.
move=>G t A p1' q1' i r H1 H2 H3 H4 H5; apply: cons0=>//=; eauto. 
by case=>[x|e] m H; [apply: H3 | apply: H4]; apply: H.
Qed.

Implicit Arguments cons_gh1 [G A p1' q1' i r p1].

(******************************************************)
(* associativity lemmas. do not emit any vc's         *)
(* per se, but enable vc generation by exposing       *)
(* commuting conversions between a bind and/or a ttry *)
(******************************************************)

(* eval_try_try shows the equivalence of specs for   *)
(*                                                   *)
(* ttry ttry c1                                      *)
(*     then[x2] c2 x2                                *)
(*     else[e3] c3 e3                                *)
(* then[x4] c4 x4                                    *)
(* else[e5] c5 x5                                    *)
(*                                                   *)
(* and                                               *)
(*                                                   *)
(* ttry c1                                           *)
(* then[x2] ttry c2 x2                               *)
(*          then[x4] c4 x4                           *)
(*          else[e5] c5 e5                           *)
(* then[e3] ttry c3 x3                               *)
(*          then[x4] c4 x4                           *)
(*          else[e5] c5 e5                           *)

(*
Lemma try_try (A B C : Type) (p1 : spec A)
              (p2 : A -> spec B)
              (p3 : exn -> spec B)
              (p4 : B -> spec C)
              (p5 : exn -> spec C) (i : heap) (r : ans C -> heap -> Prop) : 
        verify i (try_s p1 (fun x2 => try_s (p2 x2) p4 p5) 
                           (fun e3 => try_s (p3 e3) p4 p5)) r -> 
          verify i (try_s (try_s p1 p2 p3) p4 p5) r.
Proof.
move=>A B C p1 p2 p3 p4 p5 i r.
case=>h [[i1 i2 H1 <- {i} [H2 [H3 H4]] [<- {h}]] H5].
have T : try_pre p1 p2 p3 i1.
- do 2![split=>//]=>y i; [move/(H3 y i)|move/(H4 y i)];
  move=>[h1 h2 H6 <- {h} [H7 _] _]; 
  by apply: frame_pre H7.
exists i2; split.
- apply: (starI _ _ H1)=>//.
  split; first by apply: (starI i1 empty); rewrite ?unh0 //; apply: unDl H1.
  split=>x h []; 
    move/(_ i1 empty); rewrite unh0; move/(_ (unDl H1) T (refl_equal _));
    move=>[h1 [->{h} [H7 H6]]]; rewrite unh0 in H6 *;
    by case: H7; move=>[y [h2 [H7]]];
       [move: (H3 _ _ H7)| move: (H4 _ _ H7)];
        move=>[h3 h4 H8 H9 [H10 [H12 H13]] _ H11]; subst h2;
        move: (diff2_frame_ex H8 H11 H10)=>[m1 [H14 [<- H16]]];
        try (move: (H12 _ _ H16) || move: (H13 _ _ H16)); apply: frame_pre.
move=>y _ [h1 h2 H6 -> H7 _ H8]. 
rewrite -(unKhl H1 H6) in H7 => {h1 H6}. 
apply: H5; apply: star2I=>//.
by case: H7; move=>[x [h [[]]]];
   move/(_ i1 empty); rewrite unh0; move/(_ (unDl H1) T (refl_equal _));
   move=>[h1 [H6 [H9 H10]]] H11; rewrite unh0 in H6 H10; subst h1;
   (case: H9; move=>[z [h1 [H12 H13]]]; [left|right]);
   exists z; exists h1; split=>//;
   split=>h5 h6 H17 H18 H19; subst h1;
   case: H18=>H18 [H19 H20];
   move: (diff2_frame_ex H17 H13 H18)=>[m1 [H21 [H22 H23]]]; subst h;
   try (move: (H19 _ _ H23)||move: (H20 _ _ H23))=>H24; 
   move: (diff2_frame_ex H21 H11 H24)=>[h1 [H25 [H26 H27]]];
   exists h1; do 2![split=>//]; [left|left|right|right];
   exists x; exists m1.
Qed.

Lemma bnd_try (A B C : Type) (p1 : spec A)
              (p2 : A -> spec B) 
              (p3 : exn -> spec B)
              (p4 : B -> spec C) 
              (i : heap) (r : ans C -> heap -> Prop) : 
        verify i (try_s p1 (fun x2 => bind_s (p2 x2) p4) 
                           (fun e3 => bind_s (p3 e3) p4)) r -> 
          verify i (bind_s (try_s p1 p2 p3) p4) r.
Proof.
move=>A B C p1 p2 p3 p4 i r.
case=>h [[i1 i2 H1 <- {i} [H2 [H3 H4]] [<- {h}]] H5].
have T : try_pre p1 p2 p3 i1.
- do 2![split=>//]=>y i; [move/(H3 y i)|move/(H4 y i)];
  move=>[h1 h2 H6 <- {h} [H7 _] _]; 
  by apply: frame_pre H7.
exists i2; split.
- apply: (starI _ _ H1)=>//.
  split; first by apply: (starI i1 empty); rewrite ?unh0 //; apply: unDl H1.
  move=>x h []. 
  move/(_ i1 empty); rewrite unh0; move/(_ (unDl H1) T (refl_equal _)).
  move=>[h1 [H8 [H7 H6]]]; rewrite !unh0 in H6 H8 *; subst h.
  by case: H7; move=>[y [h2 [H7]]];
     [move: (H3 _ _ H7)| move: (H4 _ _ H7)];
      move=>[h3 h4 H8 H9 [H10 H11] _ H12]; subst h2;
      move: (diff2_frame_ex H8 H12 H10)=>[m1 [H14 [<- H16]]]; 
      move: (H11 _ _ H16); apply: frame_pre.
move=>y _ [h1 h2 H6 -> H7 _ H8]. 
rewrite -(unKhl H1 H6) in H7=>{h1 H6}. 
apply: H5; apply: star2I=>//.
case: H7=>[[x [h [[]]]]|[x [-> []]]];
  move/(_ i1 empty); rewrite unh0; move/(_ (unDl H1) T (refl_equal _));
  move=>[h1 [H6 [H9 H10]]]; [move=> H11 | idtac]; 
   rewrite unh0 in H6 H10; subst h1;
   by case: H9; move=>[z [h1 [H12 H13]]]; [left|right];
      exists z; exists h1; do 2![split=>//]=>h5 h6 H17 H18 H19; subst h1;
      case: H18=>H18 H19;
      move: (diff2_frame_ex H17 H13 H18)=>[m1 [H21 [H22 H23]]]; try (subst h || subst h3); 
      try (move: (H19 _ _ H23)=>H24; move: (diff2_frame_ex H21 H11 H24)=>[h1 [H25 [H26 H27]]]; 
           exists h1; do 2![split=>//]; left; exists x; exists m1) ||
          (exists m1; do 2![split=>//]; right; exists x).
Qed. 

Lemma try_bnd (A B C : Type) (p1 : spec A) 
              (p2 : A -> spec B) 
              (p3 : B -> spec C) 
              (p4 : exn -> spec C) (i : heap) (r : ans C -> heap -> Prop) : 
        verify i (try_s p1 (fun x2 => try_s (p2 x2) p3 p4) p4) r -> 
        verify i (try_s (bind_s p1 p2) p3 p4) r.
Proof.
move=>A B C p1 p2 p3 p4 i r.
case=>h [[i1 i2 H1 <- {i} [H2 [H3 H4]] [<- {h}]] H5].
have T : bind_pre p1 p2 i1.
- split=>// y i.
  move/(H3 y i); move=>[h1 h2 H6 <- {h} [H7 _] _].
  by apply: frame_pre H7.
exists i2; split.
- apply: (starI _ _ H1)=>//.
  split; first by apply: (starI i1 empty); rewrite ?unh0 //; apply: unDl H1.
  split=>x h []; 
    move/(_ i1 empty); rewrite unh0; move/(_ (unDl H1) T (refl_equal _));
    move=>[h1 [->{h} [H7 H6]]]; rewrite unh0 in H6 *;
    case: H7; last by [move=>[_ [[<-] H7]]; move: (H4 _ _ H7)];
    move=>[y [h2 [H7]]] //;
    move: (H3 _ _ H7);
    move=>[h3 h4 H8 H9 [H10 [H12 H13]] _ H11]; subst h2;
    move: (diff2_frame_ex H8 H11 H10)=>[m1 [H14 [<- H16]]];
    by try (move: (H12 _ _ H16)||move: (H13 _ _ H16)); apply: frame_pre.
move=>y _ [h1 h2 H6 -> H7 _ H8].
rewrite -(unKhl H1 H6) in H7=>{h1 H6}. 
apply: H5; apply: star2I=>//.
case: H7; move=>[x [h [[]]]];
move/(_ i1 empty); rewrite unh0; move/(_ (unDl H1) T (refl_equal _));
move=>[h1 [H6 [H9 H10]]] H11; rewrite unh0 in H6 H10; subst h1;
case: H9; try by [move=>[e []] //; move=>[<-]; right; exists x; exists h];
move=>[z [h1 [H12 H13]]]; left;
  exists z; exists h1; split=>//;
  split=>h5 h6 H17 H18 H19; subst h1;
  case: H18=>H18 [H19 H20];
  move: (diff2_frame_ex H17 H13 H18)=>[m1 [H21 [H22 H23]]]; subst h;
  try (move: (H19 _ _ H23)||move: (H20 _ _ H23))=>H24;
  move: (diff2_frame_ex H21 H11 H24)=>[h1 [H25 [H26 H27]]];
  exists h1; do 2![split=>//]; [left|right];
  by exists x; exists m1.
Qed.

Lemma bnd_bnd (A B C : Type) (p1 : spec A) 
              (p2 : A -> spec B) 
              (p3 : B -> spec C) 
              (i : heap) (r : ans C -> heap -> Prop) : 
        verify i (bind_s p1 (fun x2 => bind_s (p2 x2) p3)) r -> 
        verify i (bind_s (bind_s p1 p2) p3) r.
Proof.
move=>A B C p1 p2 p3 i r H.
case: H=>_ [[i1 i2 H1 <- {i} [H2 H3] [<-]] H4].
have T : bind_pre p1 p2 i1.
- split=>// y i; move/(H3 y i); move=>[h1 h2 H5 <- {i} [H6 _] _].
  by apply: frame_pre H6.
exists i2; split.
- apply: (starI _ _ H1)=>//.
  split; first by apply: (starI i1 empty); rewrite ?unh0 //; apply: unDl H1.
  move=>x h [];
    move/(_ i1 empty); rewrite unh0; move/(_ (unDl H1) T (refl_equal _));
    move=>[h1 [->{h} [H7 H6]]]; rewrite unh0 in H6 *.
  case: H7=>[[y [h2 [H7]]]|[e [H9 //]]].
  move: (H3 _ _ H7).
  move=>[h3 h4 H8 H9 [H10 H11] _ H12]; subst h2.
  move: (diff2_frame_ex H8 H12 H10)=>[m1 [H14 [<- H16]]].
  by move: (H11 _ _ H16); apply: frame_pre.
move=>y _ [h1 h2 H6 -> H7 _ H8]. 
rewrite -(unKhl H1 H6) in H7=>{h1 H6}. 
apply: H4; apply: star2I=>//.
case: H7=>[[x [h [[]]]]|[e [-> {y} []]]];
move/(_ i1 empty); rewrite unh0; move/(_ (unDl H1) T (refl_equal _));
move=>[h1 [H6 [H9 H10]]]; rewrite unh0 in H6 H10; subst h1;
case: H9=>[[z [h1 [H12 H13]]]|[e1 [H12 //]]]; 
last by [move: H12=>[<- {e1}] H12; right; exists e];
[move=>H14; left | left];
exists z; exists h1; split=>//;
split=>h5 h6 H17 H18 H19; subst h1; 
case: H18=>H18 H19;
move: (diff2_frame_ex H17 H13 H18)=>[m1 [H21 [H22 H23]]];
[subst h; move: (H19 _ _ H23)=>H24; 
 move: (diff2_frame_ex H21 H14 H24)=>[h1 [H25 [H26 H27]]] | subst h2]. 
- by exists h1; do 2![split=>//]; left; exists x; exists m1.
by exists m1; do 2![split=>//]; right; exists e.
Qed.

Program 
Definition test_try_bnd : STsep bool (stsep.emp, fun y i m => stsep.emp m /\ y = Val false) :=
  Do (ttry ret 4 ;; ret tt then ret false else ret true).
Next Obligation.
by case: H=><-; apply: try_bnd; apply: try_ret; apply: try_ret; apply: val_ret.
Qed.

Program 
Definition test_bnd_bnd : STsep bool (stsep.emp, fun y i m => emp m /\ y = Val false) :=
  Do ((ret 4 ;; ret tt) ;; ret false).
Next Obligation.
by case: H=><-; apply: bnd_bnd; apply: bnd_ret; apply: bnd_ret; apply: val_ret.
Qed.

Program 
Definition test_try_try (e : exn) : STsep bool (stsep.emp, fun y i m => stsep.emp m /\ y = Val true) :=
  Do (ttry ttry ret 4
          then throw[nat] e
          else ret 20
      then ret false
      else ret true).
Next Obligation.
by case: H=><-; apply: try_try; apply: try_ret; apply: try_throw; apply: val_ret.
Qed. 
     
Program 
Definition test_bnd_try : STsep unit (stsep.emp, fun y i m => stsep.emp m /\ y = Val tt) :=
  Do ((ttry ret 4 then ret false else ret true);; ret tt). 
Next Obligation.
by case: H=><-; apply: bnd_try; apply: try_ret; apply: bnd_ret; apply: val_ret.
Qed.
*)

Hint Resolve val_ret try_ret bnd_ret 
             val_read try_read bnd_read
             val_write try_write bnd_write 
             val_alloc try_alloc bnd_alloc
             val_dealloc try_dealloc bnd_dealloc
             val_throw try_throw bnd_throw
             try_do bnd_do (* try_try bnd_try try_bnd bnd_bnd *).

(******************************************)
(* Lemma for dealing with ghost variables *)
(******************************************)
(* In case the postcondition has the special form (forall x. p -> q), and *)
(* p uniquely determines x (i.e., it is invertible), then we can pull p   *)
(* out of the postcondition. *)

Lemma ghost (A B : Type) (i : heap) (p1 : spec A) 
            (p : B -> Prop) (q : B -> ans A -> heap -> Prop) (t : B) : 
        (forall x, p x -> verify i p1 (q x)) -> 
         p t -> 
         verify i p1 (fun y m => forall x, p x -> q x y m).
Proof.
move=>A B i p1 p q t H1 H2.
move: (H1 _ H2)=>[H4 H5].
split=>// y m H x. move/H1=>[H6 H7].
by apply: H7.
Qed.

Implicit Arguments ghost [A B i p1 p q].

Lemma ghost1 (A B : Type) (i : heap) (p1 : spec A)
            (p : B -> Prop) (q : B -> ans A -> heap -> Prop) (t : B) :
        (forall x, p x -> t = x) ->
           verify i p1 (q t) ->
             verify i p1 (fun y m => forall x, p x -> q x y m).
Proof.
move=>A B i p1 p q t H1 [H2 H3].
split=>// y m H x; move/H1=><-.
by apply: H3.
Qed.

Implicit Arguments ghost1 [A B i p1 p q].

Lemma ghost2 (A B : Type) (i : heap) (p1 : spec A)
            (p : B -> Prop) (q : B -> ans A -> heap -> Prop) (t : B) :
        (forall x y m, p x -> q t y m -> q x y m) ->
           verify i p1 (q t) ->
             verify i p1 (fun y m => forall x, p x -> q x y m).
Proof.
move=>A B i p1 p q t H1 [H2 H3].
by split=>// y m H x H4; apply: H1=>//; apply: H3.
Qed.

Lemma binconj (A : Type) (i : heap) (p : spec A) 
              (q1 q2 : ans A -> heap -> Prop) : 
          verify i p q1 -> verify i p q2 -> 
           verify i p (fun y m => q1 y m /\ q2 y m). 
Proof.
move=>A i p q1 q2 [H1 H2] [H3 H4].
by split=>// y m H; split; [apply: H2 | apply: H4].
Qed.

Lemma allconj (A B : Type) (i : heap) (s : spec A) 
              (q : B -> ans A -> heap -> Prop) : 
         (forall x, verify i s (q x)) -> B -> 
         verify i s (fun y m => forall x, q x y m).
Proof.
move=>A B i s q H x.
move: (H x)=>[H1 H2].
split=>//.
move=>y m H3 z.
move: (H z)=>[H1' H2'].
by apply: H2'.
Qed.  

(* packaging up the lemmas into a tactic that selects them appropriately *)             

Definition pull A x (v:A) := (unC (x :-> v), unCA (x :-> v)).
Definition push A x (v:A) := (unCA (x :-> v), unC (x :-> v)).

Ltac hstep := 
  let H := fresh "H" in rewrite -?unA;  
  match goal with 
    | |- is_true (def ?h) -> verify ?h (fr (ret_s _)) _ => 
      move=>H; apply: val_ret=>//; move: H 
    | |- is_true (def ?h) -> verify ?h (fr (try_s (ret_s _) _ _)) _ =>
      move=>H; apply: try_ret=>//; move: H
    | |- is_true (def ?h) -> verify ?h (fr (bind_s (ret_s _) _)) _ => 
      move=>H; apply: bnd_ret=>//; move: H

    | |- is_true (def ?h) -> verify ?h (fr (read_s _ ?l)) _ =>
      move=>H; rewrite -!(pull l) in H *; apply: val_read=>//; move: H
    | |- is_true (def ?h) -> verify ?h (fr (try_s (read_s _ ?l) _ _)) _ =>
      move=>H; rewrite -!(pull l) in H *; apply: try_read=>//; move: H
    | |- is_true (def ?h) -> verify (?h) (fr (bind_s (read_s _ ?l) _)) _ => 
      move=>H; rewrite -?(pull l) in H *; apply: bnd_read=>//; move: H

    | |- is_true (def ?h) -> verify (?h) (fr (write_s ?l _)) _ =>
      move=>H; rewrite -?(pull l) in H *; apply: val_write=>{H}//
    | |- is_true (def ?h) -> verify (?h) 
             (fr (try_s (write_s ?l _) _ _)) _ => 
      move=>H; rewrite -?(pull l) in H *; apply: try_write=>{H}//
    | |- is_true (def ?h) -> verify (?h) 
              (fr (bind_s (write_s ?l _) _)) _ => 
      move=>H; rewrite -?(pull l) in H *; apply: bnd_write=>{H}//

    | |- is_true (def ?h) -> verify ?h (fr (alloc_s _)) _ => 
      move=>H; apply: val_alloc=>//; move: H
    | |- is_true (def ?h) -> verify ?h (fr (try_s (alloc_s _) _ _)) _ => 
      move=>H; apply: try_alloc=>//; move: H
    | |- is_true (def ?h) -> verify ?h (fr (bind_s (alloc_s _) _)) _ => 
      move=>H; apply: bnd_alloc=>{H}//

    | |- is_true (def ?h) -> verify ?h (fr (allocb_s _ _)) _ => 
      move=>H; apply: val_allocb=>//=; move: H
    | |- is_true (def ?h) -> verify ?h (fr (try_s (allocb_s _ _) _ _)) _ => 
      move=>H; apply: try_allocb=>//=; move: H
    | |- is_true (def ?h) -> verify ?h (fr (bind_s (allocb_s _ _) _)) _ => 
      move=>H; apply: bnd_allocb=>{H}//=

    | |- is_true (def ?h) -> verify ?h (fr (dealloc_s ?l)) _ => 
      move=>H; rewrite -?(pull l) in H *; apply: val_dealloc=>{H}//
    | |- is_true (def ?h) -> verify ?h (fr (try_s (dealloc_s ?l) _ _)) _ => 
      move=>H; rewrite -?(pull l) in H *; apply: try_dealloc=>//; move: H
    | |- is_true (def ?h) -> verify ?h (fr (bind_s (dealloc_s ?l) _)) _ =>
      move=>H; rewrite -?(pull l) in H *; apply: bnd_dealloc=>{H}//  

    | |- is_true (def ?h) -> verify ?h (fr (throw_s _ _)) _ =>
      move=>H; apply: val_throw=>//; move: H
    | |- is_true (def ?h) -> verify ?h (fr (try_s (throw_s _ _) _ _)) _ => 
      move=>H; apply: try_throw=>//; move: H
    | |- is_true (def ?h) -> verify ?h (fr (bind_s (throw_s _ _) _)) _ => 
      move=>H; apply: bnd_throw=>//; move: H

    (* special instances when e = empty *)

    | |- is_true (def ?h) -> verify ?h (fr (read_s _ ?l)) _ =>
      move=>H; rewrite -!(pull l) in H *; apply: val_read0=>//; move: H
    | |- is_true (def ?h) -> verify ?h (fr (try_s (read_s _ ?l) _ _)) _ =>
      move=>H; rewrite -!(pull l) in H *; apply: try_read0=>//; move: H
    | |- is_true (def ?h) -> verify (?h) (fr (bind_s (read_s _ ?l) _)) _ => 
      move=>H; rewrite -?(pull l) in H *; apply: bnd_read0=>//; move: H

    | |- is_true (def ?h) -> verify (?h) (fr (write_s ?l _)) _ =>
      move=>H; rewrite -?(pull l) in H *; apply: val_write0=>{H}//
    | |- is_true (def ?h) -> verify (?h) 
             (fr (try_s (write_s ?l _) _ _)) _ => 
      move=>H; rewrite -?(pull l) in H *; apply: try_write0=>{H}//
    | |- is_true (def ?h) -> verify (?h) 
             (fr (bind_s (write_s ?l _) _)) _ => 
      move=>H; rewrite -?(pull l) in H *; apply: bnd_write0=>{H}//

    | |- is_true (def ?h) -> verify ?h (fr (alloc_s _)) _ => 
      move=>H; apply: val_alloc0=>//; move: H
    | |- is_true (def ?h) -> verify ?h (fr (try_s (alloc_s _) _ _)) _ => 
      move=>H; apply: try_alloc0=>//; move: H
    | |- is_true (def ?h) -> verify ?h (fr (bind_s (alloc_s _) _)) _ => 
      move=>H; apply: bnd_alloc0=>{H}//

    | |- is_true (def ?h) -> verify ?h (fr (allocb_s _ _)) _ => 
      move=>H; apply: val_allocb0=>//; move: H
    | |- is_true (def ?h) -> verify ?h (fr (try_s (allocb_s _ _) _ _)) _ => 
      move=>H; apply: try_allocb0=>//; move: H
    | |- is_true (def ?h) -> verify ?h (fr (bind_s (allocb_s _ _) _)) _ => 
      move=>H; apply: bnd_allocb0=>{H}//

    | |- is_true (def ?h) -> verify ?h (fr (dealloc_s ?l)) _ => 
      move=>H; rewrite -?(pull l) in H *; apply: val_dealloc0=>{H}//
    | |- is_true (def ?h) -> verify ?h (fr (try_s (dealloc_s ?l) _ _)) _ => 
      move=>H; rewrite -?(pull l) in H *; apply: try_dealloc0=>//; move: H
    | |- is_true (def ?h) -> verify ?h (fr (bind_s (dealloc_s ?l) _)) _ =>
      move=>H; rewrite -?(pull l) in H *; apply: bnd_dealloc0=>{H}//  

  end.

Lemma blah A (p : loc) (l : A) : def (p :-> l) -> (p :-> l) \In p --> l.
Proof. move=>A q l H. apply/swp. by []. Qed.

Hint Immediate blah.  

Lemma blah2 A (v1 v2 :A) q : def (q :-> v1) -> v1 = v2 -> q :-> v1 \In q --> v2.
Proof. move=>A v1 v2 q D E. apply/swp. split=>//. rewrite E. by []. Qed.

Hint Immediate blah2. 

Ltac hauto D := (do ?econstructor=>//; 
                  try by [move: D; defcheck; auto |
                          eapply blah2; move: D; defcheck; auto])=>//.

Ltac hhauto := (do ?econstructor=>//; try by [heap_congr])=>//.
Ltac hdone := repeat progress hhauto=>//=. 
Ltac heval := do ![hstep | by hhauto].
